/*
   This is the main program. see README
   exact diagonalization of Heisenberg/Hubbard/tJ-systems
   
   ToDo: Simplify the program to make the code better readable!!!!

   startvec = \sum_j c^+_{i_j,s} ... |0> = 4u 2d 3u |0>    
   
   complex? phase factor later added during computation of H(i,j)
    multiply phase to base in a way that Hamiltonian and Vector is reell!?

   set parameters
   <OP> as pointer of OP-function
   static int (*a)(int b); o.ae.

   tU -> ...u..u. * ..d..d divide into up and down part (2 tables)   
   tJ -> ...uu... ...--d.d => (N over u)*(N-u over d)
   JJ -> ...
   
   Warn: operation x<<32 leads to errors for 32bit (no shift)
  
   result of (~0)>>1==~0 != (~0u)>>1  depends on compiler?
   
*/

#include <string.h>
#include <unistd.h> /* gethostname */
#include "block.h"
#include "spins.h"
#include "tbase.h"
#include "ansatz.h"
#include "pipe.h"
#include "vector.h"
#ifdef CONFIG_PTHREAD   /* ToDo: movto mthreads.h, or write new OpenMP-code */
#ifndef _OPENMP /* OPENMP is replacing PTHREAD */
#include <pthread.h>
pthread_attr_t pt_attr;  /* SunOS has SCOPE_PROCESS as default, we need SCOPE_SYSTEM */
pthread_cond_t pt_cond1;  /* conditions master -> childs */
pthread_cond_t pt_cond2;  /* conditions childs -> master */
pthread_mutex_t pt_mutex1; /* mutex */
pthread_mutex_t pt_mutex2; /* mutex */
#endif /* ~_OPENMP */
#else  /* ~PTHREAD dummy definitions to eleminate ifdef constructions */
#define pthread_mutex_lock(x)
#define pthread_mutex_unlock(x)
#endif

#ifdef CONFIG_MPI
#include <mpi.h>
#endif

/* 2015-12-31 testwise, ToDo: pthread-alternative */
#ifdef _OPENMP
#include <omp.h>
#else
#define omp_get_max_threads() 1
#define omp_get_thread_num() 0
#endif

#ifdef CFG_CPUSET // switch on if sched_setaffinity will benefit
/* 2017-01: stackoverflow: use -D_GNU_SOURCE and __USE_GNU is set internaly */
#ifndef __USE_GNU
#  warning "add Makefile: DFLAGS+= -DCFG_CPUSET=1 -D_GNU_SOURCE"
#endif
/* 2017.01.21 sched.h needs __USE_GNU to define CPU_SETSIZE for icc16 */
# include <unistd.h>  // declare sleep()
# include <sched.h>   // declare sched_setaffinity() 
# ifndef _FEATURES_H
#  warning "_FEATURES_H is not defined, try top add #include <features.h>"
# endif
#endif

#include "version.h"

#define Korrelations 1  /* 1=with 0=without spin-correlations */

#ifndef Ham1
#define Ham1	1	/* SiSj - ninj/4 + (ham1)?ninj/4:0 */          
#endif
#ifndef Ham2            /* results differ in U*(N/4-(nu+nd)/2) */
#define Ham2	0	/* 0 = U nu nd ,  1 = U(nu-.5)(nd-.5) */
#endif
#ifndef Ham3
#define Ham3	0	/* 1 = with JH-Oles-term cc and cpcp, better sym */
#endif 


#include "spinstri.h"	/* spinstri.c::outtri() */
#include "mem.h"	/* mymalloc();myfree() */
#include "symmetry.h"	/* lattice symmetries */
#include "diag.h"
#include "myprint.h"	/* maple-Ausgabe */
#include "xval.h"	/* expectation values */
#include "h_file.h"	/* read/write Hamilton matrix on Htmp */
#include "error.h"	/* err_fatal */
#include "hilbert.h"
#include "modeldef.h"

#ifndef Add34
#define Add34 0.0	/* SiSj+Add34*ninj shifting energy, default is 0.0 */
			/* may be +0.75 or -0.25 (or +0.25) is a good idea */
#endif

/* against egcs-2.90.29 980515 (egcs-1.0.3 release)
 * warning: initialization of non-const reference "class a_cplx &" from
 *    rvalue "a_cplx" */
/* ToDo: remove h_xypara+h_zpara, replaced by .wjz[] */
double  h_xypara=1.0;
double  h_zpara=1.0;	/* H=h_zpara*SizSjz+h_xypara*(SixSjx+SiySjy),
				 z-anisotropy Feb99, xy Jun99 */
int     algo;	/* choice of diagonalization + basis generation algorithm
        algo.bit0..2:  0=lanzcos,1=richter,2=wenzel2x2 
        algo.bit3:     8=generate_base_by_H(experimental newns)
        algo.bit4..5:  16=recursive_ns 32=(parallel)seq_linpack_scan
        algo.bit6:     64=MPI_Abort after l1-base-generation
*/
//int     chkpt_mode=0; /* 1=setup after crash possible, 2=continue old: save_mode */
int     checkpoint=0;   /* checkpoint for crash status 0=no, 6=latest */
int	pew;		/* number eigenvalue to check */
int     x_p;		/* output format for vector |x>   (0..3) */
ULong	n1;		/* Matrixsize , n2=Linsize obsolete */
int     n1_bits;        /* bits needed to store n1, used for MPI packing */
double  h_nz=0;         /* number of nonzero matrix elements (statistics) */
ULong   min_b_sn1;      /* pseudo minimum stored n1 of all blocks (MPI) */
#ifndef CFG_FTLM  /* ~ Finite-Temperature Lanczos Method */
double ew[NEW];		/* eingenvalues (Eigenwerte)  (max 4)	*/
#else 
double ew[MAX_ITR];
#endif
ULong  *l2b;		/* Hilfs-Tabellen storeh2 */
ULong opt_perf_test=0; /* 0: normal operation, reduce n1 to p.e. 1e6 */

/* ToDo: norm2 remove it, not necessary (tiny speedup) but complex code */
// ATnorm2 *norm2=NULL;	/* vector idx->sym-idx, norm(sym-idx)	*/
// minfo_t m_n2 = {NULL,0,0,"tmp/n2.map",NULL}; /* norm2 companion */
// int base2ok=0;			/* flag: norm2 not yet inititialized */


ULong hfmax=0;	/* max. num of hxy[]-elements on disk, 0 = unlimited */
#if 0
FILE *hfile,*fhxy1,*fhxy2;	/* File enthaelt Hamiltonien		*/
#endif
/* ToDo: 2013-04 better use system limits? joblimit.vmem=... or ulimit -v ... */
 /* ULong maxmem; */ /* maximum memory for vectors -> no paging! removed 2013 */

int cfg_sisj;	/* option for correlations, bit0=SiSj bit1=SS bit2=ZiZj(?) */

tbase1 p_mask[NN][NN];		/* bit masks for parity()  */

/* max/min/sum hnz_blocks per thread (balance in storeh/geth_block) */
ULong bsnz_max, bhnz_max, bhnz_min, bhnz_sum;
ULong mpi_traffic=0,     /* MPI statistics, data volume (/band_width) */
      mpi_syncs=0,       /* number of MPI transactions (*latency) */
      mpi_maxlen=0,      /* maximum len data block in units scplx/ulong */
      /* mpi statistics packets with packetsize up to: 1k 4k 16k 64k 256k 1M */
      mpi_stats_npkts[8];   /* number of packets with size < 1k<<(2*idx) */
/* old: mpi_z_pkts=0; */ /* number of zero blocks, wasted communications */

char nodename[22]="\0HOSTNAME............"; /* 2016-04-08 set global for err() */

/* local variables for pthreads (tasks), communication to/from tasks */
typedef struct {
  int num;     /* to identify itself (ro) */
  void *data;  /* to send data-block (set by thread 0, read by workers) */
  int  task;   /* master set 1 to run task + waits for 0, task sets 0 @end */
  void (*worker)(void *data); /* parallel working function */
  /* mutex for each thread? */
} main_pts;
volatile main_pts main_pt[B_NUM]; /* [un]locked by mutex1 */

#ifdef CONFIG_PTHREAD
/* MPI code is in not thread safe in general, we have to care about it */
//#ifndef _OPENMP
pthread_mutex_t token_ring[B_NUM*4]; /* of double(quad) length */
//#endif

/* send jobtype (task) to threads and restart them */
/* 2016-10 g++-6.2 + pthread + blas segfaults for stacksize 8K (ulimit -s) */
void parallel_start( int thread_i, void (*worker)(void *data), void *data ){
#ifndef _OPENMP
    pthread_mutex_lock(&pt_mutex1);
#endif
    // printf("DBG: send work to task %d\n", thread_i);
#pragma omp critical
{
    main_pt[thread_i].data=data;
    main_pt[thread_i].worker=(/*volatile*/ void (*)(void *))worker;
    main_pt[thread_i].task=1; /* signal worker thread i to run (last!) */
}
    /* ToDo: do next only once after storing all tasks
             or send signal only to the job */
#ifndef _OPENMP
    pthread_cond_broadcast(&pt_cond1); /* data send to threads */
    pthread_mutex_unlock(&pt_mutex1);
#else
#pragma omp flush (main_pt)
#endif
}

/* waiting for threads */
void parallel_wait( int thread_i ){
#ifndef _OPENMP
  pthread_mutex_lock(&pt_mutex1);
  while ( main_pt[thread_i].task )  /* volatile! */
    pthread_cond_wait(&pt_cond2,&pt_mutex1); /* unlock+wait_for_child+lock */
  pthread_mutex_unlock(&pt_mutex1);
#else
  while ( main_pt[thread_i].task )  /* volatile! */
    usleep(10000); /* ToDo  */
#endif
}
#else /* ~CFG_PTHREAD dummy parallel functions (serialized pthreads) */
void parallel_start( int thread_i, void (*worker)(void *data), void *data ){
  worker(data); }
void parallel_wait( int thread_i ){}
#endif /* CFG_PTHREAD */

#if defined(__SICORTEX__)
/* Alltoallv is 1.4x(600tasks) to 15x(5700tasks) slower on SiCortex2009 */
/* try this if you have same problems on your system */
#define USE_Alltoallv_Emulation 1
#else
#define USE_Alltoallv_Emulation 0 /* better on IB.mpich3, IB.ompi18 */
//#define USE_Alltoallv_Emulation 1 /* testing correctness of code only */
#endif

#if CONFIG_MPI   /* 2017-02-01 */
#if USE_Alltoallv_Emulation  /* added 2017-01-17 slow SiCortex-mpich2.A2av */
int My_MPI_Alltoallv(
  const void *d1,const int *cnt1,const int *ofs1,MPI_Datatype t1,
  const void *d2,const int *cnt2,const int *ofs2,MPI_Datatype t2,MPI_Comm cm)
{ int n1, n2, len1, len2, z1, z2, rc=0, id; MPI_Status status; void *p1, *p2;
  MPI_Type_size(t1, &z1); MPI_Type_size(t2, &z2); /* get data-size */
  /* correct? or use type MPI_BYTE only! */
  for (id=0;id<mpi_n;id++){ /* pre2017-Alltoallv-emulation */
    n1=(mpi_i+id)%mpi_n;  n2=(mpi_i+mpi_n-id)%mpi_n; /* sendto + rcvfrom-node */
    len1=cnt1[n1];      p1=&((char *)d1)[ ofs1[n1]*z1 ];
    len2=cnt2[n2];      p2=&((char *)d2)[ ofs2[n2]*z2 ];
    rc|=MPI_Sendrecv(p1,len1,t1,n1,0,
                     p2,len2,t2,n2,0,cm,&status);  } // for node00..
  return(rc); }
#else   /* default implementation, icc16 needs args + line breaks "\\" */
/* #define My_MPI_Alltoallv(d1,c1,o1,t1,d2,c2,o2,t2,cm) \
              MPI_Alltoallv(d1,c1,o1,t1,d2,c2,o2,t2,cm)  */
#define My_MPI_Alltoallv    MPI_Alltoallv
#endif
#endif /* CONFIG_MPI */

/* spins.h: inline double norm(double r){return(r*r);} */
char *sclock(long t){  /* 0..960s 16m..960m 16h... => 00:00:00 */
   static char s[64]; if (t==~0l) t=itime();
   char s2[24];
   /* this format is a mixture for simple parse and human read */
   /* this format is human readable but not simply enough to parse */
   if(t>4*60*60l)sprintf(s2,"%ldh",(t+1799)/3600); else
   if(t>   4*60l)sprintf(s2,"%ldm",(t+29)  /  60); else sprintf(s2,"%lds",t); 
   /* 0.1-9999.9min is best compromize, 5s..150h=6d */
   /* a recursive call like sprintf(s,"%s",s) is dangerous */
   sprintf(s,"%6.2f =%s",t/60.,s2);
   return(s);}

char *outc(cplx r){  /* output [complex] float number */    
    static char s1[100];
#if VecType & 4
    double phase=arg(r)/M_PI; if (phase<0) phase+=2;
    /* sprintf(s1,"(%g,%g)",real(r),imag(r)); -> |r|*exp(phase*I*Pi) */
    if (Cabs(r)<0.0001 || Cabs(r)>9.99999) /* 2017.1 fabs to Cabs */
         sprintf(s1,"%12.5e*phase(%5.3f)",Cabs(r), phase);
    else sprintf(s1,"%12.9f*phase(%5.3f)",Cabs(r), phase);
#else /* % 12.9f=-0.123456789 vs. % 12.5e=-1.12345e-09 */
    if (fabs(r)<0.0001 || fabs(r)>9.99999)
         sprintf(s1,"% 12.5e",D(r));
    else sprintf(s1,"% 12.9f",D(r)); /* 12.9f or 12.5e */
#endif
    return(s1);}

void nl(FILE *f){ fprintf(f,"\n");return;}
/* output eigenvalues */
void outew(FILE *f, const double *ew, const int n_ew){
   int k;
   for (k=0;k<n_ew;k++) fprintf(f," %12.8f", ew[k]);
   return;
}


/* ---------------------------------------------------- *
 *      statistics about hamiltonien  (sym,num_nz,...)	*
 *	-	-	-	-	-	-	
 *  double counts to 2^53=9e15 (54bit+10bit) exp e+307(9bit)
 *  
 */
struct tstatH {    
  /* max. and 2nd max number nz elements of a line (weight to?) */
  ULong nzxmin,  nzxmax,  nzxmin2,  nzxmax2,  /* nz per line */
        nzxminw, nzxmaxw, nzxmin2w, nzxmax2w; /* density or weight of nzx */
  cplx sum12, sum21;  /* sum-upperright,lowerleft r to detect errors */
/* dbl hh12, hh21;  +=<r|r> removed, some elements of H may not be summed
# s=1 N=27 triangle 53+1 n1=2 2012-01-26
#   min,max zaehlt nicht summierte Elemente, deshalb sum += r**2 != 0
# sum_h12   hh12 :  1.414213538*phase(1.000)  3.99999960e+00
# sum_h21   hh21 :  1.414213538*phase(1.000)  1.99999980e+00
# sum h21-h12^*  :  0.00000e+00*phase(0.000) -1.99999980e+00 ToDo: wegnehmen! 
# Stoerungstheorie (Perturbation theory)?
#   1. lowest diag, 2. follow biggest non-diag to nearest diag, ... ?? ToDo
 */
      double traceH,       /* Trace of H for manual error detection */
             maxH, maxH12; /* maximum absolute diagonal and off-diagonal element */
      /* statistics for data transfers within and between nodes */
      /* MPI_Allreduce dont allow LONG_LONG */
      double hnz_nd,	  /* number of non-diagonal elements */
             hnz_d,	  /* number of diagonal elements     */
             hnz_local_thread,
             hnz_inter_thread,
             hnz_inter_node; /* ToDo: [nodes] for min/max/mean%??? */
      } 
      statH[B_NUM]; /* ToDo: -> statH[B_NUM] distinct cache lines */
            
void statH_ini(){ /* reset statistic counters */
 int i;
 for(i=0;i<pt_n;i++){ /* thread safe */
   statH[i].nzxmax=0l;
   statH[i].nzxmin=999999l;
   statH[i].nzxmax2=0l;
   statH[i].nzxmin2=999999l;
   statH[i].nzxmaxw=0l;
   statH[i].nzxminw=0l;
   statH[i].nzxmax2w=0l;
   statH[i].nzxmin2w=0l;
   statH[i].sum12=C(0,0);
   statH[i].sum21=C(0,0);
   statH[i].traceH=0.0;
   statH[i].maxH=  0.0;
   statH[i].maxH12=0.0;
   statH[i].hnz_d=0;
   statH[i].hnz_nd=0;
   statH[i].hnz_local_thread=0;
   statH[i].hnz_inter_thread=0;
   statH[i].hnz_inter_node  =0;
 }
}
/* called by hamilton_puth_line() on each line for statistic analysis
 * also check for potential errors
 * ToDo: replace by tbx=bx%pt_n (thread block) and gbx=tbx+mpi_i*pt_n (global block)
 *            or    tbx=bx & (1<<pt_n_bits)-1;
 * called per num_ahead_lines
 */
void add_statH(thbuf *hbuf){
  double rr;
  ULong xy, x, n_old, y;
  int node_x, node_y, thread_x, thread_y, ii; /* was bx, by */
  cplx r;
  thread_x=hbuf->block;
  node_x=mpi_i;
  if (thread_x>=pt_n)     F1("thread_x>=pt_n thread_x=%d", thread_x);
  x=hbuf->i - hbuf->i%hbuf->num_ahead_lines; /* ??? */
  // ToDo: check x > n1+num_ahead_lines !?
  if (x>=n1 || x>=node_n1) return; // happens if i+ahead > n1
// if (x>=n1) F1("x>=n1 x=%.0f",1.*x)
// if (x>=node_n1) F1("x>=node_n1 x=%.0f",1.*x)
  if (x>=b_ofs[thread_x+1]) {
     static int reported=0;
     if (!reported)
       printf("\n %d.%d: x>=b_len   x=%.0f (may happen if i+ahead>n1)",
          mpi_i, thread_x, 1.*x);
     reported=1;
     return; // happens on extra calls for balancing???
  }
  x+=mpi_ofs[node_x]+b_ofs[thread_x]; /* x=0..n1-1, ToDo: remove? */
  n_old=0;
  if ((vvv&64) && mpi_i==0 && x<x_max)
    printf("\nvvv&64 add_statH.addh x_max=%ld hbuf.n=%d n1=%ld:",
           x_max, hbuf->n, (long)n1);

  /* synchronize mpi output? */

  for (ii=0;ii<hbuf->n;ii++) {  /* for each hnz element */
    xy     = hbuf->jj[ii];   /* 0..dest_node_n1-1 */
    node_y = hbuf->abj[ii]; /* destination node 0..mpi_n-1 */
    if (node_y>=mpi_n) ERR(9999); /* never happens? ~init! */
    r      = hbuf->rr[ii];
    y      = xy & ~XY_FLAG;
    for (thread_y=0; thread_y<pt_n; thread_y++)
      if (y<b_ofs[thread_y+1]) break; /* bad: we need remote table here */
    y     += mpi_ofs[node_y]; /* 0..n1 */

    if (  y>=n1 || node_y>=mpi_n
       || y>=mpi_ofs[node_y+1]
       || y< mpi_ofs[node_y  ]) {
      printf("\nERR.statH: ii=%d of %d x,y,rr= %d.%d.%lu %d.%d.%lu"
             " % .3e %+.3ei (bad y)",
             ii, hbuf->n, node_x, thread_x, x, node_y, thread_y, y,
              D(r), D(Im(r)));
    }

    if ((vvv&64) && x<x_max && (Norm(r)>1e-5 || x==y)){
      /* show 1..x_max H-lines, 2015 skip zero off-diag elements */
      /* printf("\n x=%8lu y=%8lu h_xy=%s",x,y,outc(r)); --- old --- */
      /* try to find exact expressions for maple/maxima input */ 
      if (ii==0) sleep(mpi_i*pt_n+hbuf->block+1); /* kind of sync */
      printf("\n addh(%5lu,%5lu, ",x,y);
      outmpl(stdout,D(r));			/* +(a/b)^(1/2) */
#if VecType & 4
      printf(","); outmpl(stdout,Im(r));	/* +(a/b)^(1/2) */
#endif
      printf("):");				/* for maple */
    }
    if (thread_y==thread_x && node_x==node_y)
                          statH[thread_x].hnz_local_thread++; else
    if (node_y==node_x)   statH[thread_x].hnz_inter_thread++; else
                          statH[thread_x].hnz_inter_node++;
    if (x> y) statH[thread_x].sum12+=r;	     /* summe of half H's */
    if (x< y) statH[thread_x].sum21+=r;	     /* should be conjugated value */
    if (x==y) statH[thread_x].traceH+=Re(r); /* should be base invariant */
    rr=Cabs(r);				/* cplx => real != sqrt(Norm()) */
    // if (n1<32 && vvv && x==y && rr)  /* 2015-12 */
    //  printf("\nDBG: Hdiag %2d %s", (int)x, outc(r));
    if (!(rr<1e36)) { /* maxF32=3.4e38 2016-01 p.e. float=(double)1e50 highS */
      /* if using f32 + Tnorm2=f64 (big-s1-systems, try to use double only
       *    found it on N=12 S=5/2 12+48 n1= 25751  2016-01 */
      /* no idea how to solve it, for N=12 S=5/2 */
      printf("\n statH x=%5lu hbuf.ii=%d y=%5lu r=%e",x,ii,y,D(r));
      E0(                 "rr>1e36  (f32 overflow)");
#if VecType & 4 == 0  /* float32 */
      F0(" have f32 only, try f64 or other solutions");
#else  /* float64 */
      if (!(rr<1e300)) F0("rr>1e300 (f64 overflow)");
#endif
    }
    if (rr) {
      if (x==y && statH[thread_x].maxH<rr)  /* max diagnal element */
                  statH[thread_x].maxH=rr;
      if (x!=y && statH[thread_x].maxH12<rr) /* max off-diagonal element */
                  statH[thread_x].maxH12=rr;
      if (x==y)   statH[thread_x].hnz_d++;   /* count diagonal elements */
      else        statH[thread_x].hnz_nd++;  /* count off-diagonal elements */
    }
    if (xy&XY_FLAG) { /* EOL, start next matrix line */
//printf("\nDBG nzx=%lu n_old=%lu ii=%d x=%lu",(ULong)ii+1-n_old,n_old,ii,x);
      if (statH[thread_x].nzxmax==ii+1-n_old) {
          statH[thread_x].nzxmaxw++; }
      if (statH[thread_x].nzxmin==ii+1-n_old) {
          statH[thread_x].nzxminw++; }
      if (statH[thread_x].nzxmax2==ii+1-n_old) {
          statH[thread_x].nzxmax2w++; }
      if (statH[thread_x].nzxmin2==ii+1-n_old) {
          statH[thread_x].nzxmin2w++; }
      if (statH[thread_x].nzxmax<ii+1-n_old) {
          statH[thread_x].nzxmax2= statH[thread_x].nzxmax;
          statH[thread_x].nzxmax2w=statH[thread_x].nzxmaxw; 
          statH[thread_x].nzxmax=ii+1-n_old;
          statH[thread_x].nzxmaxw=1; }
      if (statH[thread_x].nzxmin>ii+1-n_old) {
          statH[thread_x].nzxmin2= statH[thread_x].nzxmin;
          statH[thread_x].nzxmin2w=statH[thread_x].nzxminw; 
          statH[thread_x].nzxmin=ii+1-n_old;
          statH[thread_x].nzxminw=1; }
      if (statH[thread_x].nzxmax> ii+1-n_old
       && statH[thread_x].nzxmax2<ii+1-n_old ) {
          statH[thread_x].nzxmax2=ii+1-n_old;
          statH[thread_x].nzxmax2w=1; }
      if (statH[thread_x].nzxmin< ii+1-n_old
       && statH[thread_x].nzxmin2>ii+1-n_old ) {
          statH[thread_x].nzxmin2=ii+1-n_old;
          statH[thread_x].nzxmin2w=1; }
      x++; n_old=ii+1;
    }
  }
  return;
}

/* 2015-05-28 bug 11.05.15 n1=3 nzx=3.67
   sum_h12/n1   :  1.85037e-17*phase(0.500)
   sum_h21/n1   :  1.86159e-17*phase(0.535)
   sum_h21-12^* :  1.11191e-16*phase(0.518) not hermitian + abort
   fix: hermitian + return error if not hermitian (old void)
   return: 0=OK, 1=(n1==0) 2=not_hermitian 
 */
int statH_out(){		/* output statistics vvv&2 */
  int i;
  if (n1==0) return 1;
  for(i=1;i<pt_n;i++){      /* sum all threads to statH[0] */
    statH[0].hnz_local_thread += statH[i].hnz_local_thread;
    statH[0].hnz_inter_thread += statH[i].hnz_inter_thread;
    statH[0].hnz_inter_node   += statH[i].hnz_inter_node;
    statH[0].sum12  += statH[i].sum12;
    statH[0].sum21  += statH[i].sum21;
    statH[0].traceH += statH[i].traceH;
    if (statH[0].maxH    < statH[i].maxH)
        statH[0].maxH    = statH[i].maxH;
    if (statH[0].maxH12  < statH[i].maxH12)
        statH[0].maxH12  = statH[i].maxH12;
    statH[0].hnz_nd   += statH[i].hnz_nd;
    statH[0].hnz_d    += statH[i].hnz_d;
    if (statH[0].nzxmax  == statH[i].nzxmax)
        statH[0].nzxmaxw += statH[i].nzxmaxw;
    if (statH[0].nzxmin  == statH[i].nzxmin)
        statH[0].nzxminw += statH[i].nzxminw;
    if (statH[0].nzxmax  <  statH[i].nzxmax) {
        statH[0].nzxmax  =  statH[i].nzxmax;
        statH[0].nzxmaxw =  statH[i].nzxmaxw; }
    if (statH[0].nzxmin  >  statH[i].nzxmin) {
        statH[0].nzxmin  =  statH[i].nzxmin;
        statH[0].nzxminw =  statH[i].nzxminw; }
  }
#ifdef CONFIG_MPI
  { double add[7], sum[7];
    add[0]=0; add[1]=0; // unused 2012-01-26
    add[2]=statH[0].traceH;
    add[3]=Re(statH[0].sum12); add[4]=Re(statH[0].sum21);
    add[5]=Im(statH[0].sum12); add[6]=Im(statH[0].sum21);
    MPI_Allreduce(add, sum, 7, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    statH[0].traceH=sum[2];
    statH[0].sum12 =C(sum[3],sum[5]);
    statH[0].sum21 =C(sum[4],sum[6]);
  }
  { double add[7], sum[7];
    add[0]=statH[0].hnz_local_thread;
    add[1]=statH[0].hnz_inter_thread;
    add[2]=statH[0].hnz_inter_node;
    add[3]=statH[0].hnz_nd;
    add[4]=statH[0].hnz_d;
    MPI_Allreduce(add, sum, 5, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    statH[0].hnz_local_thread=sum[0];
    statH[0].hnz_inter_thread=sum[1];
    statH[0].hnz_inter_node  =sum[2];
    statH[0].hnz_nd=sum[3];
    statH[0].hnz_d =sum[4];
  }  
  { double add[2], sum[2];
    add[0]=statH[0].maxH;
    add[1]=statH[0].maxH12;
    MPI_Allreduce(add, sum, 2, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
    statH[0].maxH=  sum[0];
    statH[0].maxH12=sum[1];
  }  
  { ULong add[7], sum[7];
    add[1]=statH[0].nzxminw;
    add[2]=statH[0].nzxmaxw; // num min/max = weight
    add[0]=statH[0].nzxmax;
    MPI_Allreduce(add, sum, 1, MPI_UNSIGNED_LONG, MPI_MAX, MPI_COMM_WORLD);
    if (statH[0].nzxmax!=sum[0]) add[2]=0;
    statH[0].nzxmax=sum[0];
    add[0]=statH[0].nzxmin;
    MPI_Allreduce(add, sum, 1, MPI_UNSIGNED_LONG, MPI_MIN, MPI_COMM_WORLD);
    if (statH[0].nzxmin!=sum[0]) add[1]=0;
    statH[0].nzxmin=sum[0];
    add[0]=0; /* unused */
    MPI_Allreduce(add, sum, 3, MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
    statH[0].nzxminw=sum[1];
    statH[0].nzxmaxw=sum[2];
  }
#endif
  /* MPI_ToDo: sum over all nodes, may fail for multithreads */
  if (!mpi_i) {
    double sum_hnz=statH[0].hnz_local_thread
                  +statH[0].hnz_inter_thread
                  +statH[0].hnz_inter_node  ;
    printf("\n vvv&2 statistics for sparse matrix H:");
    printf("\n hnz transfer local,pthread,mpi= %3.0f%% %3.0f%% %3.0f%%",
      statH[0].hnz_local_thread*100./sum_hnz,
      statH[0].hnz_inter_thread*100./sum_hnz, 
      statH[0].hnz_inter_node  *100./sum_hnz);
    /* printf("\n hnz_ndiag>eps : %9.0f",statH[0].hnz_nd); */
    /* printf("\n hnz_diag      : %9.0f",statH[0].hnz_d); */
    /* ToDo: counting long.numlines[nzx=1..1+min(nu,nd)*2*Nw/N]++ is cheaper 
     *    to compute than min2,max2 or more sorting (use [... 1+Nw]) */
    printf("\n hnz_x_min,w  : %9lu %14lu",statH[0].nzxmin,statH[0].nzxminw);
    /* min2,max2 correct for mpi=1 pt=1 only */
if (mpi_n==1 && pt_n==1) {
    printf("\n hnz_x_min2   : %9lu %14lu",statH[0].nzxmin2,statH[0].nzxmin2w);
    printf("\n hnz_x_max2   : %9lu %14lu",statH[0].nzxmax2,statH[0].nzxmax2w);
}
    printf("\n hnz_x_max,w  : %9lu %14lu (.le.Nw)",statH[0].nzxmax,statH[0].nzxmaxw);
    /* v2.27 x_max ca. min(u,d)*nw+1?, x_min ca. minAO(u,d)+1? */
    printf("\n hnz/n1 * n1  : %12.2f * %9lu", 1+statH[0].hnz_nd*1./n1, n1);
    /* 2014-06 trace normalized to mean eigenvalue, for a1,b1 range check? */
    printf("\n maxH         : %12.9f", statH[0].maxH);
    printf("\n maxH12       : %12.9f", statH[0].maxH12);
    printf("\n trace_H/n1   : %12.9f", statH[0].traceH/n1);
    printf("\n sum_h12/n1   : %s", outc(statH[0].sum12/n1));
    printf("\n sum_h21/n1   : %s", outc(statH[0].sum21/n1));
    /* debugging: */
    printf("\n sum_h21-12^* : %s", outc(statH[0].sum21-Conj(statH[0].sum12)));
    /* if its not hermitian, convergation is not ensured */
    /* 2015-05-28 +1 against near zero rounding errors */
    if(    Cabs(statH[0].sum12 -Conj(statH[0].sum21))
    >1e-8*(Cabs(statH[0].sum12)+Cabs(statH[0].sum21)+1))E0("H not hermitian");
    if(    Cabs(statH[0].sum12 -Conj(statH[0].sum21))
    >1e-2*(Cabs(statH[0].sum12)+Cabs(statH[0].sum21)+1)){E0("assume bad bug");
      return 2;}
  }
  return 0;
}
/* --------------------------------------------------------------- */  




#ifndef NUM_AHEAD_LINES
#define NUM_AHEAD_LINES 1  /* minimum, bigger is better, nhm_buffer_lines */
#endif
/* see spins.h   thread buffer:
   typedef struct shbuf { ... nhn[NAH],el[NZX*NAH]} thbuf; */
/* thbuf *buf1; buf1=alloc_hbuf() ... reset_hbuf(buf1) ... free_hbuf(buf1) */
int reset_hbuf(thbuf *buf){
  if (!buf) {ERR(9502); return -1;} /* must be alloc before reset */
  buf->n=0;	/* ini num_elements_in_buf, returned/used by nhv */
  buf->num_ahead_lines=NUM_AHEAD_LINES; /* set before nhv called */ 
  buf->block=-1;  /* line belongs to block-nr, -1 means unused, before nhv */
  buf->v0=(tvec *)NULL; /* store/stat or nHn, set before nhv called */
  buf->v1=(tvec *)NULL; /* v1+=H*v0 if !0, set by the nhv caller */
//  buf->sci.o=0;   /* option: 0=fullbase,1=minsymcfg,2=isminsymcfg */
//  buf->scj.o=0;
//  buf->sci.num=0;
//  buf->scj.num=0;
  buf->rx=0.0;  /* <nHv>, used in xval.c to compute <v0|OP|v0> */
  /* nhm_line DBG needs initialization 2013-02 valgrind */
  {int i;for(i=0;i<NUM_AHEAD_LINES;i++)buf->nhv[i]=buf->nhn[i]=0;}
  return 0;
}

thbuf *alloc_hbuf(){
  thbuf *buf;
  /* ToDo: dynamic mem alloc for hbuf->el if needed */
  buf=(thbuf *)mymalloc( sizeof(thbuf), NULL); /* no msg,  */
  if (!buf) {
    printf("\nERROR alloc_hbuf: malloc(thbuf=%.2e) failed, mpi=%d",
            1.*sizeof(thbuf), mpi_i);
    return 0;
  }
#if 0
  buf->sci.ib= (tbase *)NULL;  /* force SIGSEGFAULT if [] accessed */
  buf->sci.is=(tphase *)NULL;
  buf->scj.ib= (tbase *)NULL;  /* force SIGSEGFAULT if [] accessed */
  buf->scj.is=(tphase *)NULL;
  /* really needed? */
  buf->scj.ib= (tbase *)malloc(sizeof(tbase) *MaxSBase);
  buf->scj.is=(tphase *)malloc(sizeof(tphase)*MaxSBase);
#endif
  reset_hbuf(buf); /* set initial values */
  return buf;
}

void *free_hbuf( thbuf *hbuf ){
#if 0
  if (hbuf->sci.ib) free(hbuf->sci.ib); /* avoid memory leak */
  if (hbuf->sci.is) free(hbuf->sci.is);
#endif
  return myfree( (void *) hbuf, NULL );
}


/* store elements of one H-line to the HBlock files on disk 
 * input: hbuf.i(b_ofs[blk]...,>node_ofs) + hbuf.n + hbuf.jj+abj+rr[0..n-1]
 * output: H-file + Statistics
 * called by _nhv if m1&2 = storeh, 
 * ToDo: return !0 if not written, for output write/store fail at XX% 2016-01
    or do we have ofs in hbuf to output ofs/n1 as % ???
 */
void hamilton_puth_line(thbuf *hbuf){ /* renamed from nhmline v2.26 */
  int ii, blk, by;
  ULong xy; cplx r; HR_TYPE rr; /* h_put-variables */
  blk=hbuf->block;
  /* vvv&2 makes SH factor 2 slower on V490-sparcIV+ cache? v2.25? */
  if (vvv&2) add_statH(hbuf); /* H-statistics for debugging */

  /* b_hnz - number of hamilton-elements per compute block in thread */
  /* hfmax - maximum file space for H per block */
#if 0
  if(vvv&16) {printf("\nvvv&16 DBG L564 n=%d", hbuf->n);fflush(stdout);}
#endif
  if (b_hnz[blk]+hbuf->n < hfmax || !hfmax)
#if  STOREH == 0
  if (pipeHbuffer[blk] && pipeHlen[blk]>0)  /* 2015-04-13 no error messages */
#endif
  if (h_error==0) { /* store Hij, ToDo: per line */
    for(ii=0;ii<hbuf->n;ii++) {
      /* store a matrix element (xy=y_idx,r=Hij) into the file buffer b_hbuf
       * and write full buffer to the file; keywords: hput puth put_h h_put */
      xy=hbuf->jj[ii]; /* 0..(b_len[by]-1)[+XY_FLAG] checked in statH */
      by=hbuf->abj[ii];  /* 0..(pt_n*mpi_n-1) */
      r =hbuf->rr[ii];
#if defined(DEBUG)
#warning "DEBUGGING puth_line on"
      if (by>=mpi_n) ERR(9999); /* 2016-04 ~init abj */
      if ((xy&~XY_FLAG)>=b_len[by]) { printf("\nDBG blk %2d %2d %8ld",blk,by,xy); ERR(9999); }
#endif
#if HRMAX>0    /* use a 8-16bit index table to double to safe disk space */
      { int j; /* get hr_index or store as new hr_index else */
        /* Note: Norm(hr-r)<1e-13 can fail for some model systems, */
        /*       but hr==r is probably unecessary strong, no speed down */
        for(j=0;j<hr_len[blk];j++) if(Norm(hr[blk][j]-r)<=1e-13) break;
        /* here we should store max(hr[][]-r)<1e-13, but this takes time */
        if (j==hr_len[blk]) { /* not found */
           hr[blk][j]=r; hr_len[blk]++;
           if(hr_len[blk]>=HRMAX) err_fulltable(); /* OVL */
        }
        rr=(HR_TYPE)j;		/* store index to value */
      }
#else    
      rr=M_d2f(r);		/* store value direct, ToDo: d2f needed? ToDo */
#endif    
      /* blk is threadnum to ensure cpu local memory,
       *   struct thxy b_hbuf defined in block.c and used in h_file.c (not good) */
      if (b_bnz[blk]>=HBLen) F2("OO b_hbuf blk=%d i=%d", blk, b_bnz[blk]);
      b_hbuf[blk].rr[ b_bnz[blk] ]=rr;
      b_hbuf[blk].by[ b_bnz[blk] ]=(BY_TYPE)by;
      b_hbuf[blk].xy[ b_bnz[blk] ]=(XY_TYPE)xy;
// printf("\nDBG puth by,y,rr,flag= %d %lu %g %lu", by, xy&~XY_FLAG, r, xy&XY_FLAG);
      /* store in Hblock-buffer, ToDo: store only complete lines? */
      b_bnz[blk]++; /* count buffered elements, store at buffer too? */
      if (xy&XY_FLAG) b_bn1[blk]++; 
/* ToDo 2014-07: better fill rest with zeros (dummy) 
     and store buf if next line will not fit to the buffer or last line? 
     usually HBLen=32k(*8) nline=13-29=25 possible-overhead<29/32k
 */
      if (b_bnz[blk]==HBLen){   /* flush buffer to disk-file */
        if (h_error) break; /* break ii loop */
        if (my_fwrite((void *)&b_hbuf[blk],Hsize,1,blk)!=1) {
          /* ToDo: 2014-07 here we have a big problem! it FAILES */
          /* ToDo: rollback everything within this block!!! */
          hfmax=b_snz[blk]; /* ToDo 2014: cmp + set to mpi_min */
          b_bn1[blk]=0; b_bnz[blk]=0;
          if((vvv&16) && blk+mpi_i*pt_n<8)
            printf("\nvvv&16 h_write failed task=%2d.%2d sn1=%lu nz/HB=%lu",
               mpi_i,blk,b_sn1[blk],b_snz[blk]/HBLen);
          break; /* hard write error, stop writing (but mpi imbalance) */
          /* ToDo: breaks MPI balance?, store dummy block! */
        } /* if write failed */
        /* reset counter buffered_nonzero-entries */
        b_sn1[blk] += b_bn1[blk]; b_bn1[blk]=0; /* not sharp!? */
        b_snz[blk] += b_bnz[blk]; b_bnz[blk]=0; 
      } /* if HBLen full */
    } /* for ii = 0 .. hbuf->n */
#if 0 // 2014-07-25
    if (ii==hbuf->n) { /* no error in ii loop (break)? 2014-07-24 */
      /* snz = stored nz, sn1 = stored n1, ToDo: bnz = buffered nz, bn1 ... */
      b_snz[blk] += hbuf->n; /* count stored elements, full ahead lines ToDo: full line */
      if (hbuf->i+hbuf->num_ahead_lines <= b_ofs[hbuf->block+1])
           b_sn1[blk]+=hbuf->num_ahead_lines; /* count lines stored */
      else b_sn1[blk]= b_ofs[hbuf->block+1]; /* maximum stored */
    }
#endif
  } /* if b_hnz < hfmax */
  /* hfmax is set by hand or by malloc error above */
  if (b_hnz[blk]        <=hfmax
   && b_hnz[blk]+hbuf->n>=hfmax  && hfmax
//   && mpi_i*pt_n+blk<4   // it may be essential seeing this errors too!
   ) { // assuming that this error will appear on all nodes, ToDo: MPI_send err
     /* 2015-04-09: ERR(630) generates lot of outputs on 1000 threads */
     /* ERR(630); */ /* v2.33, maxfile reached, stop writing, ToDo: free mem/dsk */
     if (mpi_i*pt_n+blk<8) printf(
       "\nWARN: %2d.%2d nz==hfmax, b_hnz=%4lu*HB+%5lu %6.2f%% %4lu*AH n=%.0f",
         // %% added 2016-01 not sure if node_n1 is correct or b_ofs[blk+1]-b_ofs[blk]-1
       mpi_i,blk,b_hnz[blk]/HBLen,b_hnz[blk]%HBLen,
       hbuf->i*100./max_b_len, hbuf->i/NUM_AHEAD_LINES, hbuf->n*1.);
      if(mpi_n*pt_n>1 && !mpi_i)
    printf("\nWARNING stored_nz/HBLen>0 and rest_n1/AH>0 may be buggy 2017\n");
    // strange effects seen, of randomly big bfactors, or crashs SH=1..99%
    // better store AH*NZX-blks, or fill AH-buffers from storage
  } // if hfmax
  if (b_hnz[blk]+hbuf->n<b_hnz[blk]) ERR(610); /* counter overflow, warn */
  b_hnz[blk] += hbuf->n; /* count all nonzero matrix elements */
  /* check if all elements of this matrix line have been stored */
  return;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * do b2i = scfg2idx with buffered cfgs,
 *   minsymcfg is already called within nhm (ToDo: do it to here)
 * called by hamilton_nhv()            v0 may used to get |v1>+=|H|v0>, <v0|H|v0>
 *       and wop_block()   v2.36 m1=0, v0 used to get <v1|Op|v0>
 * if num_ahead_lines=1 it is per line, but since v2.30 its for
 *    num_ahead_lines lines.
 * must be called by each mpi-process and thread to avoid deadlock v2.33
 * careful use MPI within threaded part here!
 * ToDo: call minsymcfg/sq(norm) from here cached/streamed
 *  m1&4 means l1 is not present, dont use cfg2idx for storeh2
 *  diagonal term must be stored first to get sci.norm2
 */

#ifndef DBG_NHM
#define DBG_NHM 0  /* 0=no debug, 1=debug ToDo: DBG_NHM=16 limit output */
#endif
void minsymcfg_blk_lNbrk(thbuf *hbuf); /* ToDo: declare in hilbert.h 2015-04 */
void minsymcfg_blk_bn(thbuf *hbuf); /* ToDo: declare in hilbert.h 2015-04 */
void minsymcfg_blk_bnVS(thbuf *hbuf); /* ToDo: declare in hilbert.h 2015-04 */
void minsymcfg_blk_bnVS_no_ud(thbuf *hbuf); /* ToDo: declare in hilbert.h 2015-04 */
void blk_scfg2idx(tbase *cfg, ULong *idx, int n); /* defined in hilbert.c */


// static  /* used now in xval.c too v2.36 */
void nhm_line(thbuf *hbuf, int m1) { /* cfg-to-idx via MPI for a H-line */ 
  int line_start, /* hold index to first element of a line 0..hbuf->n-1 */
      ii,         /* index from 0 to (hbuf->n-1 < NZXMAX*NUM_AHEAD) */ 
      i2=0;       /* counts through 0..NUM_AHEAD-1 (H-line index) */
  cplx sum=0;
  static scplx r0b[B_NUM][NUM_AHEAD_LINES+1]; /* buffer for <nHv>[] */
  scplx  *r0 = r0b[hbuf->block];
  /* 2013-02-07 index 0..ahead-1 */

  hbuf->nhn[i2]=0; /* 2013-01-17 fix diag2 2x2 */
//{ /* shift from nhm to nhm_line in v2.35 */
#if DBG_NHM /* DEBUG */
{ static int x=0; if (x<64/(mpi_n*pt_n)) { x++;
  if(mpi_i>=4) usleep(10000*8); /* wait for others */
  if(mpi_i< 4) { int i0; usleep(10000*mpi_i); /* sort mpi output */
    printf("\nDBG.nhm_line.in: %2d.%2d hbuf.n=%6d maxahead= %d i= %8lu",
        mpi_i, hbuf->block, hbuf->n, hbuf->num_ahead_lines,hbuf->i);
         // >= nahead
    if (mpi_i*pt_n+hbuf->block<2 || mpi_i*pt_n+hbuf->block>mpi_n*pt_n-2) 
    if(x<10) for (i0=0;i0<hbuf->n;i0++) {  /* NZXMAX*NUM_AHEAD_LINES */
      if (i0==8 && hbuf->n>10) i0= hbuf->n - 2;
      printf("\nDBG.nhm_line.in: %2d.%2d hbuf.i=%6d %s %8.4f %3s",
        mpi_i, hbuf->block, i0, ccfg(hbuf->acj[i0]), D(hbuf->rr[i0]),
        ((hbuf->jj[i0] & XY_FLAG)?"X++":"   "));
    }  usleep(10000*(8-mpi_i)); /* sync-back */
  }
}} /* static int x=0 */
  /* ToDo2015: would be good for debug, to write mincfg(acj) to new array? */
#endif
  if (hbuf->n && !node_n1) { /* zero length rest block, 2016-03 */
    printf("\nERROR.nhm_line: blk %d %d node_n1=%lu, hbuf->n=%d but 0"
           " expected!\n", mpi_i, hbuf->block, node_n1, hbuf->n);
    hbuf->n=0; /* not good, but workaround */ }
  /* ToDo2015: call minsymcfg_dflt with array-args for better SIMD/cache use */
  if (hbuf->n==1 /* one line and one diagonal-element only (faster op_zizj) */
     && hbuf->acj[0] == hbuf->tcfg) { /* ... where tcfg is scfg */
     hbuf->nhn[i2]+=D(hbuf->rr[0]); /* ??? */
  } else /* more than just the diagonal element */
#if 0  /* TEST: 1=old_serial_call_version  0=new_block_call_version(SIMD) */
  /* 2015-09-07 1 = ok, 0 = buggy 2nd run! (ToDo!) */
  {
  struct sbase *sci=&(hbuf->sci); /* symcfg(sci) = input */
  struct sbase *scj=&(hbuf->scj); /* buffer for symcfg(scj=Hij*sci) */
  // scj->o=0;
  for (line_start=0,ii=0;ii<hbuf->n;ii++) {  /* NZXMAX*NUM_AHEAD_LINES */
    /* get minimum symmetric config scj from acj[], speedup by extra out buf??? */
    /* ToDo: will be replaced by one call for whole block hbuf */
    // MIPS 2.75m _inlined 2.75m, inline 2.73m (SH.t s1-tri 46+8 2012-07-03)
    // i586 1.00m _inlined 1.00m, static 0.98m, inline 0.98m
    /* ToDo 2015-04: do loop inside minsymcfg to use cache+SIMD for speedup */
    minsymcfg_dflt(hbuf->acj[ii], scj); /* 98% cpu-time, ToDo: minsymcfg_hbuf */
    /* phase==0: 32+8 true=0..3% sym_k-depending, important for scfg2i() */
    if (line_start==ii) *sci=*scj;  /* store original target config |i> */
    else // remove?
    {
      hbuf->acj[ii]  = scj->scfg;
#if DBG_NHM
      hbuf->asj[ii]  = scj->sign;  /* used only for debug 2015-09 */
      hbuf->an2[ii]  = scj->norm2; /* used only for debug 2015-09 */
#endif      
      hbuf->rr[ii]  *= scj->sign;
      /* 20+20: true=(.2e6/431e6)^2<1%, see n2stats */
      if (sci->norm2!=scj->norm2)
        hbuf->rr[ii] *= sqrt(sci->norm2/(double)scj->norm2);
    }
    /* ToDo: v2.35 must be changed to nhn[AHEAD] or done outside */
    if (sci->scfg==scj->scfg) hbuf->nhn[i2]+=D(hbuf->rr[ii]); /* Hii ~cplx */
    if (hbuf->jj[ii] & XY_FLAG) {
      line_start=ii+1; i2++; hbuf->nhn[i2]=0; /*2013-02*/
    }
  } /* ii=0..hbuf->n */
  } /* else */
#else /* test for better speed 2015-04-15, buggy sign 2015-09-08 2nd run */
  /* else hbuf->n!=1 */ {
    // minsymcfg_blk_lNbrk(hbuf); /* block operation hbuf->n=ca.1400, also slow */
    if(nu==nd && sym.wud) minsymcfg_blk_dflt(hbuf); /* block operation hbuf->n=ca.1400, also slow */
                     else minsymcfg_blk_dflt_no_ud(hbuf); /* 15% faster JJ40 */
               /* link to minsymcfg_blk_bnVS_no_ud(hbuf) dflt=bnVS */
    /* 2015-09-08 ToDo: .scj=0 for diag-term cant be! whats wrong? bug! */
    for (line_start=0,ii=0;ii<hbuf->n;ii++) {  /* NZXMAX*NUM_AHEAD_LINES */
    //  minsymcfg_dflt(hbuf->acj[ii], scj); /* 98% cpu-time, ToDo: minsymcfg_hbuf */
    //  hbuf->acj[ii]=scj->scfg;
    //  hbuf->asj[ii]=scj->sign;
    //  hbuf->an2[ii]=scj->norm2;
      /* phase==0: 32+8 true=0..3% sym_k-depending, important for scfg2i() */
      if (line_start==ii) {}  /* store original target config |i> */
      else // remove?
      {
        hbuf->rr[ii] *= hbuf->asj[ii]; /* symcfg sign */
#if 0
        if (hbuf->acj[ii]==0x0202 && hbuf->rr[ii]!=0){
          int isminsymcfg_lNbrk(tbase ib1, struct sbase *b);
          struct sbase b;
          printf("\nDBG: x0202  ii=%d ismin=% d", ii, isminsymcfg_lNbrk(hbuf->acj[ii],&b));
          printf("\nDBG: x0202  rr=%f %+fi",  Re(hbuf->rr[ii]), Im(hbuf->rr[ii]));
          printf("\nDBG: x0202 sgn=%f %+fi\n",Re(hbuf->asj[ii]),Im(hbuf->asj[ii]));
        }
#endif
        /* 20+20: true=(.2e6/431e6)^2<1%, see n2stats */
        if (hbuf->an2[line_start]!=hbuf->an2[ii])
          hbuf->rr[ii] *= sqrt(hbuf->an2[line_start]/(double)hbuf->an2[ii]);
      }
      /* ToDo: v2.35 must be changed to nhn[AHEAD] or done outside */
      if (hbuf->acj[line_start]==hbuf->acj[ii]) hbuf->nhn[i2]+=D(hbuf->rr[ii]); /* Hii ~cplx */
      if (hbuf->jj[ii] & XY_FLAG) { line_start=ii+1; i2++; hbuf->nhn[i2]=0; }
    } /* ii=0..hbuf->n */
  } /* if hbuf->n==1 else */
#endif
#if DBG_NHM /* 2015-09-07 debug asj + an2 */
{ static int x=0; // for mpi use mpiexec --tag-output ...
  if (x<2000/*(vvv&7)*/ && n1<1000) { int i0; /* debug */ x++;
    if (mpi_i) usleep(10000*mpi_i);
    printf("\nDBG.nhmline.out: blk %d.%d hbuf->n=%d maxahead= %d",
        mpi_i, hbuf->block, hbuf->n, hbuf->num_ahead_lines); // >= nahead
    for (i0=0;i0<hbuf->n;i0++) {
      if (i0==8 && hbuf->n>10) i0= hbuf->n - 2;
      printf("\nDBG.nhmline.out: blk %d.%d i= %2d     %s"
             " sj=%+d n2=%3d rr=% 8.4f %3s",
             mpi_i, hbuf->block, i0, /* i0 not shown for diff ahead=1 vs. n? */
             ccfg(hbuf->acj[i0]),    /* minsymcfg */
             (int)D(hbuf->asj[i0]),  /* sign */
             (int)hbuf->an2[i0],     /* norm2 + rr */ D(hbuf->rr[i0]),
        ((hbuf->jj[i0] & XY_FLAG)?"X++":"   "));
    }   usleep(10000*(mpi_n-mpi_i));
  } // if (x<8 && n1<16)
} // static int x=0 ...
#endif
  /* compression: add two equal elements gives only small boost (one core) */
  /* sq: 32+8  hnz/n1= 27.1663  -8.22686824 162s 164s       */
  /* sq: 32+8  hnz/n1= 27.2569  -8.22686824 159s 164s +0.3% */
  /* sq: 15+15 hnz/n1= 31.582 31.79  (3..51)   compressed 3m20s     */
  /* sq: 15+15 hnz/n1= 31.834 32.04 (15..61) uncompressed 3m15s +1% */
  /* spin1/s1chain.gpl: N=20 20+20 19.30 vs. 19.35 +0.25% */
  /* s1chain+AddSS: N=16 8+8 n1=19 nzx=11.00 vs. 1927.95 Bug! SS*Nw */
  /* sq+AddSS: 4+36 n1=632 nzx=115.54 vs. 159.02 */
//}
  /* cfg-to-idx dont work for storeh2, no l1 */
  if (m1&4) return;

  /* ToDo: like geth if (mpi_n==1) ... */
  if (!l1) for (ii=0;ii<hbuf->n;ii++) hbuf->abj[ii]=mpi_i; /* speed_test 2016-04-07 */
  if (l1) /* 2015-10-08 benchmark, l1 uninitialized */
  for (ii=0;ii<hbuf->n;ii++) {  /* hbuf->n <= NZXMAX*NUM_AHEAD_LINES */
    /* ToDo: bad for performance, but needed at the moment for b2i() */
    if (hbuf->rr[ii]==0) { /* replace by harmless dummy data */
        /* 0..0.03% hit! JJ-square40 k= 2 1 ud=8+32 30000* of 481e3*27.26 */
        hbuf->acj[ii] = getl1(0); /* local_l1[0] ToDo: set to diag-element! */
        /* already set? */
        hbuf->abj[ii] = mpi_i;    /* this node, avoid communication */
    } else {
        hbuf->abj[ii] = scfg2node(hbuf->acj[ii]);
//        printf(" ii=%d mpi_i=%d scfg=%s destnode=%d\n",
//          ii,mpi_i,ccfg(hbuf->acj[ii]),hbuf->abj[ii]);
    }
  } /* for ii=0..n */
  if (mpi_n==1 && l1/* 2015-10-08 */) { /* only local data */
   for (ii=0;ii<hbuf->n;ii++) {  /* hbuf->n <= NZXMAX*NUM_AHEAD_LINES */
    /* scfg2idx() fails, if b.sign==0 for k=pi symmetries */
    hbuf->jj[ii] = (hbuf->jj[ii] & XY_FLAG)  /* save XY_FLAG (ToDo remove) */
        |  scfg2idx(hbuf->acj[ii]);  /* 0<=jj<node_n1 */
    /* ToDo: remove XY_FLAG 1bit/nz add nzx[n1_ahead]=1B/n1 + add limit-checks
     * ToDo2015: use blk_scfg2idx, ToDo: how to detect scfg2idx_err? */
   }
  }
  /* ToDo: better do it after _nhv call when needed??? must be done by MPI! */
  if(mpi_n==1 && hbuf->v0){ /* only needed for <nHv> */
    i2=0; r0[0]=sum=0; /* scplx = cplx = 0 ToDo */
    for (ii=0;ii<hbuf->n;ii++) {
      sum += hbuf->rr[ii] * M_f2d(hbuf->v0[ hbuf->jj[ii] & ~XY_FLAG]);
      /* store diag-block <i2|H|v0> to r0[], 2013-02-07 set r0[next]=0 */
      if (hbuf->jj[ii] & XY_FLAG) { r0[i2]=sum; i2++; r0[i2]=sum=0; }
    }
  }
#ifdef CONFIG_MPI  /* if mpirun fails, try lamclean+ipcs+ipcrm */
  if (mpi_n>1) /* nhm_line.mpi should also work for mpi_n=1 */
  { /* ToDo: send only the relevant parts, not full hbuf */
    /* ToDo: pt_n>1 COMM_WORLD -> pt_n*Worlds? against deadlock MPI_Comm_Create */ 
    /* ToDo: we send hbuf including sci,scj which we dont need (lot of size) */
    int i1, id, block=hbuf->block; /* ToDo: rename block to pt_i */
    /* ToDo: simplify to array[2] idx 0 for send, 1 for recv */
#ifndef HNZ_MPI_BUFSIZE
#define HNZ_MPI_BUFSIZE (NZXMAX*NUM_AHEAD_LINES*150/100) /* 150% */
#endif
    /* ToDo: rewrite to one MPI for all PTs + Alltoallv, minimize MPI-overhead 2017.1 .14 */
    /* we sort different numbers of blockmembers and need B_NUM*maxBlockMembers */
    /* static (=OMP_SHARED) but each thread has its own area */
    /* ToDo: [B_NUM][...+64] to ensure cpus dont have same mem cached */
    static scplx bsend_rr[B_NUM][HNZ_MPI_BUFSIZE+1], /* buffer <nHv> send */
                 brecv_rr[B_NUM][HNZ_MPI_BUFSIZE+1]; /* buffer <nHv> recv */
    static ULong bsend_jj[B_NUM][HNZ_MPI_BUFSIZE],  /* buffer cfg2idx send */
                 brecv_jj[B_NUM][HNZ_MPI_BUFSIZE];
    /* ToDo: union bsend_jj,bsend_cfg of buffers? */
    /* send buffer, rename bsend.cj to bsend_cfg */
    static tbase bsend_cfg[B_NUM][HNZ_MPI_BUFSIZE], /* send remote symcfg */
                 brecv_cfg[B_NUM][HNZ_MPI_BUFSIZE];
    /* mpi_HB is a table of offset+size to data for each node/task */
    /* MPI_MAX_NUM should be greater than a CPU cache line (block.h 8K) */
    static int   bsort_bck[B_NUM][HNZ_MPI_BUFSIZE]; /* sort back index */
    /* end of fat arrays */
    static int   bsend_ofs[B_NUM+1][MPI_MAX_NUM+1]; /* offset v2.36, +[n]=size */
    static int   bsend_of2[B_NUM][MPI_MAX_NUM+1]; /* help counter 2017.1.14 */
    static int   bsend_cnt[B_NUM][MPI_MAX_NUM+1]; /* size per dest_node  */
    static int   brecv_ofs[B_NUM+1][MPI_MAX_NUM+1]; /* offset  2017.1.14 */
    static int   brecv_cnt[B_NUM][MPI_MAX_NUM+1]; /* size per rcv_node 2017.1 */
    void *psrc, *pdst; /* src,dst for better looking code v2.36 */
    int   len1,  len2, rc=0;
    /* ToDo: bsend_cfg[B_NUM*NZXMAX*AHEAD] + startidx[block]=MAX/pt,len[]  */
    /* 1. find destination node for each cfg-data (store it = new) */
    /* 2. sort data + write sort-back-index to bsort_bck (new: for all dest) */
    /* 3. alltoall data transfer (was bad scaling until 2017.1) */
    /* 4. compute data (search cfg-idx + get r*v[idx]) */
    /* 5. alltoall back transfer */
    /* 6. sum to v[i] */
    /* phase 2: count data per destination node, reorder later */
    for (id=0;id<=mpi_n;id++) bsend_cnt[block][id]=0;
    for (ii=0;ii<hbuf->n;ii++) {
      id=    hbuf->abj[ii]; /* destination node */
#if DBG_NHM /* 2016-04-07 speed_test uninitialized */
      if (id != scfg2node(hbuf->acj[ii])) ERR(9998);
      if (id>=mpi_n) ERR(9999);
#endif
      bsend_cnt[block][id]++;  /* count for each dest. node */
    }
    /* offsets from counts (last index (n) is HBLen) for MPI_Alltoallv */
    bsend_of2[block][0]= bsend_ofs[block][0]= 0;
    for (id=1;id<=mpi_n;id++) /* runs one more, to store overall size! */
      bsend_of2[block][id]   = bsend_ofs[block][id] =
      bsend_ofs[block][id-1] + bsend_cnt[block][id-1];
    if (bsend_ofs[block][mpi_n] > HNZ_MPI_BUFSIZE)
      F2("enlarge HNZ_MPI_BUFSIZE/AH=NZX to %d++ (%d%%++)",
         bsend_ofs[block][mpi_n]/NUM_AHEAD_LINES,
         bsend_ofs[block][mpi_n]*100/NUM_AHEAD_LINES/NZXMAX);
    /* ... now we sort in groups for each destination */
    for (ii=0;ii<hbuf->n;ii++) {
      id= hbuf->abj[ii]; /* destination node (array block index) = scfg2node */
      len1 = bsend_of2[block][id];
      bsort_bck[block][ii]     = len1; /* sort-back idx */
      bsend_cfg[block][len1]   = hbuf->acj[ii];   /* sym. spin cfg */
      bsend_of2[block][id]++; /* next free place in list for dest id */
#if DBG_NHM
      if (bsend_of2[block][id] > bsend_ofs[block][id+1]) ERR(9999);
#endif
    } /* ready for send */
#ifdef CKPT_HELP  /* ToDo: better only all 10 or 100 steps or dyn 1/minute */
      /* ToDo: extra Bcast time needed on big clusters ? */
      /* 2015-02-26: code helping to get clean MPI-state for checkpointing */
{     int signal=sig1;/* spins.h: sighand.h: extern volatile int sig1; */
      MPI_Bcast(&signal,1,MPI_INT,0,MPI_COMM_WORLD); sig1=signal; /* Barrier? */
      if (sig1&15) { /* volatile.int.sig1 set by SIGUSR1 in sighand.c for ckpt */
        MPI_Barrier(MPI_COMM_WORLD); /* wait for all tasks */
        if (!mpi_i) printf("\nHLP: start 30s checkpoint window ...\n");
        /* ToDo: auto trigger ckpt-script by writing to a fifo here? */
        sleep(30); /* safe checkpoint window without MPI-messages */
        if (!mpi_i) printf(  "HLP: ... end of safe checkpoint window\n");
        /* restore hostnames, PIDs, clock tdelay+=32+ckpt etc. here? */
        sleep(2); sig1=0; } }/* wait for all restored tasks */
#endif /* CKPT_HELP */
    /* ensure balanced number of MPI calls outside, using hbuf->n=0 */
    /* MPI communication part for SH starts here, MPI_Alltoallv */
    {
#ifdef _OPENMP
      static volatile int omp_token[2*B_NUM]={1,0};  // correct ini? 1 0 0 0 ... 0
#endif
//#pragma omp barrier    // ??? SH hangs
{   // but better let only masterthread do send/recv for all tasks
#ifndef _OPENMP
      pthread_mutex_lock(&token_ring[hbuf->block]);
#else
      while (!omp_token[hbuf->block]);
#endif
/* 2017.1 not symmetric! splitted elements! disadvantage of ~compression */
#if DBG_NHM  /* nhm_line */
      if (mpi_n<4 && (vvv&4) && mpi_syncs<16) { usleep((mpi_i+1)*100000); 
        printf("\nDBG: blk=%2d.%d send-size:",mpi_i,block);
        for (id=0;id<mpi_n;id++) printf(" %6d", bsend_cnt[block][id]);
        printf("  sum= %6d", bsend_ofs[block][mpi_n]); }
#endif
      /* thread safe? idx i to task i ids mpi_i, gather? */
      rc=MPI_Alltoall(bsend_cnt[block],1,MPI_INT,
                      brecv_cnt[block],1,MPI_INT,MPI_COMM_WORLD);
      if (rc) perror("MPI_Alltoall");
      brecv_ofs[block][0]=0;    /* write idx [0..mpi_n] (one more!) */
      for (id=1;id<=mpi_n;id++)  brecv_ofs[block][id] =
        brecv_ofs[block][id-1] + brecv_cnt[block][id-1];
      len1= bsend_ofs[block][mpi_n];
      len2= brecv_ofs[block][mpi_n];
      if (len2> HNZ_MPI_BUFSIZE) F1("enlarge HNZ_MPI_BUFSIZE to %d++", len2);
#if DBG_NHM  /* 2017.1 not symmetric! splitted elements! ToDo? */
  if (mpi_n<4 && (vvv&4) && mpi_syncs<16) { usleep((mpi_i+1)*100000); 
        printf("\nDBG: blk=%2d.%d recv-size:",mpi_i,block);
        for (id=0;id<mpi_n;id++) printf(" %6d", brecv_cnt[block][id]);
        printf("  sum= %6d", brecv_ofs[block][mpi_n]); }
#endif
      mpi_syncs++; mpi_traffic += len1; /* shared! */
/*  if (len1==0) mpi_z_pkts++; 2015-01 */ /* count wasted packets JS2011-10 */
      // eats time propto mpi_n = bad
      // better per pair? for (ii=0;ii<mpi_n;ii++)
      { int i3;for(i3=0;i3<7 && len1*sizeof(tbase)/mpi_n>1024u<<(2*i3);i3++);
        /* i3 is num leading zeros */ mpi_stats_npkts[i3]++; } /* mpi statistics */
      /* since alltoall v2.53 mpi_maxlen is not per destnode, but per A2a */
      if (len1>(int)mpi_maxlen) mpi_maxlen=len1; /* max mpi buffer length */
#if DBG_NHM
      if((vvv&4) && mpi_n<8 && mpi_syncs<16)  // max= 8*16 outs
        printf("\nDBG.MPI_a %3d.%d len= %6d %6d", mpi_i, block, len1, len2);
#endif
      for (ii=0;ii<mpi_n;ii++) bsend_cnt[block][ii] *= (int)sizeof(tbase);
      for (ii=0;ii<mpi_n;ii++) bsend_ofs[block][ii] *= (int)sizeof(tbase);
      for (ii=0;ii<mpi_n;ii++) brecv_cnt[block][ii] *= (int)sizeof(tbase);
      for (ii=0;ii<mpi_n;ii++) brecv_ofs[block][ii] *= (int)sizeof(tbase);
      rc=My_MPI_Alltoallv(\
        bsend_cfg[block], bsend_cnt[block], bsend_ofs[block], MPI_BYTE,\
        brecv_cfg[block], brecv_cnt[block], brecv_ofs[block], MPI_BYTE,\
        MPI_COMM_WORLD);
      if (rc) perror("MPI_Alltoallv");
      /* may hang if memory is to low! ompi-1.4 2M-- free 2015-04 */
      for (ii=0;ii<mpi_n;ii++) bsend_cnt[block][ii] /= (int)sizeof(tbase);
      for (ii=0;ii<mpi_n;ii++) bsend_ofs[block][ii] /= (int)sizeof(tbase);
      for (ii=0;ii<mpi_n;ii++) brecv_cnt[block][ii] /= (int)sizeof(tbase);
      for (ii=0;ii<mpi_n;ii++) brecv_ofs[block][ii] /= (int)sizeof(tbase);
#ifndef _OPENMP
      pthread_mutex_unlock(&token_ring[hbuf->block+1]);
#else
      omp_token[hbuf->block]=0;
#pragma omp critical   // will replace pthread_locking later???
      omp_token[(hbuf->block+1)%(2*pt_n)]=1;
#pragma omp flush (omp_token)
#endif
}
      /* we need to send .abj .acj _OR_ .rr */
      /* ToDo: b2i(cj,&blk) should return idx + abj to reduce traffic??? */
      len2= brecv_ofs[block][mpi_n];
      // if (l1/* 2015-10-08 */) for (ii=0;ii<len2;ii++)
      //   bsend_jj[block][ii] = scfg2idx(brecv_cfg[block][ii]);
      if (l1) blk_scfg2idx(brecv_cfg[block],bsend_jj[block],len2);
      /* ToDo2015: use blk_scfg2idx */
#if DBG_NHM
if((vvv&4) && mpi_n<8 && mpi_syncs<16) { usleep(10000*mpi_i); /* order */
   for (ii=0;ii<len2 && ii<8;ii++)  /* count blk_hnz (NZX*AHEAD) */
     printf("\nDBG.nhm_l.scfg2idx: blk %2d.%d ii=%4d jj=%5lu from cfg= %s",
          mpi_i,block,ii, bsend_jj[block][ii], ccfg(brecv_cfg[block][ii])); }
#endif
      if (hbuf->v0) for (ii=0;ii<len2;ii++) /* len2 <= HNZ_MPI_BUFSIZE */
        bsend_rr[block][ii]  = hbuf->v0[ bsend_jj[block][ii] ];
#ifndef _OPENMP
      /* ToDo: mutex may be slow if called mpi_n=1000 times!?, do pure mpi? */
      pthread_mutex_lock(&token_ring[hbuf->block+pt_n]);
#else
      while (!omp_token[hbuf->block+pt_n]);
#endif
#if DBG_NHM
      if((vvv&4) && mpi_n<8 && mpi_syncs<16)
        printf("\nDBG.MPI_b %3d.%d len= %6d %6d", mpi_i, block, len2, len1);
#endif
      /* ToDo: let bsend_jj[block] on remote node until geth */
      if (!hbuf->v0)  /* send back ordered data to remote node */
           { len1=Lsize;         pdst=brecv_jj[block]; psrc=bsend_jj[block]; }
      else { len1=sizeof(scplx); pdst=brecv_rr[block]; psrc=bsend_rr[block]; }
      for (ii=0;ii<mpi_n;ii++) bsend_cnt[block][ii] *= len1;
      for (ii=0;ii<mpi_n;ii++) bsend_ofs[block][ii] *= len1;
      for (ii=0;ii<mpi_n;ii++) brecv_cnt[block][ii] *= len1;
      for (ii=0;ii<mpi_n;ii++) brecv_ofs[block][ii] *= len1;
      /* send results back to asking threads (block=thread_idx) */
      rc=My_MPI_Alltoallv(psrc,brecv_cnt[block],brecv_ofs[block],MPI_BYTE,\
                          pdst,bsend_cnt[block],bsend_ofs[block],MPI_BYTE,\
                          MPI_COMM_WORLD);
      if (rc) perror("MPI_Alltoallv");
      /* may hang if memory is to low! ompi-1.4 2M-- free 2015-04 */
//      for (ii=0;ii<mpi_n;ii++) bsend_cnt[block][ii] /= len1;
//      for (ii=0;ii<mpi_n;ii++) bsend_ofs[block][ii] /= len1;
//      for (ii=0;ii<mpi_n;ii++) brecv_cnt[block][ii] /= len1;
//      for (ii=0;ii<mpi_n;ii++) brecv_ofs[block][ii] /= len1;
#ifndef _OPENMP  // ToDo: ist es nicht besser, 2 Ringe zu fahren? nee
      pthread_mutex_unlock(&token_ring[(hbuf->block+pt_n+1)%(2*pt_n)]);
#else
      omp_token[hbuf->block+pt_n]=0;
#pragma omp critical   // will replace pthread_locking later???
      omp_token[(hbuf->block+pt_n+1)%(2*pt_n)]=1;
#pragma omp flush (omp_token)
#endif
//    pthread_mutex_unlock(&token_ring[(hbuf->block+1)%(pt_n)]);
    /* sorting back results from remote nodes */
    } /* old Sendrecv-loop id=0..mpi_n-1 */
//  --- begin --- was out mpi loop until 2015-01-31 
    if (!hbuf->v0)  /* outside the mpi-loop? */
      for (ii=0;ii<hbuf->n;ii++) { /* idx_cfg | XY_FLAG */
        hbuf->jj[ii] = (hbuf->jj[ii] & XY_FLAG)
                     | brecv_jj[block][ bsort_bck[block][ii] ];
#if DBG_NHM
      if ((ii<8 || ii>hbuf->n-8) && mpi_n<8 && (vvv&4))
        printf("\nDBG.nhm.bsort: mpi=%2d ii= %5d recv_jj= %5d bsort= %5d",
               mpi_i, ii, (int)brecv_jj[block][ii], bsort_bck[block][ii]);
#endif
      } /* for ii if !hbuf->v0 */
    if (hbuf->v0) /* decompress r2 and add to r0[]=<i2|H|v0> */
    for (r0[0]=0,i1=ii=0;ii<hbuf->n;ii++) { /* <i2|H|v0> ??? */
      r0[ i1 ] += hbuf->rr[ii] * brecv_rr[block][ bsort_bck[block][ii] ];
      if (hbuf->jj[ii] & XY_FLAG) { i1++; r0[i1]=0; } /* next HLine <= AHEAD_LINES */
    }
//  --- end --- was out mpi loop until 2015-01-31 
  } /* mpi_n>1 */
#endif  /* CONFIG_MPI */
  /* ToDo: better write <nHi> <nHv> separately?
   *    using .rr for two different things confuse the reader
   *  2013-02-07 use nhv[ii] instead of hbuf->rr[ii]
   */
  if (hbuf->v0) /* store <i2|H|v0>, nhv:cplx *r0:scplx */
    for (ii=0;ii<hbuf->num_ahead_lines;ii++) hbuf->nhv[ii]= r0[ii]; /* a2 */
#if DBG_NHM /* 2013-02-07 debug */
{ static int x=0; // for mpi use mpiexec --tag-output ...
  if (x<2000/*(vvv&7)*/ && mpi_n<8 && (vvv&4)) { /* debug */ x++;
    int ii, i0;  usleep(10000*mpi_i);
    printf("\nDBG.nhmline.out: blk %d %d hbuf->n=%d maxahead= %d",
        mpi_i, hbuf->block, hbuf->n, hbuf->num_ahead_lines); // >= nahead
    for (i0=0,ii=hbuf->jj[i0];i0<hbuf->n && i0<16;i0++) {
      printf("\nDBG.nhm_l.out: blk %d.%d i= %2d ii %d %s"
             " bj=%d jj= %5d rr=% 8.4f %3s",
             mpi_i, hbuf->block, i0, ii, /* i0 not shown for diff ahead=1 vs. n? */
             ccfg(hbuf->acj[i0]),     /* minsymcfg */
                  hbuf->abj[i0],      /* bj= 0 .. mpi_n */ 
        (int)hbuf->jj[i0] & ~XY_FLAG, /* jj= 0..node_n1-1 */ 
           D(hbuf->rr[i0]),
        ((hbuf->jj[i0] & XY_FLAG)?"X++":"   "));
      if (hbuf->jj[i0] & XY_FLAG) ii++; /* next line */
    } // sleep(1);
    if (hbuf->v0) /* show <i2|H|v0> */
      for (i0=0,ii=hbuf->jj[i0];i0<hbuf->num_ahead_lines;i0++)
        if (hbuf->nhn[i0] || D(hbuf->nhv[i0]))
        printf("\nDBG nhm_line blk=%d %d ii=%2d nhn,nhv[i]= % 6.4f % 8.6f",
             mpi_i, hbuf->block, ii+i0, hbuf->nhn[i0], D(hbuf->nhv[i0]));
  } // if (x<8 && n1<16)
} // static int x=0 ...
#endif /* DBG_NHM */
} /* nhm_line() */

/* ------------------------------------------------
 * r*(j|j><i|i), Hij=r*|j><i|, i=n, j=m, hbuf->sci,scj ??? 
 *   v2.28: do cfg-to-idx later in nhm_line
 *   v2.35: do minsymcfg later in nhm_line (better caching, +FPGA)
 *  hbuf->i is the cfg index on the node (0..b_len-1) # less bits
 *  cfgj is the cfg modified by operator H, r*cfgj=H(i,j).sci
 * ToDo: store para_index instead of r
 * ------------------------------------------------
 */
static inline void nhm(tbase cfgj, cplx r, thbuf *hbuf) {
  if (r==0) return;    /* 32+8: true=0 (only if parameter==0), speedup? */
#ifdef CONFIG_ANSATZ
  if (!ansatz_test(cfgj)) return;  /* ToDo: buffered call? */
#endif
  hbuf->rr[hbuf->n]=r;    /* fill H-buffer of this thread */
  hbuf->acj[hbuf->n]=cfgj; /* by operator H modified config, H(cfg_i) */
  hbuf->jj[hbuf->n]=0; /* v2.30 reset to carry XY_FLAG (set in _nhv) */
  hbuf->n++;  if (hbuf->n>=NZXMAX*NUM_AHEAD_LINES) ERR(9200);
  return;
} 

/* ----------------------------------------------------------------
 * m1=mode: 0=<n|H*v0> (return nhn too)
 *          1=<n|H*n>   obsolete! used for 2x2-algo
 *            until v2.25 for uududd==uuddud <nHn>=0, but its != 0
 *            result correct? differences? 3+3 diverged (n1=3)
 *  uuuddd/q6, uududd/q12, ududud/q2
 *  [1/2, q(1/2), 0] [q(1/2), 1/2, q(3/2)] [0, q(3/2), -3/2] trace=-1/2
 *  nhn: 1/2  1/2 -3/2  8It e0=-2.118033988749896
 * m1=mode: 2=storeh,
 *          4:buf_..=H|n> without_sym+idx ... (called by storeh2)
 *          8:buf_..=H|n> full
 * ToDo: always return nhn (by pointer arg?) do m1=0,1 in one step! m1==1 obsolete
 *           i + nhn in buffer[]?
 * sample: 6sites
 *   |i>=|sym_ududud>=(ududud+dududu)/q2                    - different norm2
 *   |i>=|sym_ududud>=(ududud+dududu+ududud+dududu+...)/q2/6  # not q12!
 * norm2: 1*2 
 * ToDo: MPI+FPGA:
 *       if i%num_ahead_lines return from ahead_hbuf to hbuf
 *       else  fill aheadhbuf=hbuf[NZXMAX*num_ahead] + call scfg + b2i, (wop analog)
 *       return 1st line from aheadhbuff to hbuf
 *       on overflow reduce num_ahead_lines or go to num_ahead_lines before overflow
 *       aheadhbuf=hbuf+.startel pointing to el line i
 * ToDo: return <nHn> so that nhv(m1=0);nhv(m1=1) can be replaced by nhv(m1=0) 
 * ToDo: put i into hbuf (reduce arguments)
 * ToDo: split nhv to hbuf=H*|n> (H.elements) and v0*hbuf (v.elements)
 * input: (nahead configs)
 *    l1[0..(node_n1-1)]
 *    hbuf->blk= 0..(pt_n-1) // add mpi*pt_n if needed (b_ofs+b_len)
 *    i  = line index (better, use range to reduce calls)
           0<=i<b_ofs[blk+1]-b_ofs[blk]-1 (0..thread_len)
 *             move to hbuf->i  = 0..(node_n1-1)
 *    hbuf->v0 = NULL(vector.c,spins.c) or v0(xval.c_rh,diag.c) (can be moved to hbuf?)
 *    hbuf = work buffer (something set?)
 *    m1 = mode (see above)
 * modified: hbuf.{}
 * calling: nhm+puth_buf to store stream of elements
 * output: (approx nahead*hnz_per_line configs as compressed sparse matrix)
 *    hbuf.{(rx,nhn if v0 set),n,{acj,jj,rr}[0..n-1]} if v0==NULL
 *     where rr= H[ii,jj] (ii implicite given by counter + jj-flag)
 *    hbuf.{nhv[ahead],nhn[ahead] if v0 set)}
 *     where nhv[ii]= sum_jj H[ii,jj] |jj>
 * return: nhv (if v given) ToDo: return in hbuf!?
 * we need to work on range of i=i1..i2 to have blocking and less MPI-overhead
 * must be called by each mpi-process and thread to avoid deadlock (also n1=0)
 * called by:
 *   diag.c:diag2_thread                  < n|H|v0> <n|H|n> m1=0 v1=0?
 *   xval.c:wh                            <v0|H|v0>         m1=0 v1=0?
 *   xval.c:whq                           |H*v0-rh*v0|      m1=0 v1=0?
 *   spins.c:hamilton_geth_block=hamilton i<len[blk] v1+=H*v0 <v0Hv0> m1=0
 *   vector.c:randvec                     i<len[blk] <nHn> v0=0 m1=0
 *   spins.c:storeh_block=storeh          i<len[blk]       v0=0 m1=2
 *   spins.c:storeh2                      i<n1             v0=0 m1=4 (no valid l1)
 *   spins.c:fulldiag         hbuf[nHm..] i<len[blk]       v0=0 m1=8
 * nahead: num H lines to store, nahead=1..NUMAHEAD (1=old version)
 *         0 means no but call MPI block for balance v2.35
 *         the idea of nahead>1 is to reduce hamilton_nhv calls
 * iHi was last element of line Hi, now its first element v2.35
 * ToDo 2015-02: allways fill 2AHEAD buffer and return nahead results from buffer
 * (old) single config version calling ahead version + buffering
 * ---------------------------------------------------------------- */
cplx hamilton_nhv(ULong i, /* i=0..node_n1 */
    ULong nahead, /* blocksize for i block-operations, 0..AHEAD */
    thbuf *hbuf, int m1) { /* input/output+buffer, m1=mode (ToDo obsolete) */
  tbase ib1;     /* target and workconfig */
  int ip1,ip2,ii,i1,ia,ib; /* worksites, workindizes */
  ULong iahead;
  int line_start; // pointer to index of hbuf.xx[index ] start of ne line
  cplx r1;   /* ToDo: old: sum = nhv = <i|H|v0> moved to [2*i+1].rr ??? */
  double rp,  /* temporary exchange parameter for S1SYM || CM(1,1,0)
                 ToDo2015: expand for tU?  */
         sun; /* sun=diagonal sum, onlyZ? */
#if CM(0,1,1)
  int j, is;
#endif
//  hbuf->sci.o=0;                        /* minsymcfg option */

  //  make compute blocks of field length to allow vectorization + speedup
  // ToDo: better overview: fill_ahead_block of cfg by dummy0cfg  2015-02
  // ToDo: better overview: compute_ahead_block_nHm of cfg
  // ToDo: better overview: compute_ahead_block_nHminsymcfg of nHm
  // ToDo: better overview: compute_ahead_block_nhv of nHminsymcfg
{
 hbuf->n=0; /* number of stored elements in ahead buffer */
 line_start=0;
#if DBG_NHM  
  /* HBLrVS = (((NZXMAX*NUM_AHEAD_LINES+VS-1)/VS)*VS) rounded up */
  { int k;for (k=0;k<HBLrVS;k++){/* better debug output *hbuf 2015-09 */
      hbuf->asj[k]=0;
      hbuf->an2[k]=0;
      hbuf->jj[k]=0;
      hbuf->acj[k]=TBASE0;} }
#endif
 // printf("L%d mpi=%d n1=%lu nahead=%lu\n",__LINE__,mpi_i,node_n1,nahead);
 /* 2015-10-08 add hbuf->cfg,nzx for simpler code + allow early benchmark */
 if (l1) { /* else cfg filled by speed_tests, Todo: set outside?! */
   for (iahead=0;iahead<nahead;iahead++)
     if (i+iahead+b_ofs[hbuf->block]>=node_n1) {/* global ulong node_n1,b_ofs */
#if 0   
        /* happens on last block on all tasks!
         * ToDo: variable tmp_nahead? or break? or loop + restloop or fill 0
         *   this works but is badly concept, bad perfromance + triggers bugs
         */
        static int err1255=1; /* 2016-03-22 stop out of array access */
        if (err1255) printf("\nDBG-INFO: iahead=%d > node_n1, task=%d\n",
           (int)(i+iahead+b_ofs[hbuf->block]), mpi_i);
        err1255=0;
#endif
       break; } /* slow innerloop proof + break */
     else { hbuf->cfg[iahead]=getl1(i+iahead+b_ofs[hbuf->block]); }
 }
 /* get H lines ahead to group nonlocal memory access */
 for (iahead=0;iahead<nahead;iahead++) {

  /* 0 <= hbuf->block < pt_n, 0 < i <= b_ofs[blk+1]-b_ofs[blk] */
  /* dont do calculations on empty MPI blocks (for balance) */
  /* see b_ini: node_ofs=mpi_ofs[mpi_i]; b_ofs[i]=0..node_n1 */ 
  // if (l1 /* 2015-10-08*/ && ... ) /* 2016-03 bad idea */
  if (i+iahead>=b_ofs[hbuf->block+1]-b_ofs[hbuf->block]) break; /* ahead-loop */
//  if (node_n1==0) {
//    printf("\nERR mpi=%d blk=%d node_n1=0 b_ofs[1-0]=%lu l1=%p, i+ahead=%lu\n"
//   mpi_i, hbuf->block, b_ofs[1]-b_ofs[0], l1, i+iahead); ERR(9999); }                  

  hbuf->sci.scfg=ib1=hbuf->cfg[iahead]; /* sci.scfg ?? */
  hbuf->i=i+iahead; /* ToDo: remove hbuf->i!? but can be used for progress */
  sun=0;
  hbuf->acj[line_start]=ib1; /* always start with cfg_i (better code) */
  hbuf->jj[line_start]=i+iahead; /* ToDo: replace by num_of_line_elements */
  hbuf->abj[line_start]=mpi_i /* *pt_n+hbuf->block */; /* (abj,jj) == idx(cj) */
  hbuf->rr[line_start]=0; /* set at the end of this loop */
  hbuf->n++; /* minimum of one entry per line == diag element */

// #warning "Debugging code"
// if (i>=b_len[hbuf->block%pt_n+pt_n*mpi_i]) ERR(9999); /* */

#ifdef AddSS	/* only ok for JJ-model !???, AddSS must be a float number */	
    /* add AddSS(SS-ZZ)=sum_(i,j).5(SipSjm+SimSjp) to Ham, max. nu*nd terms? */
    for(ia=0;ia<nn;ia++){    ip1=geti(ib1,ia);
     if(ip1==1 || ip1==2) sun+=.5*AddSS;  /* diagonal part Z^2 */
     for(ib=ia+1;ib<nn;ib++){ ip2=geti(ib1,ib);
       if((ip1^ip2)==3) nhm(Flip(ib1,ia,ib),AddSS,hbuf);
    } }
#endif
/* ??? should we use symmetry at this point for speed up ??? i-j => pi-pj 
 * with smallest fitting  resulting configuration? */
  for(i1=0;i1<nw;i1++){
    ia=iw[i1][0];ip1=geti(ib1,ia);
    ib=iw[i1][1];ip2=geti(ib1,ib);
    if (ia==ib) continue; /* 2016-01-21 silently ignore bad self-bonds */
    rp=para[ww.wj[i1]];  /* exchange parameter J */
#if defined(CONFIG_S1SYM) && CM(1,0,0)  /* N=8 s=5/2  7*faster */
    /* 2016-01 recheck speedup for cubocta.def s=7/2  9*faster nzx? */
    /*  reducing nzx (nonzero matrix terms by compacting LM-equals)! */
    /* for composed spins take only lowest up(down) (ex: s=5,m=1: uuuudd)
     * and correct by a factor num_up(down) to speed up of (2s)^2/4
     *  ToDo: emulate by S(+/-,idx_composed)
     *  S^+|j,m>=((j-m)*(j+m+1))^(1/2)|j,m+1>
     *  S^-|j,m>=((j+m)*(j-m+1))^(1/2)|j,m-1>
     * Why C does not support break(2) to leave 2 loops?
     * subspins of composed spins must be ordered: uuu...ddd
     * ToDo: test mixed spin-lengths
     *  bonds must be sorted to let this work ??? lowest i-j first
     */
#if  CM(0,1,1)    /* bugfix 2016-01-04 not final, partly fixed? */
    if ((ip1==1 && ip2==2) || (ip1==2 && ip2==1))
#endif 
// ToDo: if pre[ia] same LM and same bit, skip (rp=0)
// ToDo:  count rp factor for every next[ia] has same bit[s]
//   if done for tU do if (!rp) continue; = faster skipping
// ToDo: check nzx is minimum for best code, add tU tJ 
// ToDo: simplify,  code is looking unnecessary complex
    if (sym.num_lm!=nn && rp && ip1!=ip2) { /* ud - du flip? ToDo: uu also! */
     int nm[4],i,node;
     if (sym.ls[ia]>1 && sym.ls[ib]>1 && sym.li[ia]==sym.li[ib]) rp=0;
     for (node=0;node<2 && rp;node++){  ii=iw[i1][node]; /* ii={ia,ib} */
       if (sym.ls[ii]>1) {  /* composed spin detected */
        if (sym.li[ii]!=0 && sym.li[ii]!=sym.ls[ii]-1) rp=0; /* JJ only */
        else { /* member 0 or s-1 */
          nm[0]=nm[1]=nm[2]=nm[3]=0; /* later add 3. - .3 flip */
          for (i=0;i<nn;i++) if (sym.lm[i]==sym.lm[ii]) /* ToDo: next! */
            for (;i<nn;i=sym.ln[i]) nm[geti(ib1,i)]++;
          if (sym.li[ii]==0)            rp*=nm[2];  /* S^+|....d> */
          if (sym.li[ii]==sym.ls[ii]-1) rp*=nm[1];  /* S^-|u....> */
        }
       }
     }
    }
#endif /* defined(CONFIG_S1SYM) */
    ii=ip1^ip2;
#if   CM(1,1,0)
    /* SiSj - ninj/4 + ham1*ninj/4 */
    /* Ham1:  energy shift -3/4 and 1/4 (Ham1=1) from/to -1 and 0 (Ham1=0) */ 
    /* [H += ninj/4] (compensation of -ninj/4), 2spinEV: -3/4, 1/4 */
# if   IFz   /* !XY-model (faster than h_zpara=0) */
    /* right place to add bond dependend anisotropy to h_zpara, ToDo */
    /* or split .wj to Jxy=.wj and Jz=.wj+.wjz! */
#  if   Ham1		
    CM({},if(ip2 && ip1),{})  /* ninj */
      sun+=(.25*h_zpara+Add34)*(rp+para[ww.wjz[i1]]);
#  endif
# endif
    if(ii==3){  /* term for different spin directions: xo, ox */
# if     IFz
      /* SiSj-ninj/4  2spinEV: -1, 0 */
      sun-=.5*h_zpara*(rp+para[ww.wjz[i1]]);
# endif
      if(rp!=0){ /* nondiagonal terms (sym: partly diagonal) */ 
        r1=h_xypara*.5*rp;
        /* signum of imag depends on whether ud or du */
        IFww( r1*=C(ww.co[i1],ww.si[i1]*((ip2==1)?-1:1)); )
        nhm(Flip(ib1,ia,ib),r1,hbuf);
      }
    }
# ifdef	DEBUG	
    if(err_fatal){ E0("nhv SxSx");printf(" nw=%d i=%d j=%d",i1,ia,ib); }
# endif
#endif  /* CM(1,1,0) */
#if   CM(0,1,1)		/* t_{ij} cp_i c_j     t=t_{ij} t^*=t_{ji} */
    if( CM( ii , ii>0 && ii<3 , ii ) ) {
      /* valid: .u  ,only tU: ud u3 ToDo: faster with tU only ii>0 ??? */
      is=parity(ib1 & p_mask[ia][ib]); /* fermionic signum without sites */
      for(j=0;j<2;j++)if(ii & (j+1)){ /* up or down */
        r1=is*para[ww.wt[i1]];        /* signum*t_{ij} */
        if( (j==0) ? ip1&2 : ip2&1 ) r1=-r1; /* fermionic sign correction */
        IFww( r1*=C(ww.co[i1],ww.si[i1]*((ip2 & (j+1))?-1:1)); )
        nhm(tbase_xor(tbase_xor(ib1,ia,j),ib,j),r1,hbuf);
      }}
#endif
  } /* for (i1=0;i1<nw;i1++) all dual-bonds */
#if CM(0,0,1)    /* tU */
#if Nv             /* four-fermion-term, inter site repulsion+exchange */
    for(i1=0;i1<Nw;i1++){          /* wv: uu,ud,du,dd,u3,d3,3u,3d,33 */
      ia=iw[i1][0];ip1=geti(ib1,ia);
      ib=iw[i1][1];ip2=geti(ib1,ib);
#if Nv & 4    /* code activated if Nv is 4 for performance reasons */
      /* hole-hole repulsion, hole = empty site = (1-niu*nid) */
      /* this Hamiltonian is used by Fatiha Ouchni, Jul2004 */
      if (ip1==0 && ip2==0) sun+=para[ww.wh[i1]];
#endif /* if Nv & 4 */
#if Ham3
#error "Code not verified. Is the sign correct?"
      if(ip1*ip2==0)if((ip1^ip2)==3) /* .3,3. -JH Oles (jxy==jz only?) */
        nhm(Flip(ib1,ia,ib),-para[ww.wj[i1]],hbuf);
        /* ensures symmetry */
#endif  /* if Ham3 */  
      /* ToDo: more general would be a 4x4 matrix ww.repulsion with
       *       sun+=para[ww.wv[i1]]*ww.repulsion[ip1][ip2]; (less if's)
       * transform ip1,ip2 to symmetric quantum numbers
       *           nij_sum=ni+nj-2=-2..2, nij_diff=ni-nj  z=nu-nd=-2..2 */
      if((ip1>0)&&(ip2>0)){
        if(ip1&1)if(ip2&1)sun+=para[ww.wv[i1]]+para[ww.wj[i1]];/* uu  U+JH */
        if(ip1&2)if(ip2&2)sun+=para[ww.wv[i1]]+para[ww.wj[i1]];/* dd       */
        if(ip1&1)if(ip2&2)sun+=para[ww.wv[i1]];		/* ud   U    */
        if(ip1&2)if(ip2&1)sun+=para[ww.wv[i1]];		/* du        */
        if((ip1^ip2)==3)				/* ud,du  JH */
         nhm(Flip(ib1,ia,ib),para[ww.wj[i1]],hbuf);}
    }
#endif /* if Nv */
#endif /* tU */
#if CM(0,1,1)    /* tU, 2015-09-23 onsite-term added to tJ */
    for(i1=0;i1<nn;i1++){ ip1=geti(ib1,i1);
//#if Ne   /* 2014: Ne was not defined v2.16-2.47 2003-2014 +earlier */
      sun += ((ip1&1)+((ip1&2)>>1)) * para[ww.we[i1]];  /* E*(nu+nd) */
//#endif /* if Ne */
#if Ham2  /* 0.25(2nu-1)(2nd-1): ..=+0.25 u.=-0.25 .d=-0.25 ud=+0.25 */
      sun += ((ip1%3)?-.25:.25)*para[ww.wu[i1]];  /* U(nu-.5)(nd-.5) */
#elif CM(0,0,1) /* tU + ~Ham2 only */
      if (ip1==3) sun += para[ww.wu[i1]];                 /* U nu nd */
#endif /* if Ham2 (+tU) */
    }
#endif /* tU */
#ifdef HFIELD
    for(i1=0;i1<nn;i1++){ ip1=geti(ib1,i1);
      sun-=((ip1&1)-(ip1>>1&1))*para[ww.wf[i1]]/2;
    }
#endif

#define S_DEBUG(a) a
#ifdef SSANISO
#ifndef CONFIG_S1SYM
# error "defining SSANISO without CONFIG_S1SYM makes no sence"
#endif
   { /* this code was introduced by Reimar Schmidt, Jan 2004 */
     int k,i2; double oZi; static unsigned int dbg=0;
     /* oZi - eigenvalue of S^z_i of a composed spin
      *  Di - ww.wd index to float parameter (defined in daten.def)
      *  code revised by Joerg v2.20 */
     /* num_lm is N for a s=1 chain of length N build by NN=2*N spins s=1/2 */
     for(k=0;k<sym.num_lm;k++){  /* k = index of a composed spin-s object */
       for(oZi=0.0,i2=NN,i1=0;i1<nn;i1++)  /* sum S^z(i) over a composed spin */
         if(sym.lm[i1]==k) { /* check if spin i is member of composed spin */
           oZi+=(3-2*geti(ib1,i1))/2.0;
           if (i1<i2) i2=i1; /* get index of 1st member of composed spin */
         }
       sun+=oZi*oZi*para[ww.wd[i2]]; /* add to the diagonal element */
       S_DEBUG( if (dbg++<10) printf("\n dbgD k=%2d oZi=%4.1f d[i]=%.2f %s", k, oZi, para[ww.wd[i2]],ccfg(ib1));)
     }
   }
#endif
#if CM(1,0,0)
   if (nq) {  /* four-site ring term, see op_rot4() too */
     int i2; tbase ib2;
   /* sample:
    *             1      J12=J23=J34=J41, J13, J24
    *           / | \    Q1234 = four-site term (cyclic exchange)
    *          /  |  \   ferri state: S_total=0, S13=S24=1 EF=-2J12+J13/2+Q1234
    *         4---+---2      J24 term? (Nedko-Solution)
    *          \  |  /
    *           \ | /
    *             3
    */
     for (i1=0;i1<nq;i1++) {
       ib2 = ib1;
       for (i2=0; i2<4; i2++) {
         ip1 = geti(ib1, iq[i1][i2]);             /* take site i */
         ib2 = puti(ib2, iq[i1][(i2+1)%4], ip1);  /* put to site i+1 */
       }
       /* rotation */
       nhm(ib2,para[ww.wq[i1]],hbuf);

       ib2 = ib1;
       for (i2=0; i2<4; i2++) {
         ip1 = geti(ib1, iq[i1][i2]);             /* take site i */
         ib2 = puti(ib2, iq[i1][(i2+3)%4], ip1);  /* put to site i-1 */
       }
       /* inverse rotation */
       nhm(ib2,para[ww.wq[i1]],hbuf);
     }
   }
#endif

#ifdef EDIAG
    sun+=EDIAG;  /* add 1000.0 if energy is -1000.00365 for better results */
#endif
    hbuf->nzx[iahead]=hbuf->n-line_start; /* 2015-10-08 for vector use, ToDo */
    /* 2015-10-08 printf("DBGnhm: %s nzx=%d\n",ccfg(ib1), hbuf->nzx[iahead]); */
    /* diag element is not always the last because sym(H*v) could be diag */
    /* nhm(ib1,sun,hbuf);  -old- store/add diagonal value <nHn> */
    hbuf->rr[line_start]=sun; /* set diagonal value <nHn> v2.35 */

    line_start=hbuf->n; /* store EOL position for next line */
    /* compression is bad for vector computing, ToDo 2015-10-08 */
    hbuf->jj[hbuf->n-1] |= XY_FLAG; /* mark EOL, compressed sparse matrix */
    //  printf("\nDBG EOL @ %d", hbuf->n);
} /* fill buffer ahead */
    /* m1&4: storeh2 = no l1[] for cfg2idx() available;
     * store EOL to hbuf for line detection;
     * do scfg2idx(remote) and if v0 sum r*v0[abj][j](remote) */
#if 0
    if ((vvv&16) && mpi_i*pt_n+hbuf->block<8) /* 2015-04-09 */
      printf("\nvvv&16 _nhv.nhm_line task=%d hbuf-n/ahead=%d/%d=%.2f",
              mpi_i*pt_n+hbuf->block, hbuf->n, (int)nahead, hbuf->n*1./nahead);
#endif
    /* do minsymcfg + scfg2idx() etc. using sym + l1[] */
    nhm_line(hbuf, m1); /* ToDo: 2015-04 do it outside hamilton_nhv?! rm m1 */
#if 0
    if ((vvv&16) && mpi_i*pt_n+hbuf->block<8) /* 2015-04-09 */
      printf("\nvvv&16 _nhv.puth_line() task=%d",mpi_i*pt_n+hbuf->block);
#endif
    if (m1&2) hamilton_puth_line(hbuf); /* write to H-file, ToDo = above */
    /* ToDo: alternativ do sum <v1|H|v0> */
} /* hbuf->n buffer filled */

#if 1
  hbuf->i=i+nahead;   /* index of configuration 0..b_len[blk]-1, set outside!? */
#else /* old bad stuff ??? */
    for (ii=0; ii<hbuf->n; ii++) {
      if (hbuf->jj[ii]  ==i
       && hbuf->abj[ii]==mpi_i)
          hbuf->nhn[i] += D(hbuf->rr[ii]);    /* <nhn> */
      if (hbuf->jj[ii]&XY_FLAG) i++; // next matrix line
    }
    if (hbuf->v0) sum= hbuf->rr[i%hbuf->num_ahead_lines];
#endif
  return(0);                /* old: return nHv or nH1(?), new return err */
}


/*
 * void-function for pthreads, read a complete H-block and do v1=H*v0
 * return hbuf->rx=<v0|H|v0> real
 * input: hbuf->{block,v0,v1}
 * working:  ToDo: MPIperLine hbuf.n,hbuf.{jj,abj,rr}[]=H|v0>
 * output: hbuf->rx=<v0|H|v0>, v1+=H*v0
 *  called by hamilton() for diag.c:inc1
 *
 * xx is only correct for the first block ?! 
 * h_get was put into geth_block (v2.26, keywords: h_get, hget, geth, get_h)
 * for better speed, this part should be read in ca. 1000 values from vy, 
 *   than from vx0, than store to vx1 to use the CPU cache optimal
 * ToDo: h_ropen/close h_files before calling this
 * ToDo: MPI, at the moment we send mpi_n*data, so it does not scale
 *       this is only for simplicity and will be changed in future
 *
 */
void hamilton_geth_block( void *ptr ) { /* thbuf is in prinzip not needed! */
  ULong i, in, ii, ihblk, nhblks; thbuf *hbuf=(thbuf *)ptr;
  int block=hbuf->block; /* block index for x and y, 0..pt_n-1 */
  static volatile int itr=0;
  ULong xx, xy;
  scplx *v0x,*v1x;   cplx r, sum_ry;
  struct thxy *tbh;           /* H storage buffer [B_NUM].{xy,by,rr}[HBLen] */
  static struct mpi_Hbuff {   /* Data to be transferred between Nodes */
#ifdef CONFIG_MPI
    /* eats lot of memory B_NUM*HBLen*(2*cplx+2*int),  2017-01
     *  p.e. 400MB for BNUM=2 HB=2M cplx ... ToDo17: union with  SH.mpi-buff!!!
     * but before HBLen must be replaced by AHEAD*NZXMAX*FAKTOR (ToDo) 
     *  big nodes are better for big buffers getting better net performance */
    /* ToDo: some systems have big speedup if MPI buffer is aligned to 64 */
    /* Ri2017.02 BFACTOR=6 for S=1 but NOS1SYM n1=2**31.6 32tasks = BUG! */
#define BFACTOR 2 /* 1 failed, ToDo: not needed if AH * FACTOR stored */
    scplx  vr[HBLen*(1+BFACTOR)]; /* recv-buf vr[yy] from remote node (sorted) */
    int    yy[HBLen*(1+BFACTOR)]; /* send-buf (stored) yy=cfg2idx to remote node */
    int    iy[HBLen];   /* index list for sorting back results (not send) */
#endif
    scplx  vy[HBLen];   /* results for all nodes unsorted */
  } mpi_thb[B_NUM];  /* two? Hbuffers for each thread */
#ifdef CONFIG_MPI   // JS 2011-10-25
  if (sizeof(int)<sizeof(XY_TYPE)) ERR(9999); /* node_n1_bits! .yy */
#endif
  if (block>=pt_n) ERR(9999);
  /* tbh points initialy to buffer space for read/write */
  tbh=&(b_hbuf[block]);  /* ->{xy,by,rr}[HBLen] */
  v0x=hbuf->v0 + b_ofs[block]; /* v2.36 */
  v1x=hbuf->v1 + b_ofs[block];
  xx=0;      /* 0..b_len-1 */
  in=b_snz[block]; /* number of stored elements in HBlock */
  if (block==0) itr++; /* count iterations for debugging */
  sum_ry=0;
  nhblks=(b_snz[block]+HBLen-1)/HBLen; /* number of Hblocks written */
  if (nhblks>bsnz_max) ERR(9999);
  ii = 0; /* number hnz element of block, ii<HBLen */
  i  = 0; /* i is only the hnz counter, i<b_len[gblock] */
  for (ihblk=0; ihblk<bsnz_max; ihblk++) { /* nz mpi balance but rest n1? */
    if (ihblk<nhblks) tbh=h_read(tbh,block); /* v2.36 Jul08 */
    if (!tbh) {
      static int thberr=0; /* print only once */
      if (!thberr)
        E3("geth_block.h_read loop.i %lu %lu %lu",ihblk,nhblks,bsnz_max);
      thberr=1; break; /* MPI deadlock? */
    }
    /* ToDo: 2014-07 if 90% stored h_read fails, how to handle it? */
    /* problem is that block nr and hlines are async, ToDo: what to do? */
    /* copy random scattered local v0y data to the mpi buffer */
    /* ToDo: sort once to HBLen-buf + idx,len[B_NUM][mpi_n] */
    /* ToDo: unify next loop with MPI part, simpler code */
    if (mpi_n==1) { /* copy cost speed, better make stored data mpi friendly */
      for (ii=0l; i+ii<in && ii<HBLen; ii++) {
        /* get vector element from index (remote if mpi) */
        xy = (ULong)tbh->xy[ ii ]; /* incl. XY_FLAG, .by==0 */
        mpi_thb[block].vy[ii] = hbuf->v0[xy & ~XY_FLAG]; /* scplx */
      } /* for (i=0..hnz,ii=0..HBLen) */
      for (;ii<HBLen;ii++) mpi_thb[block].vy[ii] = 0; // not needed!???
    }
#ifdef CONFIG_MPI
    if (mpi_n>1)  /* should work for mpi_n=1 too, but slower */
    { /* nondiagonal hnz elements, do MPI */
      /* see nhm() for multithread problem with MPI_Sendrecv */
      int len1, len2, id, i1, by=0, rc=0;   ULong yy;
      /* instead of sorting only data in mpi_n loop, we sort it now */
      /* MPI_MAX_NUM should be greater than a CPU cache line */
      static int bsend_ofs[B_NUM+1][MPI_MAX_NUM+1]; /* offset MPI_Alltoallv */
      static int bsend_of2[B_NUM][MPI_MAX_NUM+1]; /* offset2 */
      static int bsend_cnt[B_NUM][MPI_MAX_NUM+1]; /* counter MPI_Alltoallv */
      /* we need a send for remote data (not symmetric if SH not compressed */
      static int brecv_ofs[B_NUM+1][MPI_MAX_NUM+1]; /* offset  MPI_Alltoallv */
      static int brecv_cnt[B_NUM][MPI_MAX_NUM+1]; /* counter MPI_Alltoallv */
      void *psrc, *pdst; /* src,dst for better looking code */
      // output num of dummyblocks?
      /* prepare counter */
      for (id=0;id<=mpi_n;id++) bsend_cnt[block][id]=0;
      /* count objects per destination node ... */
      if (ihblk<nhblks) /* not for dummy blocks */
      for (i1=0l; i+i1<in && i1<HBLen; i1++) {
         by = tbh->by[ i1 ]; /* node<mpi_n */
         bsend_cnt[block][by]++;  /* count for each dest. node */
      } /* for (i1=0..HBLen) */
      /* generate offsets from counts (last index (n) is HBLen) ... */
      bsend_of2[block][0]= bsend_ofs[block][0]= 0;
      for (id=1;id<=mpi_n;id++) { 
        bsend_of2[block][id]=    bsend_ofs[block][id]=
        bsend_ofs[block][id-1] + bsend_cnt[block][id-1]; }
      /* ... now we sort in groups for each destination */
      if (ihblk<nhblks) /* not for dummy blocks */
      for (i1=0l; i+i1<in && i1<HBLen; i1++) {
         xy = (ULong)tbh->xy[ i1 ]; /* incl. XY_FLAG */
         by =        tbh->by[ i1 ]; /* destination node 0<=b<mpi_n */
         len1 = bsend_of2[block][by]; /* get index to next free element */
         bsend_of2[block][by]++; /* new position, ToDo: rename */
         mpi_thb[block].iy[i1] = len1;  /* sorting back index, 2017.01.19 reversed */
         mpi_thb[block].yy[len1] = (xy & ~XY_FLAG); /* cfg2idx 0..node_n1-1 */
      } /* for (i1=0..HBLen) */
#ifdef CKPT_HELP /* 2015-02-26: get clean MPI-state for checkpointing */
{     int signal= sig1; /* spins.h: sighand.h: extern volatile int sig1; */
      MPI_Bcast(&signal,1,MPI_INT,0,MPI_COMM_WORLD); sig1=signal;
      if (sig1&15) { /* volatile.int.sig1 set by SIGUSR1 in sighand.c for ckpt */
        MPI_Barrier(MPI_COMM_WORLD); /* wait for all tasks */
        if (!mpi_i) printf("\nHLP: start 30s safe checkpoint window ...\n");
        /* ToDo: auto trigger ckpt-script by writing to a fifo here? */
        sleep(30); /* safe checkpoint window without MPI-messages */
        if (!mpi_i) printf(  "HLP: ... end of safe checkpoint window\n");
        /* restore hostnames, PIDs, clock etc. here? */
        sleep(2); sig1=0; /* wait for all restored tasks */ } }
#endif
#if 0 || DBG_NHM
      if (mpi_n<=8 && (vvv&4) && itr<4) { usleep((mpi_i+1)*10000);
        printf("\nDBG: src=%2d.%d %3lu d_sz:",mpi_i,block,ihblk);
        for (id=0;id<mpi_n;id++) printf(" %6d", bsend_cnt[block][id]);
        usleep((mpi_n-mpi_i+1)*10000); }
#endif
{ /* MPI_Alltoallv but unknown recv lengths, old i+j Sendrecv-loop */
#ifdef _OPENMP
      // static volatile int omp_token[2*B_NUM]={1,0};  // correct ini? 1 0 0 0 ..
      static volatile int omp_token=0; // slower? nicer
#endif
      /* ToDo17: send HBLen*B_NUM by thread 0, pt_barrier, bigger pckts */
#ifndef _OPENMP
      pthread_mutex_lock(&token_ring[hbuf->block]);
#else
      while (omp_token!=hbuf->block);
#endif
      rc=MPI_Alltoall(bsend_cnt[block],1,MPI_INT,
                      brecv_cnt[block],1,MPI_INT,MPI_COMM_WORLD);
      if (rc) perror("MPI_Alltoall");
      brecv_ofs[block][0]=0;    /* write idx [0..mpi_n] (one more!) */
      for (id=1;id<=mpi_n;id++){  brecv_ofs[block][id] =
         brecv_ofs[block][id-1] + brecv_cnt[block][id-1]; }
      /* Problems 2017-02-10 */
      // t8.36 4 -2 2 2 quanta2.mpi=32 SH=8m 53h 19+17=z1 k1 32*21GB(old=4GB)
      //  mpirun -n 32 # 29+7=121 27+9=130 25+11=140 23+13=137 15+21=150=6m
      //                17+19=8m=134%,t0=252%,SH=58% nzx=38*717e6=218GB=32*7GB
      //                17+19=8m=136%,SH=63% maxfile=8e9=fail alloc=4.29e9
      //        32*4.7GB_used ulimit -v 8GB! maxfile=20e9 17+19=143%
      //        4GB = problem
      //         -n 32 9+27  l=12 k= 1  0-2-15-17-30-32-9-11-6-8-21-23 1-4-16-19-31-34-10-13-7-28-22-25 3-14-18-29-33-26-12-5-27-20-24-35 
      //               maxfile=5.2e+07 t0=150%,325% -n31=280% n30.t0=258%
      //          -n16=223% -n12=202% -n4,maxf=.4e9,t0=130%
      //          -n12,maxf=1e8=128% -n8,maxf=1e7,t0=152% maxf=.8e6,t0=121%
      //     maxf=1e7,-n8 v2.56 = sig11 [B_NUM+1] + BFACTOR=150 is fix v2.60
      //                     but not v2.55 !
      //  mpirun -n 16 # 29+7=121 27+9=132 25+11=131 13+23=136 15+25=%=9m
      //  meggie -n 16 # 29+7=115 27+9=126 23+13=137 v2.60
      //            32 #                   13+23=137  -520MB*32 34GB
      if(0){ static int bfactor=110;
        if (brecv_ofs[block][mpi_n]*100 > HBLen*(bfactor+1)) {
          bfactor= brecv_ofs[block][mpi_n]*100/HBLen; // DBG JS2017-02
          printf("\nDBG %2d bfactor=%3d i,stored,all %5lu %5lu %5lu itr=%d",
            mpi_i, bfactor, ihblk, nhblks, bsnz_max, itr);
           } }
      if (brecv_ofs[block][mpi_n] > HBLen*BFACTOR /* HNZ_MPI_BUFSIZE */)
        F1("enlarge BFACTOR to %d", /* Bfactor = 1.0 */
          (brecv_ofs[block][mpi_n]+HBLen-1)/HBLen);
#if 0 || DBG_NHM /* i100 */
      if (mpi_n<=8 && (vvv&4) && itr<4) { usleep((mpi_i+1)*10000);
        printf("\nDBG: dst=%2d.%d %3lu s_sz:",mpi_i,block,ihblk);
        for (id=0;id<mpi_n;id++) printf(" %6d", brecv_cnt[block][id]);
        usleep((mpi_n-mpi_i+1)*10000); }
#endif
      /* or use AHEAD*NZX*FACTOR ! */
// if((vvv&4) && itr<4 && mpi_n<8) 
//   printf("\nMPI_1 itr=%3d %2d.%d lenS=%6d lD=%6d i=%8lu ihblk= %6lu %6lu",
//        itr,mpi_i,hbuf->block,len1,len2,xx,ihblk,nhblks);
// --- here we are at i100-iteration-section = bottleneck ---
      len1= bsend_ofs[block][mpi_n]; /* extra element used for length */
      len2= brecv_ofs[block][mpi_n]; /* to big! */
      psrc= &mpi_thb[block].yy[ 0     ]; /* 0..HBLen-1 */
      pdst= &mpi_thb[block].yy[ HBLen ]; /* HBLen..(BFACTOR+1)*HBLen-1 */
      /* using type_yy = int here, on overflow use more tasks! */
      rc=My_MPI_Alltoallv(psrc,bsend_cnt[block],bsend_ofs[block],MPI_INT,\
                          pdst,brecv_cnt[block],brecv_ofs[block],MPI_INT,\
                          MPI_COMM_WORLD);
      if (rc) perror("MPI_Alltoallv");
#ifndef _OPENMP
      pthread_mutex_unlock(&token_ring[hbuf->block+1]);
#else
#pragma omp atomic   // omp-3.0 only? gcc412.omp2.5=OK? need flush? ToDo17
      omp_token++;
#endif
      for (i1=0l; i1<len2; i1++) {  // approx HBLen/mpi_n (wasting memory)
        yy = mpi_thb[block].yy[HBLen+i1]; /* index num 0..node_n1-1 */
        mpi_thb[block].vr[HBLen+i1] = hbuf->v0[yy]; } /* for (i1=0..HBLen) */
#ifndef _OPENMP
      pthread_mutex_lock(&token_ring[hbuf->block+pt_n]);
#else
      while (omp_token!=hbuf->block+pt_n);
#endif
#if 0 || DBG_NHM
      if((vvv&4) && itr<4 && mpi_n<=8)
        printf("\nMPI_2 itr=%3d %2d.%d Alltoallv", itr,mpi_i,hbuf->block);
#endif
      // send vector elements back to the caller node
      psrc=&mpi_thb[block].vr[ HBLen ];
      pdst=&mpi_thb[block].vr[ 0 ];
      for (id=0;id<mpi_n;id++) bsend_cnt[block][id] *= (int)sizeof(scplx);
      for (id=0;id<mpi_n;id++) bsend_ofs[block][id] *= (int)sizeof(scplx);
      for (id=0;id<mpi_n;id++) brecv_cnt[block][id] *= (int)sizeof(scplx);
      for (id=0;id<mpi_n;id++) brecv_ofs[block][id] *= (int)sizeof(scplx);
      rc=My_MPI_Alltoallv(psrc,brecv_cnt[block],brecv_ofs[block],MPI_BYTE,\
                          pdst,bsend_cnt[block],bsend_ofs[block],MPI_BYTE,\
                          MPI_COMM_WORLD);
//      for (id=0;id<mpi_n;id++) bsend_cnt[block][id] /= (int)sizeof(scplx);
//      for (id=0;id<mpi_n;id++) bsend_ofs[block][id] /= (int)sizeof(scplx);
//      for (id=0;id<mpi_n;id++) brecv_cnt[block][id] /= (int)sizeof(scplx);
//      for (id=0;id<mpi_n;id++) brecv_ofs[block][id] /= (int)sizeof(scplx);
      if (rc) perror("MPI_Alltoallv");
#ifndef _OPENMP
      pthread_mutex_unlock(&token_ring[(hbuf->block+pt_n+1)%(2*pt_n)]);
//    pthread_mutex_unlock(&token_ring[(hbuf->block+1)%(pt_n)]);
#else
#pragma omp atomic   // 2017-01-10 added omp version
      omp_token++;
      if (omp_token>=2*pt_n) omp_token=0;
#endif
} /* old Sendrecv-loop id=0..mpi_n-1 */
      i1=0;
      /* sort back results from remote note */
      if (ihblk<nhblks) /* not for dummy blocks */
        for (i1=0l; i+i1<in && i1<HBLen; i1++)
           // 2017.01.19 old seq_read+rnd_write rplaced by rnd_read+seq_write
           // mpi_thb[block].vy[ mpi_thb[block].iy[i1] ] = mpi_thb[block].vr[i1];
           mpi_thb[block].vy[i1]=  mpi_thb[block].vr[ mpi_thb[block].iy[i1] ];
      for (ii=i1;ii<HBLen;ii++) mpi_thb[block].vy[ii] = 0; // not needed!??? ihblk>=nhblks?
      // output num of dummyblocks?
    } /* MPI mpi_n>1 */
    // if ((vvv&4) && itr<4) printf("\nMPI_3 %3d %2d %lu %lu", itr, mpi_i*pt_n+block, i, in); OK
    if (ihblk<nhblks) 
#endif  /* MPI */
/* ToDo17: replace HBLen by sum.nzx[0..AH-1] for simplifications, no FLAG */
    for (ii=0l; xx<min_b_sn1 && i<in && ii<HBLen; i++,ii++) {
      /* ToDo: unroll loop by get quad entries */
      /* this is the innermost loop, NFLOP=5*hnz*NumIterations */
      xy=    (ULong) tbh->xy[ ii ]; /* idx 0..node_n1=n1/mpi_n [+XY_FLAG] */
#if HRMAX>0
      r = hr[block][ tbh->rr[ ii ] ];
#else
      r =     M_f2d( tbh->rr[ ii ] );
#endif
      /* end of h_get */
      /* 2013-04-04 M_f2d added, was buggy for float16 */
      sum_ry += r * M_f2d(mpi_thb[block].vy[ii]);  /* <ii|H|v0> += <ii|H|yy> */
// if ((vvv&4) && itr<4 && sum_ry) printf("\nMPI_3 %3d %2d xx=%6lu yy=%6lu r=%.2f v=%.2f sum=%.2f",itr,mpi_i*pt_n+block,b_ofs[mpi_i*pt_n+block]+xx,b_ofs[mpi_i*pt_n+block]+xy&~XY_FLAG,r,mpi_thb[block].vy[ii],sum_ry);
      if (!(xy & XY_FLAG)) continue; /* EOL (xx) of sparse matrix H? */
      hbuf->rx += D(Conj(M_f2d(*v0x))*sum_ry); /* <vHv> should be real */
      /* 2013-04-04 avoid float16 += float16 ToDo: += (float16)float32 relevant? */
      (*v1x)    =   M_d2f(M_f2d(*v1x)+sum_ry); /* seq. r/w memory, flush cache? */
// if ((vvv&4) && itr<4 && sum_ry) printf("\nMPI_4 %3d %2d xx=%6lu v1x+=%.2f ", itr, mpi_i*pt_n+block, b_ofs[mpi_i*pt_n+block]+xx, sum_ry);
      xx++; v0x++; v1x++; sum_ry=0; /* next line */
    } /* for (i=0..hnz,ii=0..HBLen) */
  }  /* 0<=ihblk<hblks_max */

  /* now handle the nonstored elements of H-matrix
   * ToDo: blocking:
   *   return block+num_lines stored into, or dynamic growing buffer
   *   or call nhv with b_ofs + b_len (full block)
   *  + better call _nhv once for 1..ahead (block wise)!
   */
  if (vvv&256 && mpi_i<4 && mpi_i>=mpi_n-4) { /* v2.33 */
    printf("\nv&256 geth_block blk= %d.%d sn1= %lu b_len= %lu"
           " min_b_sn1= %lu extra_nhv: %lu",
      mpi_i, block, b_sn1[block], b_ofs[block+1]-b_ofs[block], min_b_sn1,
      ((max_b_len-min_b_sn1)+hbuf->num_ahead_lines-1)
                            /hbuf->num_ahead_lines);
  }
  /* recompute unstored matrix elements (slow) */
  /* old=unbalanced: 0 <= i < b_len[blk],
   * new=balanced:   min_b_sn1 <= i < max_b_len (zero padded ???) */
  for (i=min_b_sn1; i<max_b_len; i += hbuf->num_ahead_lines) {
    /* if (v0) return hbuf[i].rr = <i|H|v0> (true here), ignore i>=b_len */
    hamilton_nhv(i, hbuf->num_ahead_lines, hbuf, 0);
    for (ii=0; ii<(ULong)hbuf->num_ahead_lines
          && i+ii<b_ofs[block+1]-b_ofs[block]; ii++) {
      // r= hbuf->rr[ii];                      /* <ii|Op|v0> bug 2014-02 */
      r= hbuf->nhv[ii];                      /* <ii|Op|v0> fix 2014-02 */
      /* 2013-04-04 avoid float16 += (float16)r, use float16 (double+=) r */
      hbuf->v1[b_ofs[block]+i+ii] =
        M_d2f(M_f2d(hbuf->v1[b_ofs[block]+i+ii])+r); /* v1+= <ii|H|v0> */
      hbuf->rx += D(Conj(M_f2d(hbuf->v0[b_ofs[block]+i+ii]))*r); /* real <v0|H|v0> */
    }
  }
}

/*  v1=H*v0+v1 (parallel),  return r1=<v0|H*v0>=real
 *   (a-ib)*(c+id) = (ac+bd) + i(ad-bc) 
 * <r-i|Hr+Hi|r+i> = <r|Hr r>+<r|Hi i>+<i Hr i>+<i Hi r>
 * using long double ???
 * ToDo: hamilton_geth_block + hamilton_nhv_block => hamilton_block
 * called by diag.c:lanz1()
 *           diag.c:inc1()   v0+v1 set
 *           diag.c:incv1()
 *
 *   calls hamilton_geth_block(storedH) incl. hamilton_nhv_block(nonstoredH)
 * ToDo: Move pthread-code into MPI-loop
 */
/* v1=H*v0+v1, return <v0|H|v0> */
double hamilton(tvec *v0, tvec *v1, thbuf *hbuf[B_NUM]){
  // ULong i; 
  int block;
  /* thbuf * (hbuf)[B_NUM];  */
  double r2=0.0;  
  for (block=0; block<pt_n; block++) {
    h_ropen(block);             /* open I/O streams + buffer v2.26 */
#if 0 /* obsolete 2014-07 */
    hbuf[block] = alloc_hbuf(); /* better alloc local in thread? but free? */
    if (!hbuf[block]) {
      F2("ERROR hamilton: alloc_hbuf failed, mpi=%d thread=%d", mpi_i, block);
      exit(1);// 2014-07 continue;
    }
#endif
    reset_hbuf(hbuf[block]);   /* 2015-01-31 */
#if DBG_NHM
    { int i;for (i=0;i<HBLrVS;i++){/* better debug output *hbuf 2015-09 */
        hbuf[block]->asj[i]=0;
        hbuf[block]->an2[i]=0;
        hbuf[block]->jj[i]=0;
        hbuf[block]->acj[i]=TBASE0;} }
#endif
    hbuf[block]->block = block;
    hbuf[block]->v0 = v0;
    hbuf[block]->v1 = v1;
    hbuf[block]->rx = 0.0;
    parallel_start( block, hamilton_geth_block, hbuf[block] );
  }
  for (block=0; block<pt_n; block++) {
    /* if (hbuf[block]) */ parallel_wait(block); /* only for threads */
    h_rclose(block); /* v2.26 close buffer chain and finaly the file */
    // if (!hbuf[block]) continue;
    r2 += hbuf[block]->rx;    /* <vHv> */
/*    free_hbuf(hbuf[block]); obsolete ToDo */
  }
#ifdef CONFIG_MPI
  { double add=r2, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r2=sum; }
#endif
  return r2;
}

/* void-function for pthreads
 * ToDo: return compute time to calculate load balance
 *  input: blk=0..pt_n-1 (its cheaper to add mpi_i*pt_n as dividing by pt/n)
 * must be called for all blocks per mpi-process avoiding deadlocks
 */
void storeh_block( void *ptr ){
  ULong i, iout;
  double tm1, tm2; /* time in seconds 2013-04-23 */
  thbuf *hbuf=(thbuf *)ptr;
  int block=hbuf->block,            /* node block    0..pt_n-1 */
      c_blk=mpi_i*pt_n+hbuf->block; /* cluster block 0..pt_n*mpi_n-1 */
  tm1=fstime();  /* 2013-04-23 itime (long) replaced by fstime (double) */
  iout=1<<6;  /* output alive msg sparingly, N=40: 2^29 8outs (old=1<<12) */
  /* one MPI transaction per loop, max(b_len[c_blk]) for balance */
  for(i=0; i<max_b_len; i+=hbuf->num_ahead_lines){
    if ((!c_blk) && iout && i>=iout) { /* give alive signals */
      /* ToDo: sync threads and output min/max? 2013 */
      tm2=fstime()-tm1; iout<<=2; /* factor 4, iout may overflow to 0 */
      if (tm2>2) /* old: 10s .. 40s */
        printf("\n SH %6.2f%% hnz/i= %6.2f nzxmax= %3d nz/s= %8.2e"
               " t100%%= %6.2fm",
                  i*100./b_ofs[1], b_hnz[0]*1./i,
                  (int)statH[0].nzxmax,       /* 2012-01-23 speed tuning */
                  b_hnz[0]*pt_n*mpi_n*1./tm2, /* 2014-06 speed nz/s */
                  tm2/60.*b_ofs[1]*1./i);
    }
    /* ToDo: return number of lines filled into the buffer? */
    /* ToDo: send/recv Hij elements to/from other nodes (cached) */
    /* storeH to disk */
    hamilton_nhv( i, hbuf->num_ahead_lines, hbuf, 2);
  } /* loop i<max_b_len */
#if 0
  if ((vvv&4) && (c_blk<4 || c_blk>=mpi_n*pt_n-4)) 
    printf("\nvvv&4 SH_block.end %3d sn1= %9lu snz= %9lu hnz= %9lu",
             c_blk, b_sn1[block], b_snz[block], b_hnz[block]);
#endif
  if (err_fatal && (mpi_i&(mpi_i-1))==0) /* 2015-02 log output */
    printf("\nERR%d: %4d.%02d i=%.1e= %6.2f%% h_nz=%.1e (SH_block.end)",\
       err_fatal,mpi_i,block, 1.*b_sn1[block], b_sn1[block]*100./max_b_len,
        1.*b_hnz[block]); /* 2016-01-23 add b_sn1 */
  hbuf->rtime=fstime()-tm1;
}

/* store H elements on disk, reading disk is faster than computing!? Aug07 */
void storeh(thbuf *hbuf[B_NUM]){
  int block;
  double tm1;               /* fstime() = MPI_Wtime() runtime in seconds */
  /* thbuf * (hbuf)[B_NUM]; ToDo: obsolete */
  double h_nn;
  double     rtime_max,       rtime_min; /* balance + timing measurement */
  int  block_rtime_min, block_rtime_max; /*  thread index for min/max timings */
  rtime_max=rtime_min=0.0;
  block_rtime_min=block_rtime_max=0; /* avoid gcc warning */
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (!mpi_i) {
    if (vvv) printf("\n --- starting hilbert matrix generation (storeh) ---");
    /* (buffersize-overhead)/numelements=size_per_element */
    if (vvv) 
      printf("\n NZXMAX= %d  AHEAD_LINES= %d nzSize=HSize/HBLen: %dK/%dK=%d",
      NZXMAX, NUM_AHEAD_LINES, (int)Hsize/1024, HBLen/1024, (int)Hsize/HBLen);
    if (vvv&(64+2)) {  /* ToDo: measure factor! */
      printf("\n vvv&64 (=%d) to output matrix elements",vvv&64);
    }
    /* meggie 2015-03 sq40 8+32 gcc492+SSE3+3dNOW */
    printf("\n vvv&2 (=%d) statH (speed-5(2)%%), vvv&64 prtH",vvv&2);
  }
  statH_ini();	/* initialize H-statistics Oct99 */

  ini_thxy(n1); /* see h_file.c, initialize H-storage, print size needed */
  /* 2015-04-09 moved to h_file.c ini_thxy(): 
   *  if (!mpi_i) printf(" sizeof(HR_TYPE)=%d\n",(int)sizeof(HR_TYPE)); */
  tm1=fstime();
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (checkpoint<4) { // old 3
    for(block=0;block<pt_n;block++) {
      h_wopen(block); /* open I/O streams + ini buffer */
      /* big local vars (sbase), relatively bad for parallelisation */
      // hbuf[block]=alloc_hbuf(); /* ToDo: alloc data area within the thread */
      reset_hbuf(hbuf[block]); /* 2015-01-31 */
#if DBG_NHM
      { int i;for (i=0;i<HBLrVS;i++){/* better debug output *hbuf 2015-09 */
          hbuf[block]->asj[i]=0;
          hbuf[block]->an2[i]=0;
          hbuf[block]->jj[i]=0;
          hbuf[block]->acj[i]=TBASE0;} }
#endif
      hbuf[block]->block=block; /* overall block number */
      /* start compute threads, hbuf->v0=v1=NULL */
      parallel_start( block, storeh_block, hbuf[block] );
    }
    for(block=0;block<pt_n;block++){ int i, blk=block;
      parallel_wait(block); /* stop compute threads and let them wait */
      if (b_bnz[blk]!=0 && !h_error){   /* flush buffer to disk-file */
        for (i=b_bnz[block];i<HBLen;i++){
          b_hbuf[block].xy[i]=0;
          b_hbuf[block].by[i]=0;
          b_hbuf[block].rr[i]=0;
        }
        if (my_fwrite((void *)&b_hbuf[blk],Hsize,1,blk)!=1) {
          hfmax=b_snz[blk]; /* ToDo 2014: cmp + set to mpi_min */
          b_bn1[blk]=0; b_bnz[blk]=0; h_error=1;
        } else { /* reset counter buffered_nonzero-entries */
          b_sn1[blk] += b_bn1[blk]; b_bn1[blk]=0; /* not sharp!? */
          b_snz[blk] += b_bnz[blk]; b_bnz[blk]=0;
        }
      } /* if HBLen full */
      h_wclose(block);      /* flush H-buffers to disk */
      if (block==0) {
        rtime_max=rtime_min=hbuf[0]->rtime; /* balance + timing measurement */
        block_rtime_min=block_rtime_max=0;
      } else {
        if (rtime_min>hbuf[block]->rtime) {
            rtime_min=hbuf[block]->rtime; block_rtime_min=block+mpi_i*B_NUM;}
        if (rtime_max<hbuf[block]->rtime) {
            rtime_max=hbuf[block]->rtime; block_rtime_max=block+mpi_i*B_NUM;}
      }
      // free_hbuf(hbuf[block]);
    }
    MPI_Barrier(MPI_COMM_WORLD); /* block to get walltime, ToDo, better get maxtime? */
    tm1=fstime()-tm1;
    /* ToDo: output time in a form where it can be removed for comparisions
     * example: diff <(sed 's/\([0-9]{2}\)/#1/g' out1) <(sed '' out2)
     * ToDo: output trace?
     */
    h_checkpoint(1); /* write hr_saveNNN.dat for STOREH!=0 */
    checkpoint=5; if(!mpi_i)fputs(" checkpoint5.w", stdout); // SH=45 old=3
    if(!mpi_i){FILE *f;f=fopen(FCHKPT,"wt");fputs("5\n",f);fclose(f);}
  } else h_restore(1); /* checkpoint>=4 read hr_saveNNN.dat */
  /* storeh complete, print statistics */
  
#ifdef CONFIG_MPI
  { ULong  t1[3], t2[3]; /* ULLong may be a problem for MPI */
    t1[0]=h_error;
    MPI_Allreduce(t1,t2,1,MPI_UNSIGNED_LONG, MPI_MAX, MPI_COMM_WORLD);
    h_error=t2[0];
    t1[0]=hfmax;
    MPI_Allreduce(t1,t2,1,MPI_UNSIGNED_LONG, MPI_MIN, MPI_COMM_WORLD);
    if (hfmax>t2[0] && !mpi_i) printf("Warning: hfmax reduced by mpi\n");
    hfmax=t2[0];
  }
#endif  // Apr09

#if 0 /* ToDo: 2014-07-17 disable, or do later, SIGSEGV L1295 */
  /* h_error may be set by some nodes */
  if (h_error){  /* error during write is fatal, ignore all things written */
    /* not a good solution, if 90% SH fits to mem, nothing is used = slow */
    if (!mpi_i) E0("disk/mem full? stored data will not be used (ToDo: change!)"); 
    h_error=0; 
    h_nn=0;
    for(block=0;block<B_NUM;block++){ b_snz[block]=0; b_sn1[block]=0; }
  }
#else
  if (h_error){  /* error during write is fatal, ignore all things written */
    /* not a good solution, if 90% SH fits to mem, nothing is used = slow */
    if (!mpi_i) E0("disk/mem full? try to use it partly, may fail (ToDo: change!)"); 
    h_error=0; 
    /* check 2014-08-07 4 threads 72% stored diff 1e-8 !? */
  }
#endif
  
  /* max/min/sum hnz_blocks per thread (balance?) */
  bsnz_max=bhnz_sum=bhnz_max=0;
  bhnz_min=((b_hnz[0]+HBLen-1)/HBLen)*mpi_n*pt_n;
  min_b_sn1=max_b_len;  /* pseudo minimum stored n1 of all threads */
  /* ToDo: measure balance, hnz_min*pt_n/h_nz%..hnz_max... */
  for (block=0;block<pt_n;block++) { /* ToDo: nondiag_blocks per thread */
    /* ToDo: using HBLen as unit is confusing versus AHEAD
           because HBLen should be (max of all ahead blocks of) AHEAD*nzx */
    ULong nblks=(b_hnz[block]+HBLen-1)/HBLen;
    if  (nblks!=(b_hnz[block]+HBLen-1)/HBLen) F0("ULong overflow, enlarge HBLen");
    /* encode info about blk-src */
    if (bhnz_max<nblks*mpi_n*pt_n) bhnz_max=nblks*mpi_n*pt_n+mpi_i*pt_n+block%pt_n;
    if (bhnz_min>nblks*mpi_n*pt_n) bhnz_min=nblks*mpi_n*pt_n+mpi_i*pt_n+block%pt_n;
    nblks=(b_snz[block]+HBLen-1)/HBLen;
    if (bsnz_max<nblks) bsnz_max=nblks;
    bhnz_sum+=nblks;
    if (b_sn1[block]<b_ofs[block+1]-b_ofs[block]) 
      if (min_b_sn1>b_sn1[block])
          min_b_sn1=b_sn1[block];
  }
  for (h_nz=0,block=0;block<pt_n;block++) h_nz+=(double)b_hnz[block];
  for (h_nn=0,block=0;block<pt_n;block++) h_nn+=(double)b_snz[block];
#ifdef CONFIG_MPI
  { double r1[2],r2[2]; int i0;
    ULong  t1[10],t2[10]; /* ULLong may be a problem for MPI */
    /* stored + overall hnz + non stored n1 */
    t1[0]=bsnz_max; t1[1]=bhnz_max; t1[2]=mpi_maxlen;
    MPI_Allreduce(t1, t2,      3, MPI_UNSIGNED_LONG, MPI_MAX, MPI_COMM_WORLD);
    bsnz_max=t2[0]; bhnz_max=t2[1]; mpi_maxlen=t1[2];
    t1[0]=min_b_sn1;
    MPI_Allreduce(t1, t2,      1, MPI_UNSIGNED_LONG, MPI_MIN, MPI_COMM_WORLD);
    min_b_sn1=t2[0];
    t1[0]=bhnz_min;
    MPI_Allreduce(t1,&bhnz_min,1, MPI_UNSIGNED_LONG, MPI_MIN, MPI_COMM_WORLD);
    t1[0]=bhnz_sum;
    MPI_Allreduce(t1,&bhnz_sum,1, MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
    r1[0]=h_nz; r1[1]=h_nn;
    MPI_Allreduce(r1, r2,      2, MPI_DOUBLE,        MPI_SUM, MPI_COMM_WORLD);
    h_nz=r2[0]; h_nn=r2[1];
    t1[0]=mpi_traffic;
    t1[1]=mpi_syncs;
    /* t1[2]=mpi_z_pkts; */
    for (i0=0;i0<8;i0++) t1[2+i0]= mpi_stats_npkts[i0];
    MPI_Allreduce(t1,t2,       3, MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
    mpi_traffic=t2[0];
    mpi_syncs=t2[1];
    for (i0=0;i0<8;i0++) mpi_stats_npkts[i0]= t1[i0+2];
    /* mpi_z_pkts=t2[2]; */
  }
#endif  // Jan09
  /* see comments in doc/speed_estim.html */ 
  if (!mpi_i) {
    /* parallel computing balance: (compare to b_hnz = length of htmp*) */
    printf("\nPRF: SH.t_min=  %5.2f blk= %d", rtime_min/60., block_rtime_min);
    printf("\nPRF: SH.t_max=  %5.2f blk= %d", rtime_max/60., block_rtime_max);
    printf("\nPRF: SH.t[min]= %5.2f", tm1/60.);
    /* time in seconds, no div 0, t/nz is per task, speed is agreagated */
    printf( "\nPRF: SH_speed.nz/s= %8.2e  t[ns]/nz= %7.1f (per core) %3.1f*%d",
     h_nz*1./((tm1)?tm1:1)/(mpi_n*pt_n), tm1*1.0e9/h_nz*mpi_n*pt_n,
     tm1*1.0e9/h_nz*mpi_n*pt_n/(1+sym.numall), 1+sym.numall);
    printf( "\nPRF: SH_speed.nz/s= %8.2e  t[ns]/nz= %7.1f (%d cores)",
     h_nz*1./((tm1)?tm1:1), tm1*1.0e9/h_nz, mpi_n*pt_n);
      /* ToDo: per task! ? */
    if (opt_perf_test) printf("\nPRF: speed is overestimated (ToDo)\n");
  }
#ifdef CONFIG_MPI
  if (vvv && mpi_n!=1 && !mpi_i) {
    /* t=bytes/BW+pkts*latency, maxlen in data objects per mpi-task */
    printf("\n MPI traffic  pkts/task=2*n1/ahead*mpi_n= %8g ?= %8g (counted)",
   2.*(((n1+mpi_n-1)/mpi_n+NUM_AHEAD_LINES-1)/NUM_AHEAD_LINES)*mpi_n*mpi_n,
    mpi_syncs*2.);
    /* for lot of zero_packets we should higher AH or remove sendrcvs */
    /* max_pcksize=HNZ_MPI_BUFSIZE=NZXMAX*NUM_AHEAD_LINES 
                     mpi_maxlen*sizeof(tbase) */
    printf("\n MPI traffic  bytes/nz= %d "/* zpkts= %8g */
                                         " maxpktlen= %lu*tbase",
            (int)(sizeof(tbase)+Lsize),
       /* mpi_traffic/hnz=1, */ /* mpi_z_pkts*1., */ mpi_maxlen);
    /* ToDo: max + mean packet size, ahead dependent? */
    printf("\n MPI traffic ..size/kB=%6d numpkts=%8g",1<<0,mpi_stats_npkts[0]*1.);
    printf("\n MPI traffic ..size/kB=%6d numpkts=%8g",1<<2,mpi_stats_npkts[1]*1.);
    printf("\n MPI traffic ..size/kB=%6d numpkts=%8g",1<<4,mpi_stats_npkts[2]*1.);
    printf("\n MPI traffic ..size/kB=%6d numpkts=%8g",1<<6,mpi_stats_npkts[3]*1.);
    printf("\n MPI traffic ..size/kB=%6d numpkts=%8g",1<<8,mpi_stats_npkts[4]*1.);
    printf("\n MPI traffic ..size/kB=%6d numpkts=%8g",1<<10,mpi_stats_npkts[5]*1.);
    printf("\n MPI traffic ..size/kB=%6d numpkts=%8g",1<<12,mpi_stats_npkts[6]*1.);
    printf("\n MPI traffic size../kB=%6d numpkts=%8g",1<<12,mpi_stats_npkts[7]*1.);
  }
  mpi_traffic=0;
  mpi_syncs=0;
  mpi_maxlen=0;
  /* mpi_z_pkts=0; */
  mpi_stats_npkts[0]=0;
  mpi_stats_npkts[1]=0;
  mpi_stats_npkts[2]=0;
  mpi_stats_npkts[3]=0;
  mpi_stats_npkts[4]=0;
  mpi_stats_npkts[5]=0;
  mpi_stats_npkts[6]=0;
  mpi_stats_npkts[7]=0;
#endif
  if (vvv && !mpi_i) {
    printf("\n n1 = hilbert dimension,  hnz = number of nonzero matrix elements");
    /* ToDo: better output counted hnz: n1 hnz/n1 ..., 99.9999% as 99%!? */
    printf("\n n1= %.2e nzx=hnz/n1= %.3f  hnz_stored= %3d%%  sn1_stored= %.0f",
            1.*n1, h_nz/n1, ((h_nz)?(int)((100*h_nn)/h_nz):0), 1.*min_b_sn1);
    /* balance=max_efficience=max(tm1/(n*tn))=sumhnz/(n*blkhnzmax)
     * ToDo: better hnz instead of blocks (case 1block/thread) !? */
    printf("\n hnz_blks/threads= %9.2f"
           " balance= %.0f..%.0f max_eff=%.1f%% min,maxblk=%lu,%lu",
             bhnz_sum*1./(pt_n*mpi_n),
             /* min,max has its blocknum encoded, decode it */
             (bhnz_min/(mpi_n*pt_n))*1.,
             (bhnz_max/(mpi_n*pt_n))*1.,
             (bhnz_sum*100./(pt_n*mpi_n))/(bhnz_max/(pt_n*mpi_n)),
             (bhnz_min%(mpi_n*pt_n)),
             (bhnz_max%(mpi_n*pt_n)) );
  }
  if (mpi_i==0 && mpi_n>1 && 2*(h_nz/node_n1)<NZXMAX)
    printf("\nWarning 2hnz/n1<NZXMAX: waste of memory, see v&2");
  if (mpi_i==0 && mpi_n>1 && MaxSBase>1)
    printf("\nWarning MaxSBase>1: waste of MPI bandwidth (factor %.1f),"
           "\n please reduce MaxSBase in spins.h to 1",
           sizeof(thbuf)*1./(sizeof(thbuf)
              -(MaxSBase-1)*(sizeof(tbase)+sizeof(tphase))));
  if ((!mpi_i) && h_nn<h_nz) {
    printf(
     "\nWarning: H stored incompletely, need maxfile=%.2e for best speed.",
            1.*Hsize*h_nz/HBLen/(mpi_n*pt_n)+Hsize);
  }
  if (vvv&2) { 
    statH_out(); /* ToDo: speed_test using rnd cfgs, error handling */
    /* ToDo: set errornumber in hbuf !? */
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
  }
  return;
}

/* --------------------- Jan13 2000 -------------------------------
 * new way for storing l1 and H, is it faster than the old way?
 * it should generate the symmetric states itself by
 * - starting with symmetric config
 * - using iterativ H, and symconfig and sort to generate new sym-configs
 * - until H generates no new configs (m1=0 run1 for n1, m1=1 run2 for sh)
 * algo&8:
 * ToDo: make MPI ready! sort in using ising-ergs + hashing!
 *       or mean-(diag|ising)erg of 0th and 1st excitations
 *       to have best blocking and small number hnz in of diagonal blocks
 * use kind of MC? 
 *    1st step generate mpi_n*pt*100 states and part it to mpi_n*pt blocks
 *    generate all states between start_block_i and start_block_i+1
 * mpi_n = 1 !!!
 */
void storeh2(ULong m1, thbuf *hbuf){
  ULong tm1,i1,i2,i3,i4,i5,i6,i0; int j1,j2,block;
  struct sbase b1,b2; tbase ib; double h_nz, h_nn;
  /* 2014-07-17 old local: thbuf *hbuf; */
  /* hbuf=alloc_hbuf();  */ /* very big */
  reset_hbuf(hbuf);  /* 2015-01-31 */
  hbuf->num_ahead_lines=1; /* bad, but code is not ready */
  if (vvv) printf("\n storeh2 (warning, buggy code, needs review)");
  // b1.o=0; 
  statH_ini();	/* Oct99 */
  /* das speichern des letzten hnz der Zeile macht probleme! da sich n1 aendert! */
  ini_thxy(n1); /* compute bytes needed for storing H elements, ini files */
  tm1=itime();
  i1=0;		/* pointer to last unchecked state */
  i2=1;		/* pointer to next free entry */ 
  if(m1==0){		/* starting state == Neelstate (lowest ising erg) */
    ib=TBASE0;i0=0l;
    for(j1=0;j1<2*nd;j1+=2) ib=puti(ib,j1%nn+j1/nn,2); /* ib=Neel */
    /* MC steps to lower Ising energy if possible, only JJ-model, ToDo */
    /* ToDo: more simple: erg=nHn + nhv to get next n for next better min? */
    {
      int *k1, *k2;   /* pair correlations == ising erg */
      tbase ib2, ib3=ib;      /* test state */
      int z1, z2, j3;
      for (j3=2;j3>0;j3--) { /* down counter */
        k1=get_dcorr(ib ,3, NULL);
        for (j1=0   ;j1<nn-1;j1++)
        for (j2=j1+1;j2<nn  ;j2++)
          if(j1!=j2 && (z1=geti(ib,j1))!=(z2=geti(ib,j2))) {
          /* now j1 and j2 are different neighbours */
          ib2=puti(puti(ib,j2,z1),j1,z2); /* j1 <=> j2 */
          k2=get_dcorr(ib2, 1, NULL); /* get NumIsingExcitations, ToDo: ising, not dcorr */
          /* prefer more movable pairs, ToDo: check all corrs */
          if (k1[1*CM(2,4,5)]< k2[1*CM(2,4,5)] ||  /* smallest j1 */
             (k1[1*CM(2,4,5)]==k2[1*CM(2,4,5)] &&
              k1[2*CM(2,4,5)]< k2[2*CM(2,4,5)])) { 
                 j3=2; ib=ib2; free(k1); k1=k2; j1=j2=nn;
                 minsymcfg_dflt(ib,&b1);
                 if (b1.sign!=0) ib3=ib; /* store sign!=0 */
              }
          else free(k2);
        }
        free(k1);
      }
      ib=ib3;
    }
    printf("\n start konfig= %s",ccfg(ib));
    minsymcfg_dflt(ib,&b1); ib=b1.scfg;
    printf("\n         scfg= %s  sign=%+g",ccfg(b1.scfg), D(b1.sign));
    if( b1.sign==0
#ifdef CONFIG_ANSATZ
     || 0==ansatz_test(hbuf->acj[j1])		/* variational ansatz */
#endif
    ) {
    printf("\n bad start konfig (sign==0 || proj==0), try another");
    for(i0=0;i0<10000;i0++){ ib=ll1(i0); minsymcfg_dflt(ib,&b1); ib=b1.scfg;
      if( b1.sign!=0
#ifdef CONFIG_ANSATZ
      && 0!=ansatz_test(hbuf->acj[j1])		/* variational ansatz */
#endif
      )break; }
    }
    putl1(0,ib,l1);	/* first state = starting state, n1=1 ??? */
    if (l2b) l2b[0]=0;	/* sorting-idx */
    printf("\n start konfig= %s",ccfg(ib));
  } else i2=m1;   /* start with picked state */
  h_wopen(0);
  if (mpi_i==0 && mpi_n>1) E0("missing MPI code in storeh2()");
  MPI_Barrier(MPI_COMM_WORLD);
  for (;;) {
    hbuf->n=0;
    /* ToDo: i1 must match 0..(b_len[blk]-1) v2.27 */

    /* ToDo: set num_ahead_lines=1 ? */
    hamilton_nhv( i1, 1, hbuf, 4);    /* store nHm only in buf_ */
    if (err_fatal) { E1(" err_fatal: i1=%lu",i1); break;}
    if (i1<100 && (vvv&64)) { printf("\n H|i1> = H|%s> i1=%5lu i2=%5lu",
                       ccfg(getl1(i1)),i1,i2);
                       get_dcorr(getl1(i1),2,NULL);
    }
    minsymcfg_dflt(getl1(i1),&b2);	/* need for b2.norm2 */
    if(Norm(b2.sign)<1e-6) E0(" should not happen!");

    /* set scfg (compress extracted in v2.29, getbase replaced) */
    for (j1=0;j1<hbuf->n;j1++) if (hbuf->rr[j1]!=0) /* test all H|cfg> */
    {					/* every cfg with <cfg|H|i1>!=0 */
      minsymcfg_dflt(hbuf->acj[j1],&b1);		/* dont forget is==0!!! ??? */
      /* representator, better use w2w ??? */
      hbuf->rr[j1]*=b1.sign;
      hbuf->acj[j1] =b1.scfg;
      if(b1.norm2!=0 && b2.norm2!=b1.norm2)
         hbuf->rr[j1]*=sqrt((double)b2.norm2/(double)b1.norm2);
    }		/* debugging friendly src-code ??? */

    /* compress */
    for (j1=0;j1<hbuf->n;j1++) if (hbuf->rr[j1]!=0) /* test all H|cfg> */
    {					/* every cfg with <cfg|H|i1>!=0 */
      /* search hbuf->acj[j1] in rest of hbuf, compression */
      for(j2=j1+1;j2<hbuf->n;j2++)
        if (hbuf->rr[j2]==0
         || hbuf->acj[j2]==hbuf->acj[j1]) {
        hbuf->rr[j1] += hbuf->rr[j2];
        hbuf->n--;
        hbuf->rr[j2] = hbuf->rr[hbuf->n];
        hbuf->acj[j2] = hbuf->acj[hbuf->n]; /* compress */
        j2--;
      }
    }		/* debugging friendly src-code ??? */

    /* remove zeros */
    for (j1=0;j1<hbuf->n;j1++) if (hbuf->rr[j1]==0) /* test all H|cfg> */
    {					/* every cfg with <cfg|H|i1>!=0 */
        hbuf->n--;
        hbuf->rr[j1] = hbuf->rr[hbuf->n];
        hbuf->acj[j1] = hbuf->acj[hbuf->n]; /* compress */
        j1--;
    }		/* debugging friendly src-code ??? */

/* parallel ???????, braucht immense Zeit */
    i4=i2;
    for(j1=0;j1<hbuf->n;j1++)if(hbuf->rr[j1]!=0){	/* allready stored? */
      if(hbuf->acj[j1]==getl1(i1)){ hbuf->jj[j1]=i1;continue; }
      i3=i2;
      if(l2b)
      {	/* suche im sortierten hilfsindex - schneller?! */
        for(i6=1;i6<=i4;i6<<=1);	/* i4 - hintere bits */
        for(i5=0;i6;i6>>=1){		/* interval-Halbierung */
          if (i5+i6>=i4) continue;      /* i+i/2 > max */
          if (getl1(l2b[i5+i6])<=hbuf->acj[j1]) i5+=i6;
        }
        if(getl1(l2b[i5])==hbuf->acj[j1]){
              hbuf->jj[j1]=l2b[i5];continue; } /* found, min_j? */
        if(i4>=n1){ hbuf->rr[j1]=0;continue; } /* 0-Projection, count? */
        /* insert new element, max_j? -> max(maxj-minj) v-size? */
        for(i6=i4;i6>i5;i6--)l2b[i6]=l2b[i6-1];
        if(i5==0 && hbuf->acj[j1]<=getl1(l2b[i5])) l2b[0]=i4; else
        l2b[i5+1]=i4;
      } else {	/* sequential searching is very slow! */
        for(i3=0;i3<i2;i3++)if(hbuf->acj[j1]==getl1(i3))break;
        if(i3<i2){ hbuf->jj[j1]=i3;continue; }	/* already stored */
      }
      if(i4>=n1){ hbuf->rr[j1]=0;continue; }
      putl1(i4,hbuf->acj[j1],l1); hbuf->jj[j1]=i4;
      i4++;
      if(i4==n1) E1("n1-ovl, n1=%.0f", 1.*n1); /* report once */
    }
    if(i4==1 && hbuf->n==0){ ERR(500);break; }
    if(i4>=n1){ E0("n1-ovl");break; }
    hbuf->i=i1;
    hamilton_puth_line(hbuf);	/* write Hamiltonian!!! run2 n1=correct */
    i2=i4;i1++;if(i1>=i2)break;

  }
  if (b_bnz[0]!=0 && !h_error){   /* flush buffer to disk-file */
    int i;
    for (i=b_bnz[0];i<HBLen;i++){
      b_hbuf[0].xy[i]=0;
      b_hbuf[0].by[i]=0;
      b_hbuf[0].rr[i]=0;
    }
    if (my_fwrite((void *)&b_hbuf[0],Hsize,1,0)!=1) {
      hfmax=b_snz[0]; /* ToDo 2014: cmp + set to mpi_min */
      b_bn1[0]=0; b_bnz[0]=0; h_error=1;
    } else { /* reset counter buffered_nonzero-entries */
      b_sn1[0] += b_bn1[0]; b_bn1[0]=0; /* not sharp!? */
      b_snz[0] += b_bnz[0]; b_bnz[0]=0;
    }
  } /* if HBLen full */
  h_wclose(0); 
  tm1=itime()-tm1; h_checkpoint(1); n1=i1;
  if (vvv) printf(" SH2 nk=%lu n1=%lu\n",i2, i1);
  if (vvv) printf("PRF: SH2.t=%s\n",sclock((long)tm1));
  if (vvv&2) statH_out(); /* storeh2 ToDo: error handling hbuf->err=2 */
  /* hnz = number of stored elements, snz = number of stored lines */
  for(h_nz=0,block=0;block<pt_n;block++) h_nz+=(double)b_hnz[block];
  for(h_nn=0,block=0;block<pt_n;block++) h_nn+=(double)b_snz[block];
  if (vvv && !mpi_i)
    printf("\n h_nz = num of matrix elements, h_nn = stored elements");
  printf("\n h_nz %.3e h_nn %.3e %3.0f%% hxy %d",
             h_nz,     h_nn,(h_nn*100.)/h_nz,(int)Hsize/HBLen);
  if(h_nn<h_nz)
  printf("\nWarning: %.2e B for maximum speed needed.", 1.*Hsize*h_nz/HBLen);
  free_hbuf(hbuf);
  return;
}

int outbasis(FILE *f){
  ULong i;
  static struct sbase b1; int i1,i2,i3,i4,i6; // b1.o=0;
  if (vvv&64) { 
    if (!mpi_i) fprintf(f,"\n vvv&64: outbasis.x_max=%lu B=[", x_max);
    if (node_ofs<x_max) usleep(10000*mpi_i); /* sort parallel output */
    if (node_ofs<x_max)
    for (i=0l;i<node_n1 && node_ofs+i<x_max;i++){
      if (mpi_i!=0 || i!= 0l)fprintf(f,",");
      if (node_ofs+i<x_max) 
              fprintf(f,"\n%4lu %s ",(ULong)node_ofs+i,ccfg(getl1(i)));
      minsymcfg_dflt(getl1(i),&b1);fprintf(f,"*q(%4g)=(",(double)b1.norm2);
      /* this is ising like energy for all diff. correlations */
      /* ToDo: replace by get_dcorr */
      for(i4=0,i1=1;i1<sym.num_dcorr;i1++){ i6=0;
         for(i2=   0;i2<nn;i2++)
         for(i3=i2+1;i3<nn;i3++)
         if(sym.corr[i2][i3]==i1) {
          if(geti(b1.scfg,i2)==1 && geti(b1.scfg,i3)==1); else
          if(geti(b1.scfg,i2)==2 && geti(b1.scfg,i3)==2); else
          i6++; /* ud - symmetric */
         }
         i4+=i6;
         fprintf(f," %2d",i6); /* fprintf(f,"+%2d+%2d",i4,i6); 2012-05 */
      }
      fprintf(f," sum= %d)  ",i4);
      if (node_ofs+i==x_max-1 && n1>x_max) {/*sleep(1);*/fprintf(f," ... ");}
    } /* for i=0...node_n1 */
    MPI_Barrier(MPI_COMM_WORLD);
    if (mpi_i+1==mpi_n) { fprintf(f,"]\n"); fflush(f); }
    /* sleep(1); */ /* sort parallel output */
  } /* if (vvv&64) */
  return(0);
}

double sq( double x){ return x*x; }

/* calculate the geometrical distance^2 between site i1 and i2 */
double distance( int i1, int i2){
  double d1,d2;
  int i,j,k;
  d1=sq(pos[i1][0]-pos[i2][0])
    +sq(pos[i1][1]-pos[i2][1])
    +sq(pos[i1][2]-pos[i2][2]);
  for(i=-1; i<2; i++)
  for(j=-1; j<2; j++)
  for(k=-1; k<2; k++) { /* brute force, test surrounding boxes */
    d2=sq(pos[i1][0]-pos[i2][0]+i*edg[0][0]+j*edg[1][0]+k*edg[2][0])
      +sq(pos[i1][1]-pos[i2][1]+i*edg[0][1]+j*edg[1][1]+k*edg[2][1])
      +sq(pos[i1][2]-pos[i2][2]+i*edg[0][2]+j*edg[1][2]+k*edg[2][2]);
    if (d2<d1) d1=d2;
  }
  return d1;
}

/*
 ***********************************************************************
 * output a table(j1)/matrix(j1,j2) of Ising weigths (picture of state?)
 * using w2w-matrix (output via vvv&8)
 */
#define Imax (2*NN)    /* maximum number uu-pairs per bond type */
extern int w2w[NN][NN];

struct s_ising_weight {
  int blk; /* thread number */
  ULong  states[Imax*Imax];  /* number of states */
  double weight[Imax*Imax];  /* weight matrix */
  int   d1a,d1b, d2a,d2b;     /* matrix dimension (maxIsingErg=max_ud_per_bond) */
  tvec *v0;
} ising_ptr[B_NUM];

void ising_weight_thread( void *ptr ) {
  /* map an site pair nisnjs to a Ising index to count up, see get_dcorr */
  const char map[16]= /* fits to L1 cache line !? */
    {  2, 3, 3, 4 ,  3, 0, 1, 3 ,  3, 1, 0, 3 ,  4, 3, 3, 2 };
   /* .. .u .d .3   u. uu ud u3   d. du dd d3   3. 3u 3d 33 => 0..4 = 3bit */
   /* index: 0=uu+dd  1=ud+du  [2=..+33  3=.u+u.+.d+d.  [4=.3+3.]] */
  int   *buf=NULL;  /* buffer for counted pairs (IsingErg) */
  struct s_ising_weight *pp=(struct s_ising_weight *)ptr;
  int blk=pp->blk;
  ULong i1;       /* state counter */
  tbase b1;       /* cfg[i] */
  int j1, j2;     /* num excitations */
  int i2, i3, i4, ib1, ib2; /* work variables */
  buf=(int *)malloc(CM(2,4,5)*(numpara+2)*sizeof(int));
  if (!buf) { E0("ising_weight: malloc failed"); return; }
  /* upper border of IsingErgs=Neel is nn, add a overflow index nn */
  for (j1=0;j1<Imax*Imax;j1++) {
    pp->states[j1]=0;
    pp->weight[j1]=0;
  }
  pp->d1b=pp->d2b=0;
  pp->d1a=pp->d2a=Imax;

  for(i1=0l; i1<b_ofs[blk+1]-b_ofs[blk]; i1++){  /* l1[] sort for norm2[]??? */
    b1=getl1(i1+b_ofs[blk]);
    // l2.num=0; l2.o=0; minsymcfg_dflt(ib1,&l2); /* get norm2 = slow */
    /* reset IsingErg table */
    for (i4=0;i4<numpara*CM(2,4,5);i4++) { buf[i4]=0; }
    /* get Ising energies per bond */
    for (i2=   0;i2<nn-1;i2++)    /* N*(N-1) site pairs */
    for (i3=i2+1;i3<nn  ;i3++) {
           i4=w2w[i2][i3];    /* access N*N pair array */
           if (i4<0 || i4>=numpara || i4>2+1) continue;
           ib1=geti(b1,i2);  /* 0=. 1=u=o 2=d=x 3=u+d=Q */
           ib2=geti(b1,i3);
           if (CM(2,4,5)*i4+map[4*ib1+ib2]>CM(2,4,5)*(numpara)) ERR(9999);
           buf[CM(2,4,5)*i4+map[4*ib1+ib2]]++;
    }
    j1=buf[CM(2,4, 5)]/2;
    if (numpara<3) j2=0; else
    j2=buf[CM(4,8,10)]/2; /* segfault if numpara<3 ? */
    if (j1>=Imax) j1=Imax-1; /* overflow */
    if (j2>=Imax) j2=Imax-1; /* overflow */
    if (j1>pp->d1b) pp->d1b=j1; /* max matrix idx, minimum output matrix */
    if (j2>pp->d2b) pp->d2b=j2;
    if (j1<pp->d1a) pp->d1a=j1; /* min matrix idx, minimum output matrix */
    if (j2<pp->d2a) pp->d2a=j2;
    pp->states[j2*Imax+j1]++;
    pp->weight[j2*Imax+j1]+=Norm(pp->v0[b_ofs[blk]+i1]); // *l2.norm2;
  }
  free(buf);
}

int out_ising_weight(tvec *v0) {
  long  tm1=itime();
  int j1, j2, blk;
  double sum;
  ULong i1;
  if (!mpi_i)
    printf("\nout_ising_weight: numpara= %d", numpara);
  MPI_Barrier(MPI_COMM_WORLD);
  if (mpi_n>1 && !mpi_i)
    printf("\nWarning: mpi_n>1 can produce wrong results, skip (ToDo)");
  if (mpi_n>1) return 0;

  for(blk=0; blk<pt_n; blk++) {
    ising_ptr[blk].blk=blk;
    ising_ptr[blk].v0=v0;
    parallel_start(blk, ising_weight_thread, &ising_ptr[blk]);
  }
  for(blk=0; blk<pt_n; blk++) {
    parallel_wait(blk);
  }
  if (!mpi_i)
    printf("\nout_ising_weight: parallel work done.");
  MPI_Barrier(MPI_COMM_WORLD);
  /* reduce results from threads */
  for(blk=1; blk<pt_n; blk++) {
    for (j1=0;j1<Imax*Imax;j1++) {
      ising_ptr[0].states[j1] += ising_ptr[blk].states[j1];
      ising_ptr[0].weight[j1] += ising_ptr[blk].weight[j1];
    }
    if (ising_ptr[0].d1b<ising_ptr[blk].d1b)
        ising_ptr[0].d1b=ising_ptr[blk].d1b; /* max matrix idx, min output */
    if (ising_ptr[0].d2b<ising_ptr[blk].d2b)
        ising_ptr[0].d2b=ising_ptr[blk].d2b;
    if (ising_ptr[0].d1a>ising_ptr[blk].d1a)
        ising_ptr[0].d1a=ising_ptr[blk].d1a; /* min matrix idx, min output */
    if (ising_ptr[0].d2a>ising_ptr[blk].d2a)
        ising_ptr[0].d2a=ising_ptr[blk].d2a;
  }
  if (!mpi_i)
    printf("\nout_ising_weight: threads summed.");
  MPI_Barrier(MPI_COMM_WORLD);

#ifdef CONFIG_MPI
  {
    ULong  tstates[Imax*Imax];
    double tweight[Imax*Imax];  /* weight matrix */
    int   td1a,td1b, td2a,td2b;     /* matrix dimension (maxIsingErg=max_ud_per_bond) */
    MPI_Allreduce(&ising_ptr[0].states, &tstates, Imax*Imax, MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
    MPI_Allreduce(&ising_ptr[0].weight, &tweight, Imax*Imax, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    for (j1=0;j1<Imax*Imax;j1++) {
      ising_ptr[0].states[j1]=tstates[j1];
      ising_ptr[0].weight[j1]=tweight[j1];
    }
    MPI_Allreduce(&ising_ptr[0].d1a, &td1a, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
    MPI_Allreduce(&ising_ptr[0].d2a, &td2a, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
    MPI_Allreduce(&ising_ptr[0].d1b, &td1b, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
    MPI_Allreduce(&ising_ptr[0].d2b, &td2b, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
    ising_ptr[0].d1a=td1a;
    ising_ptr[0].d2a=td2a;
    ising_ptr[0].d1b=td1b;
    ising_ptr[0].d2b=td2b;     
  }
  if (!mpi_i)
    printf("\nout_ising_weight: processes summed.");
  MPI_Barrier(MPI_COMM_WORLD);
#endif

  if (!mpi_i) { /* ToDo: mass center, mean deviation, momentums? */
    double center1=0, center2=0; /* mass center of Ising weight */
    double dev1=0, dev2=0,       /* deviation of Ising weight */
           cstate1=0, cstate2=0, /* center of states */
           r1, r2;
    for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      center1+=j1*ising_ptr[0].weight[j2*Imax+j1];
      center2+=j2*ising_ptr[0].weight[j2*Imax+j1];
      cstate1+=j1*ising_ptr[0].states[j2*Imax+j1];
      cstate2+=j2*ising_ptr[0].states[j2*Imax+j1];
    }
    for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      r1=(center1-j1)*ising_ptr[0].weight[j2*Imax+j1]; dev1+=r1*r1;
      r2=(center2-j2)*ising_ptr[0].weight[j2*Imax+j1]; dev2+=r2*r2;
    }
    printf("\nPRF: iw.t= %.2f min (Ising weight)"
           "\n entropy_center1,2= %8.4f %8.4f "
           "\n mass_center1,2=    %8.4f %8.4f  deviation1,2= %8.4f %8.4f"
           "\n weight matrix (%d-%d,%d-%d):",
            (itime()-tm1)/60., cstate1/n1, cstate2/n1,
             center1, center2, sqrt(dev1), sqrt(dev2),
             ising_ptr[0].d1a, ising_ptr[0].d1b,
             ising_ptr[0].d2a, ising_ptr[0].d2b);

    for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      if (j1==ising_ptr[0].d1a) printf("\n ");
      if (ising_ptr[0].states[j2*Imax+j1])
            printf(" %11.9f", ising_ptr[0].weight[j2*Imax+j1]);
      else  printf(" %11g",   ising_ptr[0].weight[j2*Imax+j1]);
    }
    printf("\n#");
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      sum=0;
      for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
        sum+= ising_ptr[0].weight[j2*Imax+j1];
      printf(" %11.9f", sum);
    }

    printf("\n states matrix (%d-%d,%d-%d):",
      ising_ptr[0].d1a, ising_ptr[0].d1b,
      ising_ptr[0].d2a, ising_ptr[0].d2b);
    for (j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      if (j1==ising_ptr[0].d1a) printf("\n ");
      printf(" %11lu", ising_ptr[0].states[j2*Imax+j1]);
    }
    printf("\n#");
    for (j1=ising_ptr[0].d1a; j1<=ising_ptr[0].d1b; j1++) {
      for (i1=0,j2=ising_ptr[0].d2a; j2<=ising_ptr[0].d2b; j2++)
        i1+= ising_ptr[0].states[j2*Imax+j1];
      printf(" %11lu", i1);
    }
  }

  return 0;
}

/*  ************ check of the eigenvector and eigenvalue ************ */
/* Inp: v0 (ToDo: use static if v0 != v1)
 * Out: v1 (used as work array) 
 *  print H*v0, rh=<v0*H*v0>, rq=norm(H*v0-h*v0) 
 *  print out expectation values as XX=... , use gawk to extract results
 */
int chk3(tvec *v0,tvec *v1,int iv,thbuf *hbuf[B_NUM]){
    int i1,i2,i3,j1,j2,j3,deg; long l1; double d1,d2,rr;
    static double sisj[NN][NN],zizj[NN][NN];
    deg=0; /* degenerate ? */

    for(i1=0;i1<nn;i1++)
    for(i2=0;i2<nn;i2++)zizj[i1][i2]=sisj[i1][i2]=0.0;

#if !defined(Dimer)	/* better: only for generators <T>!!! Nov98 */

/* write a general diff_pair_corr(fkt{-NR},array[][],v0)
 *  and call with sisj,zizj or fkt=1,2,3... */
    l1=itime(); /* measure zizj-time */
#if Korrelations
    /* 3*<ZiZj>=<SiSj> nur, wenn <SS>=0 (singlet), auch bei Entartung (?) */
    /* could be problematicaly od degenerate states! not proofed */
    for(i3=0;i3<sym.num_dcorr;i3++){
      int sites[2]; long l2=itime();
      sites[0]=i1=sym.dcorr[i3][0];
      sites[1]=i2=sym.dcorr[i3][1];
      // old version does non-diag-MPI communication (extremly slower)
      // new variant does no interprozess communication JS-2011-10-24 
      // zizj[i1][i2]= wop(&op_zizj,NULL,2,sites,v0,v0,hbuf); /* ZiZj */
      zizj[i1][i2]= w_diag_op(&op_zizj,NULL,2,sites,v0,v0,hbuf); /* ZiZj */
      for(j3=j1=0;j1<nn;j1++)for(j2=0;j2<nn;j2++)
      if(sym.corr[j1][j2]==i3){
        zizj[j1][j2]=zizj[i1][i2]; /* copy same bonds */
        j3++;              /* count num of same bonds */
      }
      if (!mpi_i) {
        printf("\n i-j %2d %2d ZiZj %9.6f",i1,i2,zizj[i1][i2]); /* kurz+knapp */
        printf("  num= %4d  r^2= %7.3f", j3, distance(i1,i2));
        /* ToDo: no or only once? if fast, easier for diff-check */
        if (n1>>16) printf(" ij.t= %5.1f m", (itime()-l2)/60.); /* for diff-check */
      }
    } /* for(i3=0;i3<sym.num_dcorr;i3++) */
    if (!mpi_i){
      printf("\n %dzij.t= %s", sym.num_dcorr, sclock(itime()-l1));

      for(i3=-1,i1=0;i1<nn;i1++)
      if(sym.corr[i1][i1]>i3){ i3=sym.corr[i1][i1];
        printf("\n i= %2d ZiZj=",i1);	/* 3ZiZj using sh/awk (simpler!) */
        for(i2=0;i2<nn;i2++) printf(" %9.6f",zizj[i1][i2]);
      }
      printf("\n Warning: SiSj=3ZiZj, ZiZj only correct for EA=1, sum over Sz too!");/* Jun99 */
    }

    for(i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++){
      sisj[i1][i2]=3*zizj[i1][i2]; }		/* 3*ZiZj isotropic */
    if (cfg_sisj&1)	/* set in daten.i */
    { 
      int sites[2]; long l2;
      l1=itime();
      if ((vvv&1) && !mpi_i) printf(" t+=%s",sclock(~0l));
      if (!mpi_i) printf("\n cfg_sisj&1 is set, sisj will be calculated ");
      /* could be problematicaly on degenerate states! not proofed */
      for(i3=0;i3<sym.num_dcorr;i3++){
        l2=itime();
        sites[0]=i1=sym.dcorr[i3][0];
        sites[1]=i2=sym.dcorr[i3][1];
        sisj[i1][i2]=wop(&op_sisj,NULL,2,sites,v0,v0,hbuf);  /* SiSj */
        if (!mpi_i)
          printf("\n i-j %2d %2d SiSj %9.6f",i1,i2,sisj[i1][i2]);
        for(j3=j1=0;j1<nn;j1++)for(j2=0;j2<nn;j2++)
        if(sym.corr[j1][j2]==i3){
          sisj[j1][j2]=sisj[i1][i2]; /* copy same bonds */
          j3++;              /* count num of same bonds */
        }
        d1=sisj[i1][i2]-3*zizj[i1][i2];
        if (!mpi_i) {
          printf("  Sij-3Zij= %9.6f %6.0e num= %4d", d1, d1, j3);
          if (n1>>16) printf(" ij.t= %5.1f m", (itime()-l2)/60.);
          /* print time is bad for diff-check */
        }
      } /* for(i3=0;i3<sym.num_dcorr;i3++) */
      if (!mpi_i) {
        printf("\n %dsij.t= %s", sym.num_dcorr, sclock(itime()-l1));
        for(i3=-1,i1=0;i1<nn;i1++)
        if(sym.corr[i1][i1]>i3){ i3=sym.corr[i1][i1];
          printf("\n i= %2d SiSj=",i1);				/* SiSj */
          for(i2=0;i2<nn;i2++) printf(" %9.6f",sisj[i1][i2]); }
      }
    }
    if ((!(cfg_sisj&1)) && !mpi_i)
      printf("\n cfg_sisj&1 is not set, sisj will not be calculated ");

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++) d1+=sisj[i1][i2]; 
    d2=fabs(d1); d2=fabs(-1.0+sqrt(1+4*d2));    /* 2S */
    if (!mpi_i) {
      printf("\n SS= %5.3f 2S= %5.3f % 5.0e", d1, d2, d2-(int)(d2+.5));
      if (cfg_sisj&1) printf(" (as sum sisj)");
      else            printf(" (as sum 3zizj)");
    }

    if ((vvv&1) && !mpi_i) printf(" t+=%s",sclock(~0l));
    for(i3=-1,i1=0;i1<nn;i1++) if(sym.corr[i1][i1]>i3){	/* oct98 */
      if (!mpi_i) printf("\n i= %2d Zi=",i1);		/* Zi */
      i3=sym.corr[i1][i1];
      // d1=wop(&op_zi,NULL,1,&i1,v0,v0,hbuf); /* Zi 2017-01-27-- */
      d1=w_diag_op(&op_zi,NULL,1,&i1,v0,v0,hbuf); /* Zi 2017-01-28++ */
      if (!mpi_i)
        printf(" %9.6f Zi-(nu-nd)/(2N)= %9.6f",d1,d1-(nu-nd)/(2.*nn));
    }
    if ((vvv&1) && !mpi_i) printf(" t+=%s",sclock(~0l));
    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++) d1+=fabs(zizj[i1][i2]); 
    d2= d1= fabs(d1/(nn*nn)); /* ZMag=1/N^2 Sum |ZiZj| */
    if (!mpi_i) printf("\n ZMag= %10.8f",d1);

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++)
      d1+=fabs(sisj[i1][i2]-zizj[i1][i2]); 
    d1=fabs(d1/(nn*nn));
    if (!mpi_i) printf(  " XYMag= %10.8f",d1);		/* XYMag=1/N^2 Sum |XYiXYj| */

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++) /* SMag=1/N^2 Sum |SiSj| */
      d1+=fabs(sisj[i1][i2]);
    d1=fabs(d1/(nn*nn));
    if (!mpi_i) printf(  " SMag= %10.8f SMag-3ZMag= % 10.8f",d1,d1-3*d2);

    /* this type of magnetization is for test if SMag(E) in same order? */
    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++) d1+=zizj[i1][i2]*zizj[i1][i2]; 
    d2= d1= sqrt(d1/(nn*nn));
    if (!mpi_i) printf("\n Z2Mag= %10.8f",d1); /* Z2Mag=(1/N^2 Sum ZiZj^2)^.5 */

    for(d1=i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++) d1+=sisj[i1][i2]*sisj[i1][i2]; 
    d1=sqrt(d1/(nn*nn));
    if (!mpi_i) printf(  " S2Mag= %10.8f ",d1);  /* S2Mag=(1/N^2 Sum SiSj^2)^.5 */

#endif /* Korrelations */

#if 1 /* check again energy and its error: 0=yes 1=no */
    if (!mpi_i)
    printf("\nChange L%d in spins.c to"
           " check the energy again (slow!)\n",__LINE__-8);
#else /* nonparallel and slow version, makes no sence */
    r=wh(v0);	 if (!mpi_i) printf(" E%d/N= %12.8f",iv,D(r)/nn);
    		 if (!mpi_i) printf(  " E%d= %12.8f",iv,D(r));
    r=whq(v0,r); if (!mpi_i) printf(" %5.0e" ,D(r));   /* (H*v-<vHv>*v)^2 */
#endif
    /* if((fabs(r)>1.e-5) && (v1!=NULL)) return deg; */ 

    if(sym.numsym<60 || (double)n1<1e6) /* takes time, change ??? */
    /* wss is slow, if we have ss from sisj we dont need it again */
    if ((!mpi_i) && !(cfg_sisj&2))
      printf("\n cfg_sisj&2 is not set, 2S will not be calculated ");
    if(cfg_sisj&2) {
      if (!mpi_i) printf("\n cfg_sisj&2 is set, 2S will be calculated,");
      l1=itime();
      rr = wss(v0,v0,hbuf);
      if (!mpi_i) printf(" 2S= %5.3f", rr);		/* 2S */
      if (!mpi_i) printf(" 2S.t=%s", sclock(itime()-l1));
    }
#if  CM(0,0,1)
    rr = D( wjj(v0,v0,hbuf) );
    if (!mpi_i) printf(" 2J= %5.3f", rr);		/* 2J */
#endif

    MPI_Barrier(MPI_COMM_WORLD);
#if defined(SITE_DIST)
    /* --- staggered Magnetization --- */
    for(rr=i2=0;i2<nn;i2++)rr+=((site_dist(0,i2)&1)?-1:1)*sisj[0][i2];
    if (!mpi_i) printf(" StaggMag= %9.6f", rr/nn);
#endif
# ifdef Spin1chain
    stringorder2(stdout,v0);
# endif    
#else /* defined(Dimer) */
				/* SiSj */
    stringorder(stdout,v0);
    for(i1=0;i1<nn;i1++)sisj[0][i1]=3*zizj[0][i1];/* sowieso schon Dez99 */
    for(i1=0;i1<nn;i1++)sisj[2][i1]=3*zizj[2][i1];
    /* ##### C_x=Sum_(i!=j,(i,j aus X)){|<SiSj>|} ##### */
    d1=.0;							/* C_tot */
    for(i1=0;i1<nn;i1++)if(i1!=0)  d1+=fabs(sisj[0][i1]);
    for(i1=0;i1<nn;i1++)if(i1!=2)  d1+=fabs(sisj[2][i1]);
    if (!mpi_i) printf(" C_tot= %8.6f",(double)d1/(2*(nn-1)));
    d1=.0;							/* C_d */
    for(i1=0;i1<nn;i1++)if(i1%4<2)if(i1!=0) d1+=fabs(sisj[0][i1]);	
    if (!mpi_i) printf(  "C_d= %8.6f",(double)d1/(nn/2-1));
    d1=.0;							/* C_p */
    for(i1=0;i1<nn;i1++)if(i1%4>1)if(i1!=2) d1+=fabs(sisj[2][i1]);	
    if (!mpi_i) printf(  "C_p= %8.6f",(double)d1/(nn/2-1));
    d1=.0;							/* C_pd */
    for(i1=0;i1<nn;i1++)if(i1%4<2)if(i1!=2) d1+=fabs(sisj[2][i1]);	
    if (!mpi_i) printf(  "C_pd %8.6f",(double)d1/(nn/2));
#endif /* Dimer */
    /* samples for 'exotic' corelation terms: (please double check results) */
#if 0 /* 1: compute site-rotation-terms, 0: disabled (not computed) */
    void op_rot4(void *para,int ns,int *iw,thbuf *hbuf); /* xval.c */
    /* example for H. Mosadegh, Sep09: Heisenberg model only! untested */
    if (nn>4) {
      int iw[4]={0,1,2,3};
      rr = wop(&op_rot4,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s_rot(0,1,2,3)> i= 0123 % 9.6f", rr);
    }
    if (nn>6) {
      int iw[6]={0,1,2,3,4,5};
      rr = wop(&op_rot4,NULL,6,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s_rot(0,1,2-5)> i= 012345 % 9.6f", rr);
    }
#endif
#if 0 /* 1: compute 4-site-terms, 0: disabled (not computed) */
    void op_diff_sisj_sksl(void *para,int ns,int *iw,thbuf *hbuf); /* xval.c */
    void op_mult_sisj_sksl(void *para,int ns,int *iw,thbuf *hbuf); /* xval.c */
    /* example for F. Ouchni, makes sence for tU-chains N>4 */
    if (nn>4) {
      int iw[4]={2,4,0,2};
      rr = wop(&op_diff_sisj_sksl,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s(i)s(i+2)-s(i-2)s(i)> i= 2 % 9.6f", rr);
    }
    /* example for J. Richter, makes sence for J-chains N>=4 */
    if (nn> 4) {
      int iw[4]={2,4,0,2};
      rr = wop(&op_mult_sisj_sksl,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s(i)s(i+2)*s(i-2)s(i)> i= 2 % 9.6f", rr);
    }
    if (nn>=4) {
      int iw[4]={0,1,2,3};
      rr = wop(&op_mult_sisj_sksl,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s(0)s(1)*s(2)s(3)> % 9.6f", rr);
    }
    if (nn>=4) {
      int iw[4]={0,1,0,1};
      rr = wop(&op_mult_sisj_sksl,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s(0)s(1)*s(0)s(1)> % 9.6f", rr);
    }
    if (nn>=4) {
      int iw[4]={0,1,1,2};
      rr = wop(&op_mult_sisj_sksl,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s(0)s(1)*s(1)s(2)> % 9.6f", rr);
    }
    if (nn>=4) {
      int iw[4]={1,2,0,1};
      rr = wop(&op_mult_sisj_sksl,NULL,4,iw,v0,v0,hbuf);
      if (!mpi_i) printf("\n <s(1)s(2)*s(0)s(1)> % 9.6f", rr);
    }
#endif
#if  CM(0,1,1)
    if (!mpi_i) printf("\n niu.=");  /* occupation value <v0|n_s|v1> */
    for(i1=0;i1<nn;i1++) {
      rr = wop(&op_nis,(void *)"1",1,&i1,v0,v0,hbuf);
      if (!mpi_i) printf(" %8.6f", rr); 			/* niu */
    }
    if (!mpi_i) printf("\n niud=");  /* double occupation value (nu*nd)(i) */
    for(i1=0;i1<nn;i1++) {
      rr = wop(&op_nis,(void *)"3",1,&i1,v0,v0,hbuf);
      if (!mpi_i) printf(" %8.6f", rr); /* niud */
    }
    if ((!mpi_i) && !(cfg_sisj&1))
      printf("\n cfg_sisj&1 is not set, slow cicj will not be calculated ");
    for(i3=0;i3<sym.num_dcorr;i3++){
      double uu, dd, ud, du;
      int iw[2];
      iw[0]=i1=sym.dcorr[i3][0];
      iw[1]=i2=sym.dcorr[i3][1];
      rr =  wop(&op_ninj,NULL,2,iw,v0,v0,hbuf); /* ninj */
      if (!mpi_i) printf("\n i-j %2d %2d nn % 9.6f", i1, i2, rr); /* ninj */
      /* for (sym_ud) nuij is wrong, better take (nuij+ndij)/2 */
      uu=wop(&op_nisnjs,(void *)"1111",2,iw,v0,v0,hbuf);
      dd=wop(&op_nisnjs,(void *)"2222",2,iw,v0,v0,hbuf);
      ud=wop(&op_nisnjs,(void *)"1122",2,iw,v0,v0,hbuf);
      du=wop(&op_nisnjs,(void *)"2211",2,iw,v0,v0,hbuf);
      rr=wop(&op_nisnjs,(void *)"3030",2,iw,v0,v0,hbuf);
      if (!mpi_i) printf(  " uu % 9.6f",uu);
      if (!mpi_i) printf(  " dd % 9.6f",dd);
      if (!mpi_i) printf(  " ud % 9.6f",ud);
      if (!mpi_i) printf(  " du % 9.6f",du);
      if (!mpi_i) printf(  " hh % 9.6f",rr);
      /* we need symmetrized operators if we use spin-inversion */
      if (nu==nd && sym.wud) if (!mpi_i) printf(  " uudd % 9.6f", 0.5*(uu+dd));
      if (nu==nd && sym.wud) if (!mpi_i) printf(  " uddu % 9.6f", 0.5*(ud+du));
      /* hopping between site iw[0] and iw[1]    *
       *    sum_s c(+,i,s)c(-,j,s)+h.c.          */
      if (cfg_sisj&1) { /* slow */
        rr = wop(&op_cicj,NULL,2,iw,v0,v0,hbuf);
        if (!mpi_i) printf(  " cc % 9.6f", rr);
      }
    }
    if (nu==nd && sym.wud && !mpi_i)
      printf("\nWarning: use uudd and uddu instead of uu, dd, ud and du");
#if 0 && CM(0,1,1)
/*  Definition by F. Ouchni (do you know about citations?)
 *  "bond-order wave" (BOW) (product of four fermionic operators)
 *  reference:PRB vol_67 page_205109
 *     BOW(r) = (-1)^r ( (\sum_{l} <B_{l,l+1} B_{l+r,l+r+1}>/N) - B^2)
 *          B = \sum_{l} <B_{l,l+1}>/N
 *  B_{l,l+1} = \sum_{spin} (C^{\dag}_{l s} C_{l+1 s} + h.c)
 *  sample: 
 *  correct? only if using symmetry? see m_tU.c
 *   m_tU 10 >daten.def  # v.0.22
 *      param=10,0,1,0,0,1,0.5,0.3333333,0.25,0.1 nud=2,2
 *   E0= 0.4921456 (ok)
 *        sum_k= -2 0 -1000 : <B(l,l+1)> l= 0  -0.642523 (see m_tU.c)
 *               <B(l,l+1) B(l+r,l+r+1)> r= 1   0.299550 OK
 *               BOW(r)                  r= 1   0.113287 OK (F.Ouchni)
 *               
 */
    /* only for chains */
    {
      int iw[4]={0,1,1,2}, r=1;
      double b2=0, b4;
      if (nn>1) printf("\n <B(l,l+1)>              l= 0 % 9.6f",
        b2= wop(&op_cicj,NULL,2,iw,v0,v0,hbuf) );
      if (nn>2)
      for (r=1; r<=nn/2; r++) {
        iw[2]=iw[0]+r;
        iw[3]=iw[1]+r;
        b4= wop(&op_cicjckcl,NULL,4,iw,v0,v0,hbuf);
        if (mpi_i) continue;
        printf("\n <B(l,l+1) B(l+r,l+r+1)> r= %d % 9.6f",r,b4);
        printf("\n BOW(r)                  r= %d % 9.6f",r,(1-((r&1)<<1))*(b4-b2*b2));
      }
    }
#endif

#else /* now only JJ */
    if ((!mpi_i) && !(cfg_sisj&4))
      printf("\n cfg_sisj&4 is not set, mrule will not be calculated ");
    if (  cfg_sisj&4 ) {
      if (!mpi_i) printf("\n cfg_sisj&4 is set, mrule will be calculated ");
      for (i1=0;i1<sym.subl;i1++){
        for(i3=i2=0;i2<nn;i2++) if (sym.sl[i1][i2]+1>i3) i3=sym.sl[i1][i2]+1;
        mrule(v0,sym.sl[i1],i3); /* Marshall+output */
      }
    }
#endif /* JJ */

    MPI_Barrier(MPI_COMM_WORLD);
    /* calculate structure factor S(q)/N^2 v2.27 */
    for (i1=0;i1<sym.subl;i1++){
      int *sl, nsl;
      sl=sym.sl[i1];
      /* get the maximum SL index */
      for(nsl=i2=0;i2<nn;i2++) if (sl[i2]+1>nsl) nsl=sl[i2]+1;
      d1=0;
      if (nsl==2) {
        /* sublattices A,B => S(q=(pi,pi)), A-A: cos=1, A-B: cos=-1 */
        for (i2=0;i2<nn;i2++) for (i3=0;i3<nn;i3++)
          d1+=sisj[i2][i3]*((sl[i2]==sl[i3])?1:-1);
        if (!mpi_i) printf("\n SL%d S(pi)/N^2= % .8f",i1,d1/(nn*nn));
      }
      if (nsl==3) {
        /* sublattices A,B,C => S(q=(4pi/3,0)), A-A: cos=1, A-B: cos=-0.5 */
        for (i2=0;i2<nn;i2++) for (i3=0;i3<nn;i3++)
          d1+=sisj[i2][i3]*((sl[i2]==sl[i3])?1:-0.5);
        if (!mpi_i) printf("\n SL%d S(q3)/N^2= % .8f",i1,d1/(nn*nn));
      }
      if ((!mpi_i) && 2<=nsl && nsl<=3) {
        printf(" SL= ");
        for (i2=0;i2<nn;i2++) printf("%c",sl[i2]+'A');
      }
    }
    MPI_Barrier(MPI_COMM_WORLD); /* sync output */

#if CM(1,0,0)
#if defined(N1) && defined(N2)
#if N1>1 /* ToDo: adapt to symmetries */
    printf("spins.c L%d: old version, please adapt to SBase\n", __LINE__);
   /* local sublattice magnetization Msi^2=<1/N*sum_{j=1}^N{ti*tj*Siz*Sjz}>
    * sublattice magnetization Ms^2=1/N*sum_{i=1}^N{Msi^2}
    * ti=-1 for i odd (A), ti=+1 for i=even (B)      */
    printf(" Z= %8.5f",    D(wz(v0)));
    wtatb(&r1,&r2,v0);
    printf(" TA= %8.5f",   D(r1));
    printf(" TB= %8.5f",   D(r2));
    printf(" Dx= %8.5f",   D(wdx(v0)));
    printf(" Dy= %8.5f",   D(wdy(v0)));
    printf( " C= %8.5f",    D(wc(v0)));
#endif /* N1>1 */
#endif /* defined betts */
#endif

    if (vvv&1) out_ising_weight(v0);
    MPI_Barrier(MPI_COMM_WORLD);

    /* output up to x_max x coefficients and its basecfg */
    if (vvv)             x_out(stdout,v0,0,x_p); /* max first */
    if (vvv && x_max<n1) x_out(stdout,v0,1,x_p); /* min first */
    MPI_Barrier(MPI_COMM_WORLD);
    if ((vvv&1) && !(n1>>20)) x_sort(stdout,v0,v1); /* slow v0=input, v1=work */
return 0; /* ToDo2016 ??? */
    if ((vvv&1) && !mpi_i) printf(" t+=%s",sclock(~0l));
    MPI_Barrier(MPI_COMM_WORLD);
    return(deg);
}
/* ====================== end of chk3 ======================= */    
    
    
/* sorting indizes using shellsort, R.Sedgewick: "Algorithmen in C" */
/* input: a,n;  output: ia; Nov98 JS +error in 1st Auflage */
/* Mar99-JS: ULong n statt int */
void shellsort(const double a[],int ia[],const ULong n){
  int v; ULong i,j,h;
  if(vvv) printf(" shellsort");
  for(i=0;i<n;i++) ia[i]=(int)i;	/* set indizes, (int) ??? Mar99 */
  for(h=1;h<=n/9;h=3*h+1);		/* using int or ULong ??? confusing */
  for(;h>0;h/=3)
   for(i=h-1;i<n;i++){
     v=ia[i]; j=i;
     while(j>=h && a[ia[j-h]]>a[v]){
       ia[j]=ia[j-h]; j-=h; }
     ia[j]=v;
   }
}

/* ================== USING LAPACK (linear algebra package) ================
  ToDo2016: runtime switch between cplx/real code (like lapack)
   also OpenBLAS: Basic Linear Algebra  (see utils/lapack_test.c n1^2-OVL32)
 DSYEV - eigenvalues and eigenvectors of a real symmetric matrix A (LAPACK)
 ZHEEV - eigenvalues and eigenvectors of a complex Hermitian matrix A
 Makefile: add DFLAGS= -DLAPACK and system specific options LIB=...
 Please test BLAS/LAPACK before using spinpack, see utils/lapack_test.c
   grep zheevd /usr/lib64/lib* # scalapack,lapack,openblas,mkl,acml
 time[s] = 4e-8 * n1^3 (wf: GS160-alpha-731MHz dsyev nev=1 double ) 
 time[s] = 8e-8 * n1^3 (wf: GS160-alpha-731MHz zheev nev=1 complex)
 example: n1=4000 4s-5min, use N(sz=0) slightly above n1_max, reduce n1 by Sz++
   LC40_5+35  k=-2 1 n1=16450
   LC40_4+36  k=-2 1 n1= 2280 nev=1
   LC18_9+9   k=-2 1 n1= 2700 nev=0 oblas0.2.14.zheev   16comp2=0.055m -14.53171376
   LC20_7+13  k=-2 1 n1= 3876 nev=0 oblas0.2.14.zheev   16comp2=0.171m -12.16910877 +9.90211303
     lapack_test -Z      3876       oblas0.2.14.zheev   16comp2=0.172m
   LC20_7+13  k=-2 1 n1= 3876 nev=1 oblas0.2.14.zheevdV 16comp2=0.327m -12.16910877 +9.90211303
     lapack_test -Z -a-1 3876 -d -V oblas0.2.14.zheevdV 16comp2=0.330m 2016-05
     lapack_test -Z      3876 -d -V oblas0.2.14.zheevdV 16comp2=1.047m
   LC20_7+13  k=-2 0 n1= 3876 nev=0 oblas0.2.14.dhyev   16comp2=0.070m -11.72911456 +10.0000000
     lapack_test         3876       oblas0.2.14.dhyev   16comp2=0.072m 2016-05
   LC20_7+13  k=-2 0 n1= 3876 nev=1 oblas0.2.14.dhyevdV 16comp2=0.146m -11.72911456 +10.0000000
     lapack_test -a-1    3876 -d -V oblas0.2.14.dhyevdV 16comp2=0.161m 2016-05
     lapack_test         3876 -d -V oblas0.2.14.dhyevdV 16comp2=0.160m
   LC20_8+12  k=-2 1 n1= 6288
   LC20_9+11  k=-2 1 n1= 8398 nev=1 oblas0.2.14.zheev.16comp2=1.652m
             lapack_test 8398 -Z -d -V 
   LC20_10+10 k=-2 1 n1= 9225 oblas0.2.14.zheev.16comp2
   LC24_12+12 k=-2 1 n1=112632 malloc_xm failed
   4x4-square-cplx sym=q_0_0_0+ n1=365 e0=-11.22848321
   jacobi=21s  lapack=1s  (alpha731MHz, real)   431e-9s*n1^3
   jacobi=     lapack=1s  (alpha731MHz, complex)
   i686-2.7GHz n1=1789 jacobi nev=1 14.1m 53MB 4x4_q00  147e-9s*n1^3 19It,rots=10.7*n^2
   i686-2.7GHz n1=1789 jacobi nev=0 5.38m 28MB 4x4_q00   55e-9s*n1^3 19It,rots=10.7*n^2
   i686-2.7GHz n1=914  jacobi nev=1 1.52m 17MB 4x4_q000 120e-9s*n1^3 17It,rots=9.85*n^2
   i686-2.7GHz n1=914  jacobi nev=0 0.53m 11MB 4x4_q000  40e-9s*n1^3 17It,rots=9.85*n^2
   LC40j2nd4   n1=1240   jaco 640MB-flt  omp1= 98%(sy=2%)  1.2m/5.7e6rots=0.17m/it(7e5rots=.49n**2)
   LC40j2nd4   n1=1240   jaco 640MB-flt  omp2=124%(sy=8%)  5.2m/5.7e6rots=0.7m/it(7e5rots=.49n**2)
   LC40j2nd4   n1=1240   jaco 640MB-flt  omp8=140%(sy=14%)    /5.7e6rots=3.7m/it(7e5rots=.49n**2)
   LC40j2nd4   n1=1240  cjaco 640MB-cflt omp8=150%(sy=14%) 25m/5.7e6rots=3.2m/it(7e5rots=.49n**2)
   LC40j2nd5   n1=8311  cjaco 640MB-cflt omp8=190%(sy=13%)
   LC40j2nd6   n1=48564 cjaco 19GB-cflt it=1:1.82m omp8=328%(sy=10%)
   LC32j2nd7   n1=52819 cjaco 11GB-flt
   LC24j2z0    n1=28968 cjaco 3.3GB-flt 12Zregions 41m++/It nrots=2e6/9m
           skip 0,0-rots: 2e6/1.97m
            pcjago nrots=16k/0.82m(omp1) 16k/1.1m(omp2) 16k/1.4m(omp8)
   LC18j2z0    n1=765   cjaco= 35s nrots=2701576 -15.17070224
                       pcjaco=100s nrots=1748922 -15.17070224 omp2=105s
                       localized?  nrots=1878070 -15.17070224 9m
   ToDo: pthread?
   i686-2.6GHz n1=1617 zheev nev=1 139s  44MB # lapack-3.0 SuSE-9.2
   i686-2.6GHz n1=1617 zheev nev=0  24s  44MB
   i686-2.6GHz n1=1617 dsyev nev=1  44s  23MB # lapack-3.0 SuSE-9.2
   i686-2.6GHz n1=1617 dsyev nev=0   8s  23MB
   i686-2.6GHz n1=3583 dsyev nev=0  85s 103MB
   i686-2.6GHz n1=4704 dsyev nev=0 189s 176MB v2.22
   AOpt-2.2GHz n1=5481 dsyev nev=0 281s 240MB v2.23 4m..5m14s (259M/231M) acml=5m11s
                 10584 dsyev nev=0  ??  880MB v2.23 (vsz reduced)
                 16852 zheev nev=0 314m       v2.24 loki-Opteron-2.2GHz
                 18018 zheev nev=0 13h=48600s v2.22 8e-9s*n1^3 marvel
                 25000 zheevd     1e5s (ca.)
  max_n1=(2^32/cplx?)**(1/2)=2^16=65536
# LC18 J=2,0.125 q=-2,0,0 n1=622 of 48620 76s   0.27m (2011-12-11)
# eigenvals= -15.17070224 -13.22659499 -11.74608234 -11.56012456 -10.96556301
# q=-2,1,x n1=1336 18min jacobi2
# eigenvals= -14.19170172 -12.39204849 -11.79140730 -10.96800408 -10.95840566
# cjacobi: 3.25m (factor 5 faster than [[Re,-Im],[Im,Re]]-trick)
# eigenvals= -14.19170172 -12.39204849 -11.79140730 -10.96800408 -10.95840566
# real: 0.25m (cplx=0.57m ompHT=0.43m)
# eigenvals= -15.17070224 -13.22659499 -11.74608234 -11.56012456 -10.96556301
 !!! Jacobi is better in parallel than QR-algorithm
 * * * WARNING * * *
 SuSE-9.2 + lapack-3.0-920 = OK (n1>5400 wrong Es, spur ok)
  valgrind: Warning: set address range perms: large range 240330888, a 0, v 1
   5481^2*8 = 240330888

 BLAS/LAPACK is a Fortran-Library. This C interface may possibly not work.
 ToDo: extract DSYEV and ZHEEV and convert to C routines
       would allow to use long (8byte) instead of int (4byte) for bufsize
    DSYEV (..,..,A=RealMatrix,LDA,...W=Eigenvalues)
      DSYTRD = tridiag (real sym) = DLATRD+DSYR2K + DSYTD2 (slow=dlarfg+dsymv+daxpy+dsyr2)
        dgemv+dlarfg+dsymv+dscal+daxpy
        dgemv=a*A*x+b*y or a*A^T*x+b*y
      DSTERF = EW or DORGTR+DSTEQR = EW+EV
    ZHEEV (..,..,A=CplxMatrix,LDA,...W=Eigenvalues)
      ZHETRD = tridiag (real sym)
        T = Q**H * A * Q , opt. blocked loops: NB= 32 (ideal), 1 (worst)
        nb-blocked code: ZLATRD + ZHER2K + rest unblocked code ZHETD2
      DSTERF(N,*D,*E,INFO) = EW or ZUNGTR + ZSTEQR = EW+EV
*  if UPLO = 'U':              
*    (  d   e   v2  v3  v4 )   
*    (      d   e   v3  v4 )   
*    (          d   e   v4 )   
*    (              d   e  )   
*    (                  d  )   
*/
/* PMMAAA=dsyev_   see https://en.wikipedia.org/wiki/LAPACK
 * P:      d=double=64bit,s=single=32bit,c=cplx=2x32bit,z=cplxd=2x64bit 
 *  MM:    sy=symmetric, he=cplx_hermitian, ge=unsymNxN
 *    AAA: ev_=simple evx=expert evd=divide-and-conquere(size+N**2) evr=MRRR */
#ifdef LAPACK
#ifdef __cplusplus
extern "C"    /* C++ shall not create special object names */
#endif
void dsyev_( /* reel symmetric eigenvalues + optional(slow) eigenvectors */
        const char *jobz,   /* (in) "N"=NoVectors, "V"=Vectors */
        const char *uplo,   /* (in) "U"=UpperTriangleA, "L"=LowerTA */
        const int *n,       /* (in) order of matrix */
        double *a,          /* a[n][lda] (in/out) */  
        const int *lda,     /* (inp) leading dimension >=max(1,n) */
        double *w,          /* w[n] (out) eigenvalues */
        double *work,       /* work[lwork] (workspace/out) */
        const int *lwork,   /* (in)   >= max(1,3*n-1) better (nb+2)*n */
        int *info);         /* (out) ==0 on success */
#ifdef __cplusplus
extern "C"    /* C++ shall not create special object names */
#endif
void dsyevd_(   /* 2016-04-26 +EV about 4-5 times faster, lapack_test.c */
        const char *jobz,   /* (in) "N"=NoVectors, "V"=Vectors */
        const char *uplo,   /* (in) "U"=UpperTriangleA, "L"=LowerTA */
        const int *n,       /* (in) order of matrix */
        double *a,          /* a[n][lda] (in/out) */
        const int *lda,     /* (inp) leading dimension >=max(1,n) */
        double *w,          /* w[n] (out) eigenvalues */
        double *work,       /* work[lwork] (workspace/out) */
        const int *lwork,   /* (in) dsyev=3*n-1, min=1
                                    dsyevd=1+6n+2n^2 for V, 2n+1 for N */
        int *iwork,         /* array[liwork], [1] returns optimal LIWORK */
        const int *liwork,  /* size = 3+5n */
        int *info);         /* (out) ==0 on success */
#ifdef __cplusplus
extern "C"    /* C++ shall not create special object names */
#endif
void zheev_( /* complex hermitian */
        const char *jobz,   /* (in) "N"=NoVectors, "V"=Vectors */
        const char *uplo,   /* (in) "U"=UpperTriangleA, "L"=LowerTA */
        const int *n,       /* (in) order of matrix */
        double *a,          /* a[n][lda][2] (in/out), diagonal part I=0 */
        const int *lda,     /* (inp) leading dimension >=max(1,n) */
        double *w,          /* w[n] (out) eigenvalues */
        double *work,       /* work[lwork][2] (workspace/out) */
        const int *lwork,   /* (in)   >= max(1,2*n-1) better (nb+2) */
        double *rwork,      /* rwork[max(1,3N-2)] workspace 0..? 1..? */
        int *info);         /* (out) ==0 on success */
#ifdef __cplusplus
extern "C"    /* C++ shall not create special object names */
#endif
void zheevd_(
        const char *jobz,   /* (in) "N"=NoVectors, "V"=Vectors */
        const char *uplo,   /* (in) "U"=UpperTriangleA, "L"=LowerTA */
        const int *n,       /* (in) order of matrix */
        double *a,          /* a[n][lda][2] (in/out), diagonal part I=0 */
        const int *lda,     /* (inp) leading dimension >=max(1,n) */
        double *w,          /* w[n] (out) eigenvalues */
        double *work,       /* work[lwork][2] (workspace/out) */
        const int *lwork,   /* (in)  zheev=2*n-1, zheevd=2n+n^2  */
        double *rwork,      /* rwork[max(1,3N-2)] workspace */
        const int *lrwork,  /* rwork-dim V: 1+5n+2n^2 */
        int *iwork,         /* array[liwork], [1] returns optimal LIWORK */
        const int *liwork,  /* size = 3+5n */
        int *info);         /* (out) ==0 on success */
/* nb = blocksize for ZHETRD returned by ILAENV = 32 */
#endif  /* ifdef LAPACK */

/* ??? use hbuf without ev! a4
 * for complex (hermitian) matrizes of type: A+iB
 *   simply diagonalyze [[ A,-B], [ B, A]], but all eigenvals twice.
 *   EVs: (u,v) -> u+iv
 * old driver for routine jacobi() for symmetric reel matrizes
 * jacobi sparc-Ultra1-143MHz(boson): time/s = 1.5e-6 dim ^ 3.2
 * ToDo: output matrix_asc.pgm for vizualization (or using gnuplot?)
 * --------------------------  Matrizen ----------------------------------- */
/* older real version replaced by complex version 2011-12-11 */
/* declaration for cjacobi from diag.c, ToDo: better sdbl[2] than scplx! */
void  cjacobi(const unsigned long n, scplx *A, double *d, scplx *U);
//void pcjacobi(const unsigned long n, scplx *A, double *d, scplx *U);

// ToDo: input global ULong n1 as argument
// 2016-11-11 add ew as argument replacing global variable ew + return fail
//      rc=0 OK, else fail         
int fulldiag(tvec *ev[], int nev, thbuf *hbuf, double *ew){ 
  // single thread hbuf (2016: means we need to be single thread here?)
  int    rc=0, *ix=NULL, blk, ii; ULong k,i,j,iv=1;
  double *xd=NULL, xr,xi,trace1,trace2,h12,i12,max_abs_H=0.0; 
  double tm1, tm2; /* compute time */
#if !defined(LAPACK)  /* ToDo16: use same double[2] as cplx emulation! */
  scplx  *xm=NULL, *xv=NULL;  /* memory O(n1**2), ToDo: dbl[iv] */
#else
  double *xm=NULL, *xv=NULL;
  int lwork; /* LAPACK */
  int liwork=-1;  /* array[liwork], [1] returns optimal LIWORK */
  int *iwork=NULL;     /* [3+5*n] */
  int lrwork=-1;
  double *rwork=NULL;  /* defined more locally */
  int aaa= (nev) ? 1 : 0; /* fastest algo aaa: 0=ev 1=evd (lapack_test.c) */
  char cev = (nev) ? 'V' : 'N'; /* optionally eigenvectors or not */
#endif
  // thbuf *hbuf;
  minfo_t m_xm = {NULL,0,0,"tmp/xm.map",NULL}; /* input matrix */
  minfo_t m_xv = {NULL,0,0,"tmp/xv.map",NULL}; /* output/work matrix */
  minfo_t m_xd = {NULL,0,0,"tmp/xd.map",NULL}; /* energies */
  minfo_t m_ix = {NULL,0,0,"tmp/ix.map",NULL}; /* sort index */
  xi=0;
  if (!mpi_i) printf("\n fulldiag cplx=%d nev=%d n1=%lu", (VecType&4),nev,n1);
  if (mpi_n>1 && !mpi_i) E0("fulldiag dont work with MPI (ToDo: ScaLapack)");
  if (mpi_n>1) return -1; // ToDo: replace by parallel code later
#ifdef LAPACK
  if (!mpi_i) printf(" using BLAS/LAPACK ...\n");
#else
  if (!mpi_i) printf(" using Jacobi-method ...\n");
#endif
  /* standart jacobi and LAPACK expects simple vectors, convert block form */
  /* add 1 for the case n1=1, where Lapack fails */
  trace1=trace2=0; h12=i12=0;
#define S_double sizeof(double)
  /* int overflow ? */
  // xv needed for LAPACK, xm+xv needed for Jacobi
#ifndef LAPACK
  if (!mpi_i) {
    xm=  (scplx *)mymalloc(n1*n1*sizeof(scplx),&m_xm); /* Matrix_H-in =  */
    if (!xm) E1("malloc xm    failed, size= %g",n1*n1*1.*sizeof(scplx));
    if (nev) {
      xv=  (scplx *)mymalloc(n1*n1*sizeof(scplx),&m_xv); /* eigenvects-out */
      if (!xv) E1("malloc xv failed, n1^2*scplx= %g",n1*1.*n1*sizeof(scplx));
    }
  }
#else /* LAPACK */
  iv=((VecType&4)>>2)+1; /* 1=real, 2=emulate_{r,i} by r[2] */
  // for LAPACK we alloc xm(work/out-array) later
  if (!mpi_i) {
    xm=(double *)mymalloc(n1*n1*iv*S_double,&m_xm); /* eigenvects-out */
    if ((!xm)) E1("malloc xm    failed, size= %g",n1*iv*n1*1.*S_double);
  }
#endif
  if (!mpi_i) {
    xd=(double *)mymalloc(      iv*n1*S_double,&m_xd); /* eigenvalues-out */
    ix=   (int *)mymalloc(         n1*S_int,&m_ix);    /* sort-idx-out */
    if ((!xd) || (!ix)) F0("malloc xd+ix failed");
    for(i=0;i<n1;i++) ix[i]=(int)i; /* initialize sort-index (LAPACK) */
  }
  // hbuf=alloc_hbuf(); // all tasks?
  reset_hbuf(hbuf); /* 2015-01-31 */
  /* 2016-11-10 avoid re-use of old values on failed malloc */
  for (j=0;j<NEW;j++) ew[j]=0.0;

  // hbuf->num_ahead_lines=1; /* ToDo: check! v2.32 */
  if (mpi_i>0) F0("missing MPI code");
  if (xm==NULL || xd==NULL || ix==NULL) rc=-1;  /* return error */
  if (xm!=NULL && xd!=NULL && ix!=NULL) {
    int xflag=0;
    cplx hh;    /* temporary store for array element double/complex */
    double ahh; /* temporary fabs(hh) (slow) */
    if (vvv && !mpi_i) printf(" store matrix H ...");
    /* --- generate a simple matrix, store all elements to xm[] --- */
    tm1=fstime();
    if (xm) for(i=0;i<iv*n1*n1;i++) { xm[i]=0.0; }
    if (xv) for(i=0;i<iv*n1*n1;i++) { xv[i]=0.0; }
    for (blk=0;blk<pt_n;blk++) for(i=0;i<b_ofs[blk+1]-b_ofs[blk];/* i++ */){
      hbuf->n=0; /* reset H-buffer, ToDo: num_ahead_lines=1 ??? */
      hbuf->block=blk;
      /* get hbuf->el.jj+r fulldiag, single thread only */
      hamilton_nhv( i, hbuf->num_ahead_lines, hbuf, 8); // all tasks?
      /* printf("\n n=%d ",hbuf->n[0]); */
      for(ii=0;ii<hbuf->n;ii++) {
        j =hbuf->jj[ii]; xflag=((j&XY_FLAG)?1:0); j&=~XY_FLAG;
        hh=hbuf->rr[ii];   /* complex matrix element */
        /* if (i+b_ofs[blk]==j) trace1+=Re(hh); */ /* just as a test */
        /* lapack zheev computes max(abs(a[i,j])) for scaling */
        if ((vvv&32)&&!mpi_i)
          printf("\n H blk= %2d xy= %3ld %3ld r= %6g ",
                       blk, i+b_ofs[blk], j, D(hh));
        if (xm) {
#ifdef LAPACK
          /* */      xm[((i+b_ofs[blk])*n1+j)*iv  ] += Re(hh);
          if (iv==2) xm[((i+b_ofs[blk])*n1+j)*iv+1] += Im(hh);
#else /* jacobi */
          /* 2013-04-04 avoid flt+=dbl, use fltA = fltA + dblB */
          xm[              (i+b_ofs[blk])*n1+j] =
           M_d2f( M_f2d(xm[(i+b_ofs[blk])*n1+j]) + hh);
#endif
        } // xm
        i12+= Im(hh)*Im(hh); /* test for a reel matrix (0 if real) */
        /* do lapack.zlanhe needed in lapack.zheev 2013-05-24 */
        ahh=Cabs(hh); /* sqrt(r*r+i*i)==hypot(r,i) */
        if (max_abs_H < ahh) max_abs_H = ahh;
        if (xflag) i++; /* ToDo: v2.32 ??? check correctness  */
      } // for hbuf->n
    } // for blk=0..pt_n-1 for ofs[]
    tm2=fstime()-tm1;
    if (vvv && !mpi_i) {
      printf("\nPRF: storeH.t[m]= %.3f",tm2/60.);
      printf("\n max_abs_H= %e", max_abs_H);
      if (max_abs_H<1e-30 || max_abs_H>1e30) printf(" WARNING!");
    }
#if 0
    { int i1,i2;
      for (i1=0;i1<n1;i1++) for (i2=i1+1;i2<n1;i2++)
        if (fabs(xm[i1*n1+i2]-xm[i2*n1+i1])>1e-12)
          printf("\n %3d %3d % e",i1,i2,xm[i1*n1+i2]-xm[i2*n1+i1]); }
#endif
    /* Epsilon       = 5.96046E-08 */
    /* Safe minimum  = 1.17549E-38 */
    /* max_abs_H < sqrt(safeminimum/eps=1e-30) sigma=sqrt()/max_abs_H */
    /* max_abs_H > sqrt(eps/safeminimum=1e+30) sigma=/max_abs_H/sqrt() */
    /* eps   = relative machine precision 1e-12? */
    /* sfmin = safe minimum, such that 1/sfmin does not overflow 1e-10? */
    if (iv==2 && i12==0 && !mpi_i)
      printf("\nWARNING: You are wasting memory"
             " and time using complex mode for real matrix H!");
    /* 1024^2*dbl=8M, 20k^2*dbl=3GB max_size=1024 max_scale=32 */
    if (vvv && !mpi_i)
      printf("\n use vvv&256 to output matrix.pgm (n1<=32*1024)\n");
    if ((vvv&256)&& n1<=32*1024 && !mpi_i) { /* compress n1>=1024 JS2011 */
      FILE *f1; unsigned int bb, ib;
      int    *sorti=NULL; 
      double *diagH=NULL, maxnondiag=0.0;
      sorti=           (int    *)malloc(n1*S_int); /* JS 2011-12-12 */
      if (sorti) diagH=(double *)malloc(n1*S_double);
      if (!sorti || !diagH) E0("malloc sorti,diagH");
#define Hfull(ii,jj) xm[(ii*n1+jj)*iv]
      printf(" use vvv&768 for sorted matrix.pgm\n");
      if (diagH) for (i=0;i<n1;i++) {
        /* ToDo: we need Jzz but use diagH */
        sorti[i]=i; diagH[i]=Re(Hfull(i,i));
        /* 2011-11-15 JS better sorted? using sum off-diag */
        for (j=0;j<n1;j++) diagH[i]+=/* -fabs */(Re(Hfull(i,j))*1e-9); 
      }
      if (diagH && (vvv&512)) shellsort(diagH, sorti, n1);
      /* 2011-11-15 JS better sorted? but roughly marked */
      if (diagH) for (i=0;i<n1;i++) diagH[i]=Re(Hfull(i,i));
      /* output of relevant Hii ranges by gray bg, 2011-11-15 JS updated */
      if (diagH && (vvv&512))
      for (ib=1,i=0;i<n1;i++) {
        if (i>0) {ib=0; if (fabs(diagH[sorti[i  ]]
                                -diagH[sorti[i-1]])>1e-6) ib=1;}
        if (ib) printf("\n %6lu  ii= %6d  Hii= %15.10f",
                        i, sorti[i], diagH[sorti[i]]);
      }  if (!mpi_i) printf("\n");
      f1=fopen("matrix.pgm","w+"); /* PGMBIN 1=black 0=white HL */
      if (diagH && f1) {
        /* 2011-11-14 JS changed from PBMBIN to PGMBIN + scale=1..maxint */
        unsigned int n1sc=n1, scale=1, is, js;
        /* screen=1024 300dpi*A4=2700px? */
        if (n1>1023) { scale=n1/1024+1; n1sc=n1/scale; }
        fprintf(f1,"P5\n# SpinPack matrix H d=%lu\n%u %u 255\n",
             n1, n1sc, n1sc);
        for(i=0;i<n1;i++)
          for(j=i+1;j<n1;j++)
            if (maxnondiag<Cabs(Hfull(i,j)))
                maxnondiag=Cabs(Hfull(i,j));
        printf(" maxnondiag= %.0e", maxnondiag);
        /* round() math.h -lm _XOPEN_SOURCE >= 600 || _ISOC99_SOURCE; ToDo */
        /*     special cases: round(0.5)=1, round(-0.5)=-1 */
        /* dbl rint(dbl): BSD4.3 _BSD_SOURCE || _SVID_SOURCE
             || _XOPEN_SOURCE >= 500 || _ISOC99_SOURCE; or cc -std=c99 */
        /*  ceil(0.5) is 1.0, and ceil(-0.5) is 0.0 (next bigger integral) */
        /* floor(0.5) is 0.0, and floor(-0.5) is -1.0 (lower integral) */
        /* round(x) replaced by ceil(x-0.5), floor(x+0.5)  2013-04-24 */
        /* need nearest, J1=0.01 J2=1 should show J2 matrix */
        for (i=0;i<n1sc;i++) { /* 2016-12 removed usage ib bit 0+1+3 */
          ib=0; /* ib = background color, showing same sorted-diag range */
          if (i>0 && (vvv&512))
            if ( fabs(diagH[sorti[ i   *scale]]
                     -diagH[sorti[(i-1)*scale]])>1e-3) ib=4;
          if (n1sc<40) printf("\n%c ",((ib)?'+':'-')); /* 2016-12 add TXT */
          for (j=0;j<n1sc;j++) {
            if (j>0 && (vvv&512))
              if ( fabs(diagH[sorti[ j   *scale]]
                       -diagH[sorti[(j-1)*scale]])>1e-3) ib^=4;
            bb=0; /* intensity of pixel propto nonzero-part, 255=white */
            for (is=0;is<scale;is++) /* scale is .ge. 1 */
            for (js=0;js<scale;js++)
              if (Re(Hfull(sorti[i*scale+is],sorti[j*scale+js]))) bb++;
            fprintf(f1,"%c", (255-bb*64/(scale*scale)) ^ ((ib&4)*4) );
            if (n1sc<40) printf(" %c",((bb)?'1'+8*bb/(scale*scale):'.'));
          }
        }
        fclose(f1);
      } /* open(f1) */
      if (sorti) free(sorti); /* JS 2011-12-12 */
      if (diagH) free(diagH);
    } /* vvv&256 == output matrix.pgm */
    if (xm) {
      for(i=  0;i<n1;i++) { trace1+=Re(xm[(i*n1+i)*iv]); } /* matrix trace */
      for(i=  0;i<n1;i++)            /* as a check sum upper left elements */
      for(j=i+1;j<n1;j++) {  h12+=Re(xm[(i*n1+j)*iv]); }
    }
    tm1=fstime();  /* start time fulldiag */
#ifdef LAPACK
    ii=(int)n1;              /* buffer for n1 */
 /* lwork=(nb+1)*n1; */  /* zheev-optimum = (32+1)*n1 working space */
 /* lwork=(nb+2)*n1; */  /* dsyev-optimum = (32+2)*n1, 2016-04-13 */  
    lwork=(int)((32+2)*n1); /* buffer for lwork >= max(1,2*n-1) better (nb+1)*n */
    /* 2013-05-24 nrot=3*n1 is less than recommended 33*n1
       and gives sometimes poor results (see bugs.tex) */
    /* for cplx (nb=32+1)*n1 (n1=489 3*n1=0.024m) */
    /* ToDo: problem if nrot is 4 byte and n1^2 is larger than 2e9 */
    if (iv==2) { /* cplx aaa=1(aldo zheevd) */
      lwork=(32+1)*n1;   /* zheev-optimum = (32+1)*n1 working space */
      lrwork= 3*n1-2;  /* added 2016-04-13 for rwork, unused for DSYEV/ZHEEV */
      if (aaa==1) lwork= (nev) ? 0+2*n1+n1*n1 +30*n1 : 1+n1; /* +30n1 faster */
      if (aaa==1) lrwork=(nev) ? 1+5*n1+2*n1*n1 : n1; /* optimum 2016-04-26 */
    } else {
      lwork=(32+2)*n1;   /* dsyev-optimum = (32+2)*n1, 2016-04-13 */
      if (aaa==1) lwork= (nev) ? 1+6*n1+2*n1*n1 +26*n1: 32+2*n1; /* dsyevd 2016-04 */
    }  /* dsyevd-optimum "V" = 2016-04-25 */
    if (aaa==1 && !mpi_i){
      liwork= (nev) ? 3+5*n1 : 1;  /* zheevd, dsyevd */
      iwork= (int *)malloc(liwork*sizeof(int));
      if (!iwork){printf(" malloc *iwork failed ... fallback ~AAd\n");aaa=0;}
    }
    /* use  ev as matrix  and  xm as workbuffer */
    if (!mpi_i) {
      myfree((void *)xv,&m_xv);
      xv=(double *)mymalloc(lwork*iv*S_double,&m_xv); /* work-space */
      if (!xv) E0("lapack.malloc workspace xv failed");
    }
    if (!mpi_i) {
      int info=-1;
      if (iv==1 && aaa) { int lwork2=-1, liwork2=-1; /* output int array!? */
        dsyevd_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork2,iwork,&liwork2,&info);
//    dsyevd_(&jobz,"U",&n1,a,&lda,w,work,&lwork2,iwork,&liwork2,&info);
        printf(" optimal workspace  work= %9.0f <= %9d\n",D(xv[0]),lwork);
        printf(" optimal workspace iwork= %9d <= %9d\n",iwork[0], liwork);
        printf("\n LAPACK dsyevd"); /* 2016-04 dsyevd is faster + parallel */
        dsyevd_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork,iwork,&liwork,&info);
      }
      if (iv==1 && !aaa) { int lwork2=-1; /* output int via cplx array!? YES! result=(32+2)*n1 */
        dsyev_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork2,&info); /* dsyevd_() */
        printf(" optimal workspace  work= %9.0f <= %9d\n",D(xv[0]),lwork);
        printf("\n LAPACK dsyev"); /* 2016-04 dsyevd is faster for nev=0 */
        dsyev_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork,&info); /* dsyev_() */
      }
      /* --- complex hermitian matrix, iv=2 --- */
      minfo_t m_rw = {NULL,0,0,"tmp/rw.map",NULL}; /* work-buffer */
      if (iv==2) {
        rwork=(double *)mymalloc(lrwork*sizeof(double),&m_rw);
        if (!rwork) { E0(" malloc *rwork failed - skip zheevd");aaa=0; }
      }
      /* valgrind shows that zhemv_ of blas-3.0 accesses WORK+len=xm+lwork...
       * so we malloc more elements than lwork=LWORK counts    */
      if (iv==2 && aaa) { int lwork2=-1; /* output int via cplx array!? YES! result=(32+1)*n1 */
        int lrwork2=-1;
        int liwork2=-1; /* detect optimum workspace size */
        zheevd_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork2,
                rwork,&lrwork2,iwork,&liwork2,&info); /* zheevd_() */
        printf(" optimal workspace  work= %9.0f <= %9d\n", D(xv[0]),lwork);
        printf(" optimal workspace rwork= %9.0f <= %9d\n",rwork[0],lrwork);    
        printf(" optimal workspace iwork= %9d <= %9d\n",iwork[0],liwork);      
        printf(" LAPACK zheevd");
        zheevd_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork,
                 rwork,&lrwork,iwork,&liwork,&info); /* zheevd_() */
      }
      if (iv==2 && !aaa) { int lwork2=-1; /* output int via cplx array!? YES! result=(32+1)*n1 */
        zheev_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork2,rwork,&info);
        printf(" optimal workspace  work= %9.0f <= %9d\n", D(xv[0]),lwork);
        printf(" LAPACK zheev");
        zheev_(&cev,"U",&ii,xm,&ii,xd,xv,&lwork,rwork,&info);
      }
      printf("_info=%d (0=ok)\n",info);
      /* free(rwork); rwork=NULL; */
      if (rwork) myfree(rwork,&m_rw); /* crashes using mpt + openblasp 2016-04 */
    } /* !mpi_i */
  /* if not freed, zheev causes segfault after about 7 runs on Tru64-cxx */
    if (iwork) { free(iwork); iwork=NULL; }
#else // not LAPACK (slower)
    if (!mpi_i) {
      printf(" JACOBI_cplx"); /* JS 2011-12-10 speedup ... for cplx */
      cjacobi(n1, xm, xd, ((nev)?xv:NULL));
      shellsort(xd,ix,n1);	/* sort xd[ix[i]], 2016-04 bad iv*n1 */
    }
#endif
    tm2=fstime()-tm1; /* pure lapack time measured without output */
    if (!mpi_i) printf(" eigenvals=");
    for (j=0;j<n1;j++){ trace2+=xd[ix[j]]; printf(" %+12.8f",xd[ix[j]]); }
    printf("\n  traceH= %e sumE= %e traceH-sumE= %e\n sum_h12= %e",
           trace1, trace2, trace2-trace1, h12);
    if (vvv&4) printf("\n number eigenvals eigenvects (sorted):");
    /* convert simple vectors to block form for calling <O>-functions */
    for (j=0;j<NEW && j<n1;j++) ew[j]=xd[ix[j]]; /* 2012-05-05 */
    if (nev)
    for (j=0;j<n1;j++) { /* all eigenvects */
      //  if (vvv&1) printf("\n%3ld %3d: %+12.8f",j,ix[j],xd[ix[j]]); /* eigenvalue */
      if (vvv&1) printf("\n%3ld: %+12.8f",j,xd[ix[j]]); /* eigenvalue */
      if (xv && xm) /* 2011-10-28 in case of low mem, no eigenvects */
      for(k=0;k<b_ofs[pt_n];k++){
#ifdef LAPACK   /* this is hopefully right, normalization? */
        xr=           xm[(k+n1*ix[j])*iv  ]; /* no ix[j] needed, sorted */
        if (iv>1) xi=-xm[(k+n1*ix[j])*iv+1];
        /* the minus sign was found empirically (N=6=4+2, k=-2 1) */
#else /* transposed ev-matrix! */
        xr= Re(xv[k*n1+ix[j]]); /* buggy until 2016-04 from 2011-12-- ? */
        xi= Im(xv[k*n1+ix[j]]);
        xr= Re(xv[k+n1*ix[j]]); /* fixed 2016-04 */
        xi= Im(xv[k+n1*ix[j]]);
#endif
        if (vvv&4) { /* output all coefficients of the eigenstate */
          printf(" ");
          if(iv>1)  printf("(");
                                 outmpl(stdout,xr);
          if(iv>1){ printf(","); outmpl(stdout,xi); printf(")"); }
        }
        /* copy ev in blk-structure */
        if (!A_ptr(ev)) continue;
        ev[0][k] = M_d2f(C(xr,xi)); /* ignore xi if not complex */
      } // k=0...b_ofs[pt_n](n1) ???
      if (!A_ptr(ev)) continue; /* 2016-04-29 added to avoid wop-output */
      /* call <ev|O|ev>-functions here, SiSj-example: */
      if(cfg_sisj&5)
      { /* new block necessary to allow definition of int sites */
        int i1,i3,sites[2];   /* must follow the left bracket */
        /* could be problematicaly on degenerate states! not proofed */
        for(i3=0;i3<sym.num_dcorr;i3++){
          sites[0]=sym.dcorr[i3][0];
          sites[1]=sym.dcorr[i3][1];
          /* Warning: wop want to have *hbuf[B_NUM] but gets *hbuf[1] */
          printf(" Z%dZ%d= %9.6f", sites[0], sites[1],
                  wop(&op_zizj,NULL,2,sites,ev[0],ev[0],&hbuf));
          if(cfg_sisj&1)
          printf(" S%dS%d= %9.6f", sites[0], sites[1],
                  wop(&op_sisj,NULL,2,sites,ev[0],ev[0],&hbuf));
        }
        for(i3=-1,i1=0;i1<nn;i1++) if(sym.corr[i1][i1]>i3){ /* oct09 Zi */
          i3=sym.corr[i1][i1];
          printf(" Z%d= %9.6f",i1,wop(&op_zi,NULL,1,&i1,ev[0],ev[0],&hbuf));
        }
      } // <ev|O|ev>-functions
      /* normvec(ev[0]);        LAPACK not normalized? */
      if(cfg_sisj&2) wss(ev[0],ev[0],&hbuf); /* output S^2 */
    } // j=0..n1-1, all eigenvects
    if (vvv&4 ) printf("\n");
    if (A_ptr(ev))
    for (j=0;j<(ULong)nev;j++){    /* copy eigenvects */
      for(k=0;k<b_ofs[pt_n];k++){ 
#ifdef LAPACK   /* this is hopefully right */
        xr=           xm[(k+n1*ix[j])*iv  ];
        if (iv>1) xi=-xm[(k+n1*ix[j])*iv+1];
#else
        xr= Re(xv[k*n1+ix[j]]); /* bad row/column ToDo 2016-04 */
        xi= Im(xv[k*n1+ix[j]]);
        xr= Re(xv[k+n1*ix[j]]); /* fixed 2016-04 */
        xi= Im(xv[k+n1*ix[j]]);
#endif
        ev[j][k] = M_d2f(C(xr,xi));
      }
    } // for $nev eigenvects
    if (!mpi_i) printf("\nPRF: fulldiag.t[m]= %.3f", tm2/60.);
  } // if(xv!=NULL && xd!=NULL && ix!=NULL) malloc=OK
  /* 2016-05 move PRF to the end, no more scrolling for PRF-tests */
  if (!mpi_i) printf("\n"); /* 2016-04 free-msg in next line! */
  if (ix) myfree((void *)ix,&m_ix);
  if (xm) myfree((void *)xm,&m_xm);
  if (xv) myfree((void *)xv,&m_xv);
  if (xd) myfree((void *)xd,&m_xd);
//    free_hbuf(hbuf);
  return rc;
}

/* ================== read/write parameters  ====================== */

char *strsub(char *sd,const char *ss,int i1,int i2){/* copy ss[i1..i1+i3-1] */
 int i3;for(i3=0;i3<i2;i3++)sd[i3]=ss[i1+i3];sd[i2]=0;return(sd);}
char *fk(const char *ss,int i2){ /* get i2-tes element from komma-list */
 static char s1[512];int i1,i3,i4,i5;char s2;
 s1[0]=0;
 i1=0;  /* pos letztes komma +1 */
 i3=0;  /* klammerebene */
 i4=0;  /* komma-nr */
 for(i5=0;i5<(int)strlen(ss)+1;i5++){
  s2=ss[i5];if(s2==0)s2=',';
  if(s2=='(') i3++;else
  if(s2==')') i3--;else
  if((s2==',') && (i3==0)){ 
    i4++; if(i4<i2) i1=i5+1; else break;  } }
 if((i4>=i2) && (i2>0))strsub(s1,ss,i1,i5-i1);
 if((s1[0]=='(')&&(s1[strlen(s1)-1]==')'))strsub(s1,s1,1,(int)strlen(s1)-2);
 return s1; }


cplx atoc(char *s){	/* (r,w) => r*exp(I*w*Pi) */
  double x=0.0, y=0.0; int i=0;
  while(s[i]!=0 && s[i]!=',')i++;
  x=atof(s); if(s[i]==',') y=atof(&s[i+1]);
#if VecType & 4     
  return polar(x,y*M_PI);}
#else
  if(y!=0)E0("imag(r)!=0"); return C(x,0);}
#endif     

void outpara(FILE *in){
  int i;
  fprintf(in," param=");
  for(i=1;i<numpara;i++){
    fprintf(in," %.9g",(double)para[i]);
    if(i<numpara-1)fprintf(in,",");
  } 
  fprintf(in,"\n# p1= %g\n# p2= %g  ... can also be used\n",
          (double)para[1], (double)para[2]);
  fflush(in);
  return;
}

int set_ww(){
  int i;
  for(i=0;i<nw;i++){
    ww.co[i]=cos(M_PI*para[ww.w[i]]*ww.ex[i]);
    ww.si[i]=sin(M_PI*para[ww.w[i]]*ww.ex[i]);
  }
  return 0;
}

int scan_ud(char *s1,int sum){
  if (sum) {  /* add values, used for loops */
    nu+=atoi(fk(s1,1));
    nd+=atoi(fk(s1,2));
  } else {
    nu=atoi(fk(s1,1));
    nd=atoi(fk(s1,2));
  }
#if CM(1,0,0)  /* 2015-03 add output pre state, 2016-03 fix nu>nn */
  if(nu+nd!=nn || nu<0 || nd<0){
    if (!mpi_i) printf("\nwarning: nu+nd!=N nu=%d nd=%d N=%d",nu,nd,nn);
    if (nu<-nn) nu=1; /* 2016-03 fix -40+140==100 */
    if (nd<-nn) nd=1;
    if (nu>nn) nu=nn-1;
    if (nd>nn) nd=nn-1;
    if (nu<0 && nd>=0) { nu=nn+nu; nd=nn-nu; } /* want have high Sz ? */
    if (nd<0 && nu>=0) { nd=nn+nd; nu=nn-nd; } /* want have high Sz ? */
    if (nu+nd!=nn && abs(nu-nd)<=nn) {nu=(nn-abs(nu-nd))/2;nd=nn-nu;} /* low Sz */
    if (nu+nd!=nn && nu<nn) nd=nn-nu; /* easy to solve */
    if (nu+nd!=nn && nd<nn) nu=nn-nd; /* easy to solve */
    if (nu+nd!=nn) { nu=nn/2; nd=nn-nu; } /* Sz=0 or minimum else */
    if (!mpi_i) printf("\nwarning: nu+nd!=N set nu=%d nd=%d",nu,nd); }
#endif  
#if CM(0,1,0)
  if(nu+nd>nn || nu<0 || nd<0){
    if (!mpi_i) printf("\nwarning: nu+nd> N nu=%d nd=%d N=%d",nu,nd,nn);
    if (nu<-nn) nu=1; /* 2016-03 fix -40+140==100 */
    if (nd<-nn) nd=1;
    if (nu<0) { nu=nn+nu; } /* want have high nu ? */
    if (nd<0) { nd=nn+nd; } /* want have high nd ? */
    while (nd+nu>nn && nu>=0 && nd>=0) { nu--; nd--; }; /* easy to solve */
    if (nd+nu>nn) { if (nd>nu) nd=nn-nu; else nu=nn-nd; }
    if (!mpi_i) printf("\nwarning: nu+nd> N set nu=%d nd=%d",nu,nd); }
#endif  
#if CM(0,0,1)
  if(nu>nn || nd>nn || nu<0 || nd<0){
    if (!mpi_i) printf("\nwarning: nu,nd> N nu=%d nd=%d N=%d",nu,nd,nn);
    if (nu<-nn) nu=1; /* 2016-03 fix -40+140==100 */
    if (nd<-nn) nd=1;
    if (nu<0) { nu=nn-nu; } /* want have high nu ? */
    if (nd<0) { nd=nn-nd; } /* want have high nd ? */
    if (nu>nn) nu=nn; /* easy to solve */
    if (nd>nn) nd=nn;
    if (!mpi_i) printf("\nwarning: nu,nd> N set nu=%d nd=%d",nu,nd); }
#endif
  return 0;
}

int scan_para(char *s1, int sum){
  int i;
  if (sum)  for(i=1;i<numpara;i++){para[i]+=atof(fk(s1,i));}
  else      for(i=1;i<numpara;i++){para[i]= atof(fk(s1,i));}
  set_ww();
  return 0;
}

/* =========================== useful functions (handling strings) ====== */

void testsym(){
}

/*
 * search end of line
 */
char *scan_eol(char *s){
  char *s2;
  s2=strchr(s,'#');if(s2) return s2; /* comment-line */
  s2=strchr(s,'%');if(s2) return s2; /* comment-line */
  s2=s+strlen(s);
  return s2;
}
/*
 * search "vnam=" in s, get position of next char, ret 0=not found
 *  ToDo: search expression "p[0-9]*=" or p%d=%f ?
 */
char *scan_var(char *s,const char *vnam){
  char *s1,*s2;
  if ((s1=strstr(s,vnam))==NULL) return NULL;
  s1+=strlen(vnam);
  s2=scan_eol(s); if(s2) if(s2<s1) return NULL; /* comment */
  if (s1[0]!='='){ if (vvv) printf(" %s=???",vnam); return NULL; }
  s1++; while(s1[0]==' ' || s1[0]=='\t') s1++;
  return s1;
}

int scan_int(char *s,const char *vnam,int *var){
  char *s1;
  if( (s1=scan_var(s,vnam))==NULL ) return -1;
  if(var){ *var=atoi(s1); if(vvv && !mpi_i) printf(" %s=%d",vnam,*var);}
  return 0;
}

int scan_inta(char *s,const char *vnam,int *var,int len){
  char *s1,*s2; int i=0;
  if( (s1=scan_var(s,vnam))==NULL ) return -1;
  s2=scan_eol(s); len=(int)(s2-s1);
  if(var){
   if (vvv && !mpi_i) printf(" %s=",vnam);
   for(i=0;i<len;i++){
    var[i]=(int)strtol(s1,&s2,10);
    if(s1==s2 || s2[0]==0) break;
    if (vvv && !mpi_i) printf(" %d",var[i]);
    s1=s2;
   }
   if (vvv && !mpi_i) printf("\n");
  }
  return i;
}

/* search "vnam=" in s, set value to var, ret 0=found,1=not found,2=err
   last modify 1998-139 */
int scan_dbl(char *s,const char *vnam,double *var){
  char *s1;
  if( (s1=scan_var(s,vnam))==NULL ) return -1;
  if(var){ *var=strtod(s1,NULL);
    if (vvv && !mpi_i) printf(" %s=%g",vnam,*var);
  }
  return 0;}

int p_mask_ini(int nn){
  int i,j;
  /* --- initialize bitmasks for faster parity() --- */
  for(i=0;i<nn;i++)for(j=0;j<nn;j++)
    if(i<j) p_mask[i][j]=(~(~(tbase1)0 << j << j)) & (~(tbase1)0 << i << i << 2);
    else    p_mask[i][j]=(~(~(tbase1)0 << i << i)) & (~(tbase1)0 << j << j << 2);
  return 0;
}

void ns_thread( void *ptr); /* see hilbert.c */
void wopij_block( void *ptr ); /* see xval.c */
void wop_block( void *ptr ); /* see xval.c */
void diag2_thread( void *ptr ); /* see diag.c */

/* ToDo: write a test_task which only outputs "(mpi.thread): I am here" */
void test_pthread( void *ptr ){
  main_pts *pp=(main_pts *)ptr;
  if (mpi_i*pt_n+pp->num<4 || mpi_i*pt_n+pp->num>mpi_n*pt_n-1-4)
    printf(" node.thread %d.%d is running as block %d\n",
           mpi_i, pp->num, mpi_i*pt_n+pp->num);
  else
  if (mpi_i*pt_n+pp->num==4)
    printf(" skip output of threads %d.%d..%d.%d\n",
      mpi_i, pp->num, (mpi_n*pt_n-1-4)/pt_n, (mpi_n*pt_n-1-4)%pt_n);
}

/* void-function for pthreads, waiting for work, send via main_pts[] */
void main_pthread( void *ptr ){
#ifdef CONFIG_PTHREAD
  /* main_pts = struct {} */
  /* why volatile was outcommended until 2016-12, think its needed */
  /* 2016-12-22 possible icpc may create endless-loop without volatile? */
  volatile main_pts *pp=( volatile main_pts *)ptr;
#if defined(CFG_CPUSET)
  /* ToDo: use CPUSET scheme of memspeed or mpi_stress !! 2014-06 */
  { // this processor bounding is useful for the case, that the jobsystem
   //   does not do this automatically
   // ToDo16: --pin_ncpu=16 --pin_ofs=0 --pin_dist=1 --pin_size=1 ???
   // ToDo 2016: comp2 getaffinity dflt= 0x0000ffff + use this mask and
   //             take one of each bit for tasks,
   //             allows dynamic taskset multibits from outside
 // 2016-12-22 replace use of hard coded mask CFG_CPUSET by getaffinity()         
 //            example use cores 4-8: taskset 0x0FF0 ./spin -t8
 // ToDo16: use KMP_AFFINITY=[<modifier>,...]<type>[,<permute>][,<offset>]
   cpu_set_t cpuset; int ii;
   /* 2016-12-22 added more info, usually 1024, bigger than numcores */
   if (mpi_i==0 && pp->num==0)
     printf("DBG_PT: cpuset[bits]=%u  CPU_SETSIZE=%u\n",
       (unsigned)(8*sizeof(cpu_set_t)), CPU_SETSIZE);
   /* 2016-12 comp2: 0x0000ffff on 16-core machine without any affinity set */
   /* ToDo2017 what about pthread_getaffinity_np ??? */
   /* int sched_setaffinity(pid_t pid, unsigned int cpusetsize,
    *                       cpu_set_t *mask);  + <sched.h>
    *   pid=0 is local thread, or use pid_t gettid(void); 
    * int pthread_getaffinity_np(pthread_t thread, size_t cpusetsize,
    *                       cpu_set_t *cpuset); + pthread.h + _GNU_SOURCE
    *   to be set on motherthread??? (not neccess.)
    *   but ntpl/pthread.h must be used? sched is better!? use pthread_self()?
    */
   if (sched_getaffinity(0, sizeof(cpu_set_t), &cpuset)) {
     if (mpi_i<8)
       printf("ERR_PT getaffinity failed for thread %2d %2d\n",mpi_i,pp->num);
   }
   /* output 2 first and 1 last (to see bad effects like t1 set by t0) */
   if (mpi_i*pt_n+pp->num < 4 || mpi_i*pt_n+pp->num >= mpi_n*pt_n-1) {
     ULong cpuset2=0L; int nbits=0;
     for(ii=0;ii<8*(int)sizeof(cpuset2);ii++) 
       if(CPU_ISSET(ii,&cpuset)) { cpuset2|=1UL<<ii; nbits++; }/* 64bit-mask */
     printf("DBG_PT: getaffinity for thread %2d %2d is cpuset=%08lx numbits=%d\n",
             mpi_i, pp->num, cpuset2, nbits); /* bad for 32/64++ cores */
   }
#if CFG_CPUSET > 1   /* 2017-01-30 added */
   { int  ncores=0, in=0; /* reduce mask to 1 bit */
     for (ii=0; ii<CPU_SETSIZE && ncores<B_NUM; ii++)
                                      if (CPU_ISSET(ii,&cpuset)) ncores++;
     for (ii=0; ii<CPU_SETSIZE; ii++) if (CPU_ISSET(ii,&cpuset)) {
         if ((mpi_i*pt_n+pp->num)%ncores==in) break;  // mycore = 0,1,2,...
         in++; }
     CPU_ZERO(&cpuset); CPU_SET(ii,&cpuset);
     usleep(80000); /* why? hold parallel outputs together */
     /* output 4 first and 4 last */
     if (mpi_i*pt_n+pp->num < 4 || mpi_i*pt_n+pp->num >= mpi_n*pt_n-4) {
       printf("DBG_PT: setaffinity for thread %2d %2d to core %2d, node=%s\n",
              mpi_i, pp->num, ii, nodename);
     }
     // no sched_setaffinity on OSF1
     if (sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) {
       /* failes if ii is above real core numbers (2016-12 comp2 size=1024) */
       printf("ERR_PT: setaffinity for core(%2d)=%2d failed\n",
               mpi_i*pt_n+pp->num, ii); usleep(100000); perror("setaffinity");
     }
     /* 2016-12-22 added getaffinity */
     if (sched_getaffinity(0, sizeof(cpu_set_t), &cpuset)) {
       if (mpi_i<8) printf("ERR_PT: getaffinity failed for task %2d\n", pp->num);
     }
     /* output 4 first and 4 last */
     if (mpi_i*pt_n+pp->num < 4 || mpi_i*pt_n+pp->num >= mpi_n*pt_n-4) {
       ULong cpuset2=0L;
       for(ii=0;ii<8*(int)sizeof(cpuset2);ii++) 
         if(CPU_ISSET(ii,&cpuset)) cpuset2|=1UL<<ii; /* 64bit-mask */
       printf("DBG_PT: getaffinity for thread %2d %2d is cpuset=%08lx %s\n",
               mpi_i, pp->num, cpuset2, nodename); /* bad for 32/64++ cores */
     }
     usleep(100000); /* why? wait for other outputs, ToDo: like ordered by OMP */
   }
#endif /* CFG_CPUSET > 1 */
  }
#endif /* CFG_CPUSET defined */
  /* test_pthread(ptr); */
#ifdef _OPENMP
  if (pp->num<8 && !mpi_i)
    printf("DBG_PT: enter worker loop %d\n", mpi_i*pt_n+pp->num); 
  /* endless wait for new work from master thread */
  for(;;) {
    while (pp->task==0)  /* volatile! */
      usleep(1000);  /* 10ms testwise */
    if (pp->task<0) break;
    pp->worker( (void *)pp->data ); /* worker thread */
#pragma omp critical
{
    pp->task=0;
}
#pragma omp flush (pp)  // ??? ToDo pointer
  }  /* for(;;) */
#else /* ~_OPENMP */
  for(;;) {  /* endless loop for worker (child-) threads */
    pthread_mutex_lock(&pt_mutex1);
    // printf("DBG: thread %d starts waiting\n",pp->num);
    while (pp->task==0)  /* volatile! */
      pthread_cond_wait(&pt_cond1,&pt_mutex1); /* unlock+wait_cond+lock */
      /* ToDo: better have one(two) condition variable(s) per thread */
    pthread_mutex_unlock(&pt_mutex1);

    if (pp->task<0) break;
    // printf("DBG: thread %d starts work\n",pp->num);
    pp->worker( (void *)pp->data ); /* call worker function for this thread */

    pthread_mutex_lock(&pt_mutex1); /* ToDo: check for race condition? */
    pp->task=0; /* ready */
    pthread_cond_signal(&pt_cond2); /* wake up master */
    pthread_mutex_unlock(&pt_mutex1);
    /* ToDo16: should we give the master a chance to lock here? race-locks? */
    /*         no. pthread_cond_wait does unlocking, which is enough */
  }
  pthread_exit(0);
#endif /* _OPENMP */
#endif /* CONFIG_PTHREAD */
} /* main_pthread() */

extern int sym_bn_m_ini_flag;  /* see hilbert.c */
void sym_bn_m_ini(int pt);
extern int chkpt_time; /* diag.c checkpoint interval in seconds (4d) 2012 */
/* ---------------------------- MAIN --------------------------------- */
int    main(int argc,char *argv[]){
  const char *name_file_i="daten.i", *name_file_m="daten.def";
  static char s1[500], *ss;
  ULong n1max=1000;
  int i, ii=0, i1, j, itr=0, iv, nev, ne0,
    nruns=0, /* count Lanczos rounds, use for srand of randvec() */
    block;
  FILE *in1=NULL; int in1line=0; long in1pos=0; /* 2015-02-26 +pos */
  /* 2016-04-15 startvec moved from global to local var, see randvec() */
  int  startvec=4; /* see vector.c randvec(), dflt: 4=random_v */
  int max_ea=1;	/* maximum expected degeneracy of ew[NE0] test-loop,dflt=1 */
  /* int? in2 for inlined .i-files, removed for simplicity 2012-10 */ 
  double r1;
  /* dynamic allocation of SH-buffer per thread, done before 
   *   allocation of rest memory for matrix, as thread-local mem */
  thbuf * (hbuf)[B_NUM];
  tvec *v0, *v1, *ev[NEV]; /* large vectors */
  minfo_t m_v0, m_v1, m_ev[NEV],  /* memory descriptors for vectors */
          m_l2 = {NULL,0,0,"tmp/l2.map",NULL};
#if defined(CONFIG_PTHREAD) && !defined(_OPENMP)
  /* stack memory is local, memory + filesystem is shared */
  pthread_t pt[B_NUM];
#endif
#ifdef CONFIG_MPI /* accept both (MPI+MPP) together */
  int mpi_thread=0; /* 2017-01-17 always output result for debugging */
  int mpi_thread_req=0; /* requested */
  const char *mpi_t[]={"SINGLE","FUNNELED","SERIALIZED","MULTIPLE"}; 
#if defined(CONFIG_PTHREAD) && defined(MPI_THREAD_SERIALIZED)
  /* try setenv LAM_MPI_THREAD_LEVEL=3
   * 0: Corresponds to MPI_THREAD_SINGLE
   * 1: Corresponds to MPI_THREAD_FUNNELED    (only main thread does MPI)
   * 2: Corresponds to MPI_THREAD_SERIALIZED  (only one does MPI at a time)
   * 3: Corresponds to MPI_THREAD_MULTIPLE
   *      test2017.01 mixed: crashes without thread-synchronization
   *                         how to use correctly? one dup_COMM_WORLD/thread?
   * y2016: ToDo: how data flow is with mixed code? intra PT inter MPI only?
   *  ToDo: faster/simpler/minMPI: one MPI/node using multiple threads y2016
   *  ToDo: finally it hould be better using (better) multithread on each node 2017.1
   */
  mpi_thread_req = MPI_THREAD_SERIALIZED;
  if (MPI_Init_thread(&argc, &argv, mpi_thread_req, &mpi_thread)
     != MPI_SUCCESS) exit(1);
#else
  if (MPI_Init(&argc, &argv) != MPI_SUCCESS) exit(1); i=0;
#endif
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_n);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_i);
  for (mpi_n_bits=0;mpi_n_bits<32;mpi_n_bits++)
     if (1<<mpi_n_bits > mpi_n-1) break;
#endif

{ /* global mpi-nodename since 2016-04-08 for better MPI-errors */
  char *s1=NULL; /* 2014-07 get hostname for debugging/errors */
// char nodename[21]="\0HOSTNAME..........."; /* 2016-04-08 set global for err() */
// #if  _BSD_SOURCE || _XOPEN_SOURCE >= 500 || _POSIX_C_SOURCE >= 200112L 
#if defined __USE_BSD || defined __USE_UNIX98  /* 2015-09-23 */
  /* ToDo: set global var and for MPI compute num remote/local connects */
  gethostname(nodename,20); /* Since glibc 2.12, unistd.h */
#endif                 
  /* not set on CentOS, call system("hostname -s > hostname"); ? no */
  s1=NULL; if (nodename[0]==0) s1=getenv("HOSTNAME");
  if (s1) strncpy(nodename,s1,20); /* getenv may return NULL, cpy(NULL)? */
  s1=NULL; if (nodename[0]==0) s1=getenv("SLURMD_NODENAME");
  if (s1) strncpy(nodename,s1,20); /* 2016-04-07 */
  nodename[21]=0; /* ensure string has an end */
}
  if (!mpi_i) { /* output only once per MPI cluster */
   
    printf("\n"
     /* __DATE__ __TIME__ avoids reproducable binaries, use file date */
      "SPINPACK version %s %s\n"
      " Copyright (C) 2001-2017 Joerg (dot) Schulenburg at ovgu.de\n"
      " released under the GNU General Public License\n"
      " base coding for %s-model (%s) NN=%d\n",
      version_string, version_date,
      CM("Heisenberg","t-J","Hubbard"), CM("JJ","tJ","tU"), NN);
    printf("Host: %s\n",nodename); /* 2014-07: debugging speed */
#if   defined(__INTEL_COMPILER)
    printf("Compiler: INTEL_COMPILER %.1f\n",__INTEL_COMPILER*.01);
    /* 1010 = 10.1,  __INTEL_COMPILER_BUILD_DATE 20070913 */
    /* __OPTIMIZE__ 1 (-O1)   CC+DBG=icce+idbe -E -dM */
    /* thread local storage (TLS) model -ftls-model= ??? */
    /* 2017-01 strange hangs using pthread -t64, see bugs  ToDo */
#ifdef __MIC__
    printf(" INTEL-CPU-Features: MIC" /* 2015-02 performance relevant */
#ifdef  __KNC__
        " KNC" /* Knight Corner = Xeon Phi */
#endif /* KNC */
#ifdef __INTEL_OFFLOAD
         " OFFLOAD" /* CoPro-offload */
#endif /* __INTEL_OFFLOAD */
         "\n");
#endif /* __MIC__ */
#endif /* icc */
#if defined(__GNUC__)
    printf("Compiler: GNUC %d.%d.%d\n",
                  __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);
    /* gcc-4.7 AVX2=8*i32 BMI2=pext64=bit_extract Haswell-2013
               bitperm64= 6 times (pext(v,p))<<32|pext(v,~p)
               -mavx2 -mbmi2 -mlzcnt == -march=core-avx2 */
#endif
#ifdef __clang_version__ /* clang -dM -E -x c /dev/null */
    printf("Compiler frontend: Clang " __clang_version__ "\n");
#endif
#ifdef __llvm__  /* clang -dM -E -x c /dev/null */
    printf("Compiler backend: LLVM\n");
#endif
#if defined(__STDC_VERSION__) /* 199901L p.e. gcc -std=c99 +_Complex +// */
    printf("Compiler: STDC %ld\n",__STDC_VERSION__);
#endif
#ifdef __DECCXX_VER
    printf("Compiler: DECCXX %d\n",__DECCXX_VER);
#endif
#ifdef __SUNPRO_CC
    printf("Compiler: SUNPRO_CC 0x%x\n",__SUNPRO_CC);
#endif
#if defined(__mips) && defined(_COMPILER_VERSION)
    printf("Compiler: MIPS_CC %d\n",_COMPILER_VERSION);
#endif
#if defined(__PATHSCALE__)
    printf("Compiler: PATHCC " __PATHSCALE__ "\n"); /* SiCortex "3.2n" */
#endif
    printf(" CPU-Features:" /* 2014-06 performance relevant */
/* gcc: include/immintrin.h __MMX__ __SSE__ ... */
/* gcc: include/cpuid.h     bit_MMX ... bit_XOP ... */
#ifdef __MMX__  // 2i32 4i16 8i8 integer only (we need min. 64bit vectors)
  //   " MMX"  // P5_1997  registers: mm0..mm7
  // MMX:  pandn(dst,src): dst=(~dst)&src;
#endif
#ifdef __SSE__
     " SSE"  // pentium3  xmm0..xmm7 128bit=4f32  ADDPS, ADDSS, MULPS, MULSS
#endif
#ifdef __SSE2__
     " SSE2"  // pentium-m pentium4  XMM+2f64,16i8,8i16,4i32,2i64
     //  __m128d == 2dbl   __m128i == int8-64-arrays  _mm_*
     //  gcc -ftree-vectorizer-verbose=2 -ftree-vectorize -O2
     //  gcc -ftree-vectorizer-verbose=2 -O3
     //  icc-10.1
#endif
#ifdef __SSE3__
     " SSE3"  // complex-arith ADDSUBPD ...PS 
#endif
#ifdef __SSSE3__
//     " SSSE3"  // atom core2  PMULHRSW, PMADDUBSW
#endif
#ifdef __SSE4a__  // see no use exept POPCNT,LZCNT
     " SSE4a"  // core2++?  PMULLD PMULDQ
     // AMD-barcelona++ EXTRQ/INSERTQ mask-shift (bit-range get/set constidx)
     // SSE4.1: PMINUD, PMAXUD packed min/max unsigned double=64bit
     //   not usefull because we need index of max/min for sign
     //     BLEND/PBLEND conditional move (vector a=(b)?c:d) b=imm or reg.
     //     for sign + cfg transfer of vmin(vcfg)
     // SSE4.2: POPCNT,LZCNT count 1-bits i64, leading zeros 64bit
#endif
#ifdef __AVX__  // 4*f64 8*f32 4*i32(only)   quanta1+2 -mavx no256bitFMA?
     " AVX"    // 256bit, needs GCC-4.6 + k2.6.30 (AVX task switch!)
    // only low128bit for integer usable!  2*get_symconfig speed
    //  __m256, __m256d und __m256i   _mm256_*  = registers ymm0...15
    // Intel-Sandy-Bridge-Mikroarchitektur 32nm Jan2011
    // Intel-Ivy-Bridge-Mikroarchitektur   22nm Apr2012 +RDRAND f16 cfgTDP
    // AMD-Bulldozer 2011Q4, Piledriver 12Q4, Jaguar 13Q2, Steamroller 14Q1
#endif
#ifdef __AVX2__  // 4*f64 4*i64(!?) 8*i32(!) 32*i8  haswell-2013 noAMD2014
     " AVX2"     // 256bit-SSE-iOPs, 256bit for integers, +FMA3(3register)
     // VPSLLVD,VPSLLVQ shift left logical, variable shifts for each word
     //   ToDo: lNv-code uses variable-shift? scaling?
     // 3Op-3bit-table-logic  
     // /usr/lib/gcc/x86_64-linux-gnu/4.7.2/include/avx2intrin.h
     // Intel-Haswell-Mikroarchitektur 22nm Jun2013 (vec=Ivy+8%, single+6%)
     // gcc -march=haswell -dM -E -x c /dev/null # __haswell__
     // haswell E5-2630v3 2(FMA) *4(256bitAVX/64bitFloat) *2(ALUs?)= 16FLOP
#endif
#ifdef __AVX512F__  // ??? 8*f64  gcc-4.9 -mavx512f  Xeon-Phi no256bitFMA?
     " AVX-512F"     // AVX512F: only 32bit/64bit-vectors, register zmm0..31
     // + k0..k7 mask registers
     //  VPERMT2PD, VPERMT2Q  Full 64-bit permute overwriting first source.
     //                         Todo: try it !!!
     //  VPTERNLOGD, VPTERNLOGQ Bitwise Ternary Logic (512*3bit to 8bit-table)
     //              incl. 0x80=dblAND,0xFE=dblOR,0xca=BitwiseBlend...XOR
     //             In: 3 Regs + 1 8bit-table, Out: 1 Reg
     //             ~i1 | i2 | i3 made in one step = 3OPs/cycle for benes
     // ToDo: 128bit = (PLH64|PHH64)<<64 | (PLL64|PHL64) = 4P64+2
     //              = 4(4P32+2)+2 = 16P32+10 = 64P16+42 = 256P8+170 = bad
     // __AVX512BW__ byte/word extensions
     // __AVX512CD__ Conflict Detection to allow more loops to be vectorized,
     //              Knights Landing and Skylake Xeon
     // __AVX512PF__ Prefetch capabilities, Knights Landing
#endif
#ifdef __BMI__  //  bit manipulation instruction ANDN,BEXTR
     " BMI"
     // BEXTR bitfield extract with register x>>start & ((1<<len)-1)
     //    one clock instead (x>>i)&1
#endif
#ifdef __BMI2__  // -mbmi2 bit manipulation instruction 2 64bit pext+pdep
     " BMI2"     // ToDo: use it to speed up getsymcfg()
     // PDEP,PEXT parallel bits deposit + bits extract (ld(64)=6 bitperm)
// ToDo: BMI2: intel-haswell-2013 amd-excavator-2015(planed) gcc-4.8
//    Parallel bit deposit and extract  (for const. runtime bitperm)
//    6 times  pext(v,m)<<32 | pext(v,~m)   6*14cycles ? vs 64*4cycles = 1:3
//   what about VBPERM 32*1B-Perm. (256bit)
// unsigned __int64 _pext_u64(unsigned __int64 src, unsigned __int64 mask);
//   see http://en.wikipedia.org/wiki/Bit_Manipulation_Instruction_Sets
//  /usr/lib/gcc/x86_64-linux-gnu/4.7.2/include/bmi2intrin.h
#endif
#ifdef __3dNOW__  /* mm3dnow.h AMD only, obsolete */
     " 3dNOW"   // k8  mmx+2f32  PFADD, PFMUL, PAND=x&y(2*64bit)
#endif     
#ifdef __XOP__ //  AMD2009 (also SSE5)  opcode 0x8f  include/x86intrin.h
     " XOP"    // VPCMOV = Vector conditional move
               // VPPERM = Packed Permute Byte (select byte-list of 2 regs)
               // VPSHLQ  Packed Shift Logical Quadwords 64bit +/-reg !!!
     // http://en.wikipedia.org/wiki/XOP_instruction_set 0x8f+3byte
#endif
// ToDo: CVT16 = F16C 16bit-float-support, f16/f32-conversion only AMD 2009
#ifdef __MIPSEB__
     " MIPS_big-endian"
#endif
#ifdef __MIPSEL__
     " MIPS_little-endian"
#endif
/* sometimes the binary is copied, not tuned for the platform, debug: */
/* gcc -dM -E - < /dev/null */
// gcc492 -mtune=native -dM -E -x c /dev/null | grep tune
#ifdef __tune_atom__  // gcc 4.7.2 atom 2008 N455-1.66GHz L2=512KB 2HT 6.5W
     " ATOM"         // N455.FSB=200MHz L1d=24kB delay=3cycles SSSE3
    // in-order + 16-stage-pipeline (cant change cmds in pipeline)
    // 2iALUs = ALU^SHFT + ALU (intel_atom_cpu*.pdf)
#endif
#ifdef __tune_core2__  // intel core2 gcc4.9.2  penryn-uarch? core-arch? 2007/2008
     // ToDo: ri2015 i5-3470 gcc472 no _tune!
     " CORE2"     // +SSE4.1
     // https://en.wikipedia.org/wiki/File:Intel_Core2_arch.svg 
     // 3ports=(ALU^FMUL/FDIV128+SSE_SHFL/ALU)
     //       +(ALU^FADD128^SSE_SHFL/MUL)
     //       +(ALU^SSE_ALU)
     // eff064=2(ALU^SHFL)+ALU      (i64)    shift?
     // eff128=1(ALU^SHFL)+SHFL+ALU (v2i64)  vshift?
     //   shuffle: reorder bytes within register (since penryn) ToDo: use?
     // J-PC-2008 E5400 2009-2011 core=Wolfdale-3M
#endif
#ifdef __tune_corei7__  
      // corei7     2009 SSE4.2
      // corei7-avx      SSE4.2+AVX+PCLMUL gcc472
      // core-avx-i      SSE4.2+AVX+PCLMUL+RDRAND+F16C gcc472 ri15.i5-3470
     " COREi7"       // advantages? +SSE4.2? more cores/clock/L2 than i3, i5
     //  (i7 for all uarchs=nehalem...haswell = business-version)
#endif
#ifdef __tune_sandybridge__  // intel sandy-bridge gcc 4.9.2 (succ_of_nehalem)
     " SandyBridge"  // advantages? AVX=256bit-FP
     // 3ports=(ALU^LEA^SHIFT^FMUL256^SIMD_VMUL/VSHIFT128)
     //       +(ALU^LEA^MUL^SIMD_VALU/VSHUF128)
     //       +(ALU^SHIFT^SHFL256^SIMD_VALU/VSHUF128)
     //  eff064=2(ALU^SHIFT)+1ALU
     //  eff128=1SHIFT+2ALU (=SIMD)
     //  PMOVSMKB helpful?
     // http://www.intel.com/content/dam/doc/manual/64-ia-32-architectures-optimization-manual.pdf
#endif
#ifdef __tune_haswell__ // haswell 2014/2015 boost (succ_of_sandybridge)
     " HASWELL"   // AVX2 = AVX + i64-SIMD256, BMI2
     // 4Aports=(ALU^SHFT^FMA256^VMUL^VSHFT)
     //        +(ALU^MUL^FMA256^VALU) +(ALU^FSHFL256^VALU^VShuffle)
     //        +(ALU^SHFT)
     // i64:  2(ALU^SHFT) + 2ALU
     // v256: VSHFT + 2VALU + i64(ALU^SHFT)
     // not in gcc4.8.3
#endif
#ifdef __tune_core_avx2__ // gcc-4.8.3 --march=core-avx2 # =haswell
    " core_avx2"
#endif
#ifdef __k8__  // AMD K8 SSE2-path 64bit only 128bit/2clks??? L2=128bit/clk
     " K8"   // L1=L2=128bit-bus L1 w=2*64bit 3ALUs64bit 2SSE-ALUs
     // 12-stage-pipeline (k7=10-stages)
     //   no exceptions explored running k10-binary on k8 (comp1)
     //   only bn8 is much slower! k8=28clks vs. k10=13clks
#endif
#ifdef __tune_amdfam10__   // AMD K10  2SSEread=2*128bit/clk = 4*f64/clk
     " K10"    // +LZCNT,POPCNT,EXTRQ,INSERTQ  L1=L2=256bit-bus L1=64k-2way
     // L1 w=2*64bit r=2*128bit  3ALUs64bit 2ALUs128bit(1FADD+1FMUL)
     //   gcc may generate same binary for k8 and k10
#endif
#ifdef _MIPS_ARCH   // Sicortex-MIPS64 "5kf" (MIPS incl. FP+caches) 2009
     " " _MIPS_ARCH  // string 
     //  MIPS-III = 64bit = R4k-series
     //   Power challenge R10k 1996 super-scalar + out-of-order-exec
     //  MIPS-IV R8000 2ALUs +1SHIFT-Unit (shared? alu^shift?)
     //           +1Mul/Div(not pipelined,4clk/MUL32,6/MUL64,DIV=21..73)
     //  nor: d=~(s|t)
     //   see http://en.wikipedia.org/wiki/MIPS_instruction_set + refs
     //  gcc412  -dM -E -x c /dev/null | grep MIPS # == pathcc
     //   _MIPS_ARCH="5kf" _MIPS_TUNE="5kf" _R4000=1 _ABI64=3 _MIPS_ARCH_5KF=1
     //   -mpaired-single __mips_paired_single_float=1 2f32 in f64-reg
     //   try: -mbranch-likely
#endif
#ifdef __SICORTEX__
     " SICORTEX"
#endif
#if defined(__LP64__) || defined(__x86_64__)  /* __WORDSIZE == 64 */
    " int64"
#endif
#if defined(__SIZEOF_INT128__)  /* __int128 available gcc4.7-4.9, ~gcc4.4 */
    " int128"
#endif
     "\n");
#if (!defined(__LP64__)) && (!defined(__x86_64__)) && NN*CM(1,2,2)>32
    printf("WARNING: NN>32 on a 32bit machine may be about 4 times slower\n");
#endif
  } /* !mpi_i */

/* we tested OpenMP very earlier, but performance was poor */
/* ToDok 2016 implementing pthread-structure using OpenMP-v3.tasks? */
#if defined(_OPENMP)  /* OpenMP (former used, but not now, later again?) */
  if(!mpi_i)printf(" OpenMP-Support:"
#if   _OPENMP >= 201511  /* YYYYMM of openmp-spec */
         " OMP-4.5" /* */
#elif _OPENMP >= 201307 
         " OMP-4.0" /* gcc 4.9 +simd+acceleraturs+... */
#elif _OPENMP >= 201107 
         " OMP-3.1" /* gcc 4.7 */
#elif _OPENMP >= 200805 
         " OMP-3.0" /* gcc 4.4  +tasks (omp task is needed 2017-01-08) */
#elif _OPENMP >= 200505  /* 2000=fortran 2002=cc/c++ 2005=f/c.combined */
         " OMP-2.5" /* gcc 4.2, icc 10 for+sections? (2017-01-09) */
         /* test g++4.1.2 warn only + exe is hanging 2017-01 */
#elif _OPENMP >= 199810  /* YYYYMM of openmp-spec */
         " OMP-1.0 (but need 2.5)" /* gcc 4.2, icc 10 */
#else /* 199810 = 1.0 */
         " unknown OMP version (but need 2.5)"
#endif /* _OPENMP values */
         /* icc10.1=200709 -openmp -E -dM _OPENMP=200505 = OMP 2.x? no task */
         /* icc10 -openmp-stubs = dummy-openmp lib, sequentiell, no macro */
         /* icc10 -cluster-openmp needs extra-LIC, 2010 discontinued */
         /* gcc4.1  _OPENMP=200505 */
      " OPENMP=%d\n",_OPENMP);
#endif /* _OPENMP defined */


/* 2016-02-25 see defined constants for MPI-1.1 MPI_SUCCESS=0 */
#if (!defined(CONFIG_MPI)) && (defined(OMPI_MAJOR_VERSION)\
      || defined(MPICH2) || defined(MPICH_VERSION) || defined(MPICH2_VERSION)\
      || defined(MPI_VERSION))
  /* ompi-1.8.4/mpi.h MPI_VERSION=3 */
  if(!mpi_i)printf("WARNING: MPI-Compiler detected, but no CONFIG_MPI set\n");
#endif

#ifdef CONFIG_MPI
  { int version, subversion;
    MPI_Get_version(&version,&subversion);
    if (!mpi_i) {
      printf("MPI:");
#if defined(MPI_SUBVERSION) /* 2016-02-25 seen in ompi-1.8.4/mpi.h */
      printf(" MPI-%d.%d",MPI_VERSION,MPI_SUBVERSION);
#elif defined(MPI_VERSION) /* +MPI_SUBVERSION */
      printf(" MPI-%d",MPI_VERSION);
#endif
#if defined(OMPI_MINOR_VERSION)  /* using stringification of macros */
      printf(" OMPI-%d.%d",OMPI_MAJOR_VERSION,OMPI_MINOR_VERSION);
#ifdef OMPI_RELEASE_VERSION
      printf(".%d",OMPI_RELEASE_VERSION);
#endif
#elif defined(MPICH_VERSION)
      printf(" MPICH-" MPICH_VERSION);
#elif defined(MPICH2_VERSION)
      printf(" MPICH2-" MPICH2_VERSION);
#elif defined(MPICH2)    /* sicortex pathscale mpicc-3.2n */
      printf(" MPICH2 (unknown)");   
#else
      printf(" non-OMPI/MPICH"); 
#endif
      printf(" using MPI-%d.%d\n", version, subversion);
      printf("MPI: MPI_MAX_NUM=%d  mpi_n_bits=%d  mpi_n= %d\n",
                   MPI_MAX_NUM, mpi_n_bits, mpi_n);
      /* 2017.01 moved here */
#if defined(MPI_THREAD_MULTIPLE)
      printf("MPI: MPI_THREAD_MULTIPLE supported\n");
#else
      printf("WARNING: MPI_THREAD_MULTIPLE not supported\n");
#endif
      printf("MPI: mpi_thread_level=%d (MPI_THREAD_%s) requested=%d \n",
        mpi_thread, mpi_t[mpi_thread&3], mpi_thread_req);
      if (mpi_thread != mpi_thread_req)
        printf("WARNING: provided != requested mpi_thread_level\n");
      printf("MPI: HNZ_MPI_BUFSIZE/NUM_AHEAD_LINES=%.3f  NUM_AHEAD_LINES=%d\n",
                HNZ_MPI_BUFSIZE*1./NUM_AHEAD_LINES, NUM_AHEAD_LINES);
#if USE_Alltoallv_Emulation /* useful on SiCortex 2009, 15x faster */
      printf("MPI: using Alltoallv_Emulation (only for bad scaling A2av)\n");
#else
      printf("MPI: using better scaling MPI_Alltoallv (since 2017-01-16)\n");
#endif
    }
  }
  if (mpi_n>MPI_MAX_NUM) {
    if (!mpi_i) printf("ERROR: please enlarge MPI_MAX_NUM"); 
    sleep(2);MPI_Abort(MPI_COMM_WORLD,-1); }
#endif

#if 0
  /* 2013-04-14 set to maximum checkpoint, from latest run */
  /*    may cause errors if l1[] is changed, so
   *    its better to set by hand to what is intended by the user (option)
   */
  if(!mpi_i){ /* read checkpoint state, may be replaced by option or ifile */
    FILE *f; int i=0;
    f=fopen(FCHKPT,"rt");
    if (f) {i=fgetc(f);
      if      (i>='0'&&i<='9') checkpoint=i-'0';
      else if (i>='a'&&i<='z') checkpoint=i-'a'+10;
      else i=0;
      fclose(f);
    }
    if (checkpoint)
      printf("WARNING: set checkpoint to %d from "FCHKPT"\n", checkpoint);
  }
#endif

#ifdef _OPENMP  /* override CONFIG_PTHREAD */
  if (!mpi_i) {  /* 2015-12 ToDo: add support, start on SH */
    int omp_n=1; // omp_set_num_threads(2);
    /* printf("DBG: compiled for OpenMP but unsupported until 2017-01\n"); */
    /* 2017-01-10 fixed mixed code MPI + OMP */
#pragma omp parallel
    omp_n= omp_get_num_threads();  /* must be called within parallel pragma */ 
    /* next line is outside parallel region */
    printf("DBG: OpenMP threads=%d\n", omp_n); /* added 2016-05-03 */
    pt_n=omp_n; /* set as default num threads */
  }
#ifndef CONFIG_PTHREAD
#error "CONFIG_PTHREAD is needed for _OPENMP (2017-01 but subject of change)"
#endif /* PTHREADS */
#endif /* _OPENMP */

  for (i=0;i<B_NUM;i++) hbuf[i]=NULL; /* better debugging 2015-07-08 */
  /* 2016-05-18 add checking OMP_NUM_THREADS, overwrite by -t */
  if (getenv("OMP_NUM_THREADS")) {
    pt_n=atoi(getenv("OMP_NUM_THREADS"));
    if (!mpi_i) printf(" set pt_n=$OMP_NUM_THREADS=%d\n", pt_n);
  }
  // Bcast_args: dat,sz,typ,snode,..
  MPI_Bcast(&checkpoint, 1, MPI_INT, 0, MPI_COMM_WORLD);
  i=0;while(++i<argc){ ss=argv[i];
    if (vvv && !mpi_i) printf("DBG: parse argument %d: %s\n",i,ss);
    if (ss[0]=='-'){
      if (ss[1]=='v'){ vvv=atoi(ss+2); continue;} /* verbose */
      if (ss[1]=='i' && ss[2]!=0) {name_file_i=ss+2; continue;} /* daten.i, ToDo: stdin */
      if (ss[1]=='m' && ss[2]!=0) {name_file_m=ss+2; continue;} /* daten.def */
      if (ss[1]=='o' && ss[2]!=0){ /* redirect stdout */
                      if (!mpi_i) /* avoid chaos (May08)  */
                       if (freopen(ss+2,"a",stdout)==NULL) E0("freopen");
                                         continue;}
      if (ss[1]=='t'){pt_n=atoi(ss+2);   continue;} /* threads per node */
    }
    /* short run, to get best estimation of performance 2017-01-13
     *   this is if CPUh are limited and we want to now how much it will eat */
    if (strncmp(ss,"--perf_test=",12)==0){
      opt_perf_test= (ULong) atof(ss+12); /* allow 1e5, 1e6 etc. */
      if (opt_perf_test && opt_perf_test<100000UL) opt_perf_test=1000000UL;
      if (opt_perf_test<2*NUM_AHEAD_LINES) opt_perf_test=2*NUM_AHEAD_LINES;
      if(!mpi_i)printf("DBG: set opt_perf_test=%lu\n",opt_perf_test);
      continue;}
    /* chkpt_mode: set checkpoint behavior, 0=no, 1=save, 2=restore */
    /* chkpt_load: set resume checkpoint, 0=start, 1=ns, 2=SH ?, 4=SH.it */
    if (strncmp(ss,"--chkpt_load=",13)==0){checkpoint=atoi(ss+13);continue;}
    if (strncmp(ss,"--chkpt_time=",13)==0){chkpt_time=atoi(ss+13);
      if(chkpt_time<15000 && !mpi_i) // 2014-09-19
        printf("WARNING: chkpt_time<25m is very small ...\n"
               " ... good choice is between 8x iteration time"
               " and 1/8 failure freq.\n");
      if(!mpi_i) /* 2015-02-21 */
        printf("DBG: int chkpt_time = checkpoint interval in seconds = %d\n",
         chkpt_time); 
      continue;
    }
//    if (strncmp(ss,"--chkpt_mode=",13)==0){chkpt_mode=atoi(ss+13);continue;}
    if (!mpi_i) printf("ERROR: unknown argument %d \"%s\"\n", i, ss);
    if (!mpi_i)
      printf("options: [-v<verbose>] [-i<daten.i>] [-m<daten.def>]"
             " [-o<outfile>] \n... [-t<threads_per_node>]"
             "\n... [--chkpt_load=4] [--chkpt_time=3600] [--perf_test=1e6]"
//             " [--chkpt_mode=3]"
             "\n");
    if (!mpi_i) /* 2016-05-18 added, grep getenv */
      printf("environment: OMP_NUM_THREADS\n");
    sleep(2); MPI_Abort(MPI_COMM_WORLD,1); return 1;
  } /* while ++i<argc */
  setvbuf(stdout, (char *)NULL, _IONBF, 0); /* no buffer */

  if (opt_perf_test==0 && !mpi_i) /* 2017-01-13 ToDo also set in daten.i?? */
    printf("INFO: use --perf_test to do a reduced matrix performance test\n");
  if (opt_perf_test!=0 && !mpi_i) /* 2017-01-13 ToDo also set in daten.i?? */
    printf("\n * * *\nWARNING: performance test only, no results!\n * * *\n");

  if (pt_n<1) pt_n=1;  /* silent correction to minimum 2014-06 */

#if defined(CONFIG_PTHREAD) && B_NUM==1
  if (!mpi_i)  /* 2014-06 */
    printf("WARNING: spinpack is compiled for pthreads"
           " but B_NUM=1 limits the tasks to 1\n");
  /* you should set B_NUM to 2 or higher if you want to use pthreads */
  /* 2016++ B_NUM can be set via CONFIG_PTHREAD set in Makefile */
#endif

  if (pt_n>B_NUM) {
      pt_n=B_NUM; /* set to maximum (1 if single-cpu code) */
    if (!mpi_i)
      printf("WARNING: number of threads is set back to maximum (B_NUM)\n");
  }
  for (pt_n_bits=0;pt_n_bits<32;pt_n_bits++)
     if (1<<pt_n_bits > pt_n-1) break;
  if (pt_n_bits==32) F0("pt_n do not fit into 32bits");

#if CFG_CPUSET < 2 /* ToDo17: set as default or set at runtime! */
  if (pt_n*mpi_n>1 && vvv && !mpi_i) printf( /* 2017.01 autoset bits */
    " please try option -DCFG_CPUSET=2 -D_GNU_SOURCE on multi-core systems\n"
           "   or use taskset, numactl, omplace, dplace to pinpoint tasks\n");
  /* ToDo: -openmp as replacement for pthread_create/start_parallel ? */
#endif
#if defined(CFG_CPUSET)
  // && !defined(CONFIG_PTHREAD)
  if (pt_n==1) /* why pt=1 only? here its for single-thread-code only */
  { //   see also main_pthread().sched_setaffinity()
   // this processor bounding is for the case if the jobsystem does not do it
   //   ToDo: define on runtime per taskset
   cpu_set_t cpuset; int ii;
   /* 2017-01-22 added more info, usually 1024, bigger than numcores */
   if (!mpi_i) printf("DBG_AF: cpuset[bits]=%u  CPU_SETSIZE=%u\n",
               (unsigned)(8*sizeof(cpu_set_t)), CPU_SETSIZE);
   /* 2016-12 comp2: 0x0000ffff on 16-core machine without any affinity set */
   if (sched_getaffinity(0, sizeof(cpu_set_t), &cpuset)) {
     if (mpi_i<8) printf("ERR_AF getaffinity failed for task %2d\n", mpi_i);
   }
   /* output 2 first and 1 last, hyperthread+8c: task0=0101 task1=0202 ... */
   if (mpi_i < 4 || mpi_i >= mpi_n-1) { /* pt=1 */
     ULong cpuset2=0L; int nbits=0;
     for(ii=0;ii<8*(int)sizeof(cpuset2);ii++)
       if(CPU_ISSET(ii,&cpuset)) { cpuset2|=1UL<<ii; nbits++; } /* bis 64bit */
     printf("DBG_AF: getaffinity for task %2d is cpuset=%08lx numbits=%d\n",
             mpi_i, cpuset2, nbits); /* bad for 64++ cores ??? */
   }
#if CFG_CPUSET > 1   /* 2017-01-30 added */
   { int  ncores=0, in=0; /* reduce mask to 1 bit */
     for (ii=0; ii<CPU_SETSIZE && ncores<mpi_n; ii++)
                                      if (CPU_ISSET(ii,&cpuset)) ncores++;
     for (ii=0; ii<CPU_SETSIZE; ii++) if (CPU_ISSET(ii,&cpuset)) {
         if (mpi_i%ncores==in) break;  // mycore = 0,1,2,...
         in++; }
     CPU_ZERO(&cpuset); CPU_SET(ii,&cpuset);
     usleep(10000); /* why? hold parallel outputs together */
     /* output 4 first and 4 last */
     if (mpi_i < 4 || mpi_i >= mpi_n-4) {
       printf("DBG_AF: setaffinity for task %2d to core %2d\n", mpi_i, ii);
     }
     // no sched_setaffinity on OSF1
     if (sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) {
       /* failes if ii is above real core numbers (2016-12 comp2 size=1024) */
       printf("ERR_AF: setaffinity for core(task=%2d)=%2d failed\n",
               mpi_i, ii); usleep(100000); perror("setaffinity");
     }
     if (mpi_i<ncores) usleep(10000*(mpi_n-mpi_i));
   }
#endif /* CFG_CPUSET > 1 */
  }
#endif /* CFG_CPUSET defined */   

  /* give vectors blocks different names for shared filesystems */
  m_v0.p=NULL; m_v0.l=0; m_v0.mode=0; m_v0.ff=NULL;
  sprintf(m_v0.name,"tmp/v0_%04d.map", mpi_i);
  m_v1.p=NULL; m_v1.l=0; m_v1.mode=0; m_v1.ff=NULL;
  sprintf(m_v1.name,"tmp/v1_%04d.map", mpi_i);

  v0=v1=(tvec *)NULL;	/* NULL = not allocated */
  for(i=0;i<NEV;i++){
    m_ev[i].p=NULL; m_ev[i].l=0; m_ev[i].mode=0; m_ev[i].ff=NULL;
    sprintf(m_ev[i].name,"tmp/ev%d_%04d.map", i, mpi_i);
    ev[i]=(tvec *)NULL;
  }

  l2b=NULL; l1=NULL; /* initialization */
  iv=0; algo=0; x_p=3; n1=0l; nn=NN; nw=0*Nw; ne0=0;
  sym.subl=0; /* if 0 auto detection will be used */
  /* hfmax=1000000l/(Hsize/HBLen); */  /* max. 1MB filesize hxy.tmp */
  hfmax=0/(Hsize/HBLen);	/* no limit for hxy.tmp size 2015-01 */
  x_max=5; if(MAXx<x_max) x_max=MAXx;   /* max. num. of output-configs */
  /* bug: x_max=min(5,MAXx); - this gives warnings on SGI-CC, JS Mar99 */
  pew=2; nev=min(pew,NEV); /* added 2016-05 */ cfg_sisj=1;
  itime();	/* init itime, maxmem=0(infty, removed 2013) */
  /* output some info for possible bug reports */

  ini_handler(mpi_i, vvv); /* signalhandler (sighand.h) after pthread_create? */

#ifdef CONFIG_PTHREAD
  /* parallel but share everything exxcept local variables */
  if (!mpi_i)
   printf("PTHREAD max=B_NUM=%d threads/mpi-task, pt_bits=%d  pt_n= %d\n"
          "PTHREAD use option -t<threads> [-v4]\n", B_NUM, pt_n_bits, pt_n);
  MPI_Barrier(MPI_COMM_WORLD); /* sync "I am here" output */
  /* ToDo17: OPENMP-alternative to pthread? define pthread-dummies? */
#ifndef _OPENMP
  /* IRIX, True64 and Linux works with NULL=default, but SunOS does not */
  pthread_attr_init (&pt_attr);
  pthread_attr_setscope (&pt_attr, PTHREAD_SCOPE_SYSTEM);
  pthread_mutex_init(&pt_mutex1,NULL);
  pthread_mutex_init(&pt_mutex2,NULL);
  pthread_cond_init(&pt_cond1,NULL); /* correct init? */
  pthread_cond_init(&pt_cond2,NULL); /* correct init? */
#endif
// ToDo: #pragma parallel (mother-thread) + #pragma openmp for (child-threads)
// #pragma omp parallel num_threads(pt_n+1)
// {  
//    #pragma omp for schedule(static,1)
//    for (0..pt_n) { 
//       if (i>0) pthread-main-thread();
//       if (i=0) { // master_thread
//    ......
//    }} // close master + omp_for
// }
// // end-of-PG
// --OR-- tested v2017-01:
//#pragma omp parallel num_threads(pt_n+1) // starts next blk on all threads
//{
//#pragma omp master // master=thread0.single, single=ok? after for?
//{
//  for (i=0; i<pt_n; i++) {
//#pragma omp task  // needs OpenMP v3
//    worker(i);
//  }  // for
//  ....
//}} // end master + parallel
// --OR-- untested v2017-01:
//#pragma omp parallel num_threads(pt_n+1) // starts next blk on all threads
//{
//  for (i=0; i<pt_n; i++) {
//#pragma omp single nowait // needs OpenMP v2.5 ???
//    worker(i);
//  }  // for
//#pragma omp single nowait // also: omp master as thread0.single
//{
//  ....
//}} // end master + parallel
  /* create main child-threads */
// pragma parallel opens all threads and starts them
//#pragma omp parallel for schedule(static,1)
#pragma omp parallel private(i) shared(main_pt) num_threads(pt_n+1)
{
//#pragma omp master /* master = single where thread=0 but no nest = OK */
//#pragma omp single /* single like master = OK */
//#pragma omp task
//{   // moved below! 2017-01++
 int i; /* 2017 needed for OMP */
// #pragma omp for schedule(static,1) private(i)
  for (i=0;i<pt_n;i++) {
#ifndef _OPENMP
    pthread_mutex_init(&token_ring[i       ],NULL);
    pthread_mutex_init(&token_ring[i+  pt_n],NULL);
    pthread_mutex_init(&token_ring[i+2*pt_n],NULL);
    pthread_mutex_init(&token_ring[i+3*pt_n],NULL);
    pthread_mutex_lock(&token_ring[i       ]);
    pthread_mutex_lock(&token_ring[i+  pt_n]);
    pthread_mutex_lock(&token_ring[i+2*pt_n]);
    pthread_mutex_lock(&token_ring[i+3*pt_n]);
#endif
//#pragma omp task  /* needs OMP-3 or higher */
#pragma omp single nowait /* needs OMP-2.5 only !? */
{
#pragma omp critical
{
    main_pt[i].num=i;     /* put thread number to child */
    main_pt[i].data=NULL; /* for first do nothing */
    main_pt[i].task=0;    /* initialize */
} /* pragma omp critical */
#ifdef _OPENMP
    if (i<8 && !mpi_i) printf("DBG_OMP: create thread %d\n",i);
    main_pthread((void *)&(main_pt[i]));  /* child = worker thread */
#else
    if( pthread_create((pthread_t *)&(pt[i]),
                       (const pthread_attr_t *)&pt_attr,
                       (void * (*)(void *)) &main_pthread,
                       (void *)  &(main_pt[i]) ) !=0 )  F0("pthread_create");
#endif
} /* pragma omp task */
  } // for i=0...
#pragma omp single
{ // master thread moved here 2017-01++ 
  usleep(200000); /* DBG_OMP better readable if ordered */
  if (!mpi_i) printf("DBG_OMP: master still alive, go on ...\n");
  if (vvv&4) // reduce DBG output
  { /* coordinate the output of "I am here!" messages */
    int ii;
    for (ii=0; ii<pt_n; ii++)
      parallel_start(ii, test_pthread, (void *)&main_pt[ii]);
    for (ii=0; ii<pt_n; ii++)
      parallel_wait(ii);
  }
#ifndef _OPENMP
  pthread_mutex_unlock(&token_ring[0     ]); /* put token into the ring1 */
  pthread_mutex_unlock(&token_ring[2*pt_n]); /* put token into the ring2 */
#endif
#else  /* ifdef CONFIG_PTHREAD */
  if (!mpi_i)
   printf("PTHREAD _not_ used, %d thread (per node, max=B_NUM=%d, bits=%d)\n",
      pt_n, B_NUM, pt_n_bits);
  if (pt_n!=1) {
    pt_n=1; pt_n_bits=0;
    if (!mpi_i) printf("Warning: reset number of threads to 1\n");
  }
#endif
  /* BY_TYPE is short (Jun08) and can hold 65536 threads */
  if (((BY_TYPE)pt_n*(BY_TYPE)mpi_n)!=pt_n*mpi_n)
     printf("\nWARNING: pt_n*mpi_n dont fit into BY_TYPE!\n\n");
  if (((((BY_TYPE)mpi_n)<<pt_n_bits)>>pt_n_bits)!=mpi_n)
     printf("\nWARNING: mpi_n<<pt_n_bits dont fit into BY_TYPE!\n\n");
  MPI_Barrier(MPI_COMM_WORLD); /* sync "I am here" messages */
  if (!mpi_i) {
    printf(
    " using Ham1=%d Ham2=%d Ham3=%d VecType=%d L1_PACKED=%d Tnorm2=%3s%d\n",
     Ham1,Ham2,Ham3,VecType,L1_PACKED, /* L1=tbase_vector */
     ((((Tnorm2)1)/2==0)?"int":"flt"),8*(int)sizeof(Tnorm2));
    printf(" h_file.h: TYPE_SIZE XY=%d BY=%d HR=%d in Byte, Hsize=%.0fkB\n",
  (int)sizeof(XY_TYPE),(int)sizeof(BY_TYPE),(int)sizeof(HR_TYPE),Hsize/1024.);
#ifdef CKPT_HELP
    printf("using CKPT_HELP (pkill -SIGUSR1 spin  for safe ckpt)\n");
#endif
#ifdef HAVE_ZLIB
    printf("using ZLIB\n");
#endif
#ifndef ReserveMB
#ifdef  OMPI_MINOR_VERSION
#define ReserveMB 150 /* comp2 22*16cores=142MB+4MB ompi-1.8 ToDo: checkit! */
#elif   CONFIG_MPI
#define ReserveMB 26  /* kautz=5800cores/18*64kB=21MB+4MB(checkpoint) mpich2 */
#else
#define ReserveMB 5   /* 4MB(checkpoint)+1MB(local data) 2015-06-10 */
#endif /* OMPI, MPI */ 
#endif /* ReserveMB */
#ifdef _Vlint_h
    printf("WARNING: using class Vlint for %d-bit emulation"
           " (about 10 times slower)\n", (int)(sizeof(vlint)*8));
#endif
    printf("use symmetric base configurations (mandatory)\n");
#ifdef CONFIG_S1SYM
    /* ToDo: replace by set of commutating symmetries */
    printf("S1SYM: use special S1-symmetric base configurations (default)\n");
#else
    printf("NOS1SYM: no special S1-symmetries (slow on S1-models, debug)\n");
#endif
#ifdef AddSS
    printf("AddSS is on (full matrix, makes the program much more slow)\n");
#endif
#ifdef DEBUG
    printf("DEBUG is on (more checks, makes the program slower)\n");
#endif
    if (((BY_TYPE)(mpi_n*pt_n))!=mpi_n*pt_n) {
      printf(" num_of_nodes * num_of_threads dont fit to BY_TYPE\n");
      ERR(9999);
    }
  } /* if !mpi_i */

  /* create the child early to save memory on pipe or fork */
  MPI_Barrier(MPI_COMM_WORLD);

  sub_ini();		/* v1.6.1 start child for gzip, ini pipeHbuffer */
  /* relax nfs, use cache 2016-03-16 kautz 5000 tasks may cause OOM */
  if (mpi_i) usleep(mpi_i/512*10000+100000); /* 5120 tasks = 0.2s */

  /* output helpful configuration for support/debug,
   * moved from after modeldef..ini_sym.{k,w,l} to before modeldef 2016-04-06
   * add Ahead */
  if (!mpi_i) {
#ifdef CplxInfo
    printf("CplxInfo: " CplxInfo "\n"); /* compiler specifique cplx def */
#endif
    printf(" long=%d int=%d  double=%d sdouble=%d  cplx=%d scplx=%d\n"
           "  hxy=%d tbase1=%d tbase=%d",
    (int)Lsize, (int)sizeof(int), (int)sizeof(double), (int)sizeof(sdouble),
                                  (int)sizeof(cplx),   (int)sizeof(scplx),
    (int)Hsize/HBLen, (int)sizeof(tbase1), (int)sizeof(tbase));
    /* added 2016-04-06 for debugging/support */
    printf(" NumAheadLines=%d\n",NUM_AHEAD_LINES);
  }
  if (sizeof(int)  != S_int) {
      printf ("Error: sizeof(int)   != S_int\n");    ERR(9999); } 
  if (sizeof(long) != S_long) {
      printf ("Error: S_long  != sizeof(long)\n");   ERR(9999); } 
            

  if (initvars(mpi_i, name_file_m)) { /* modeldef.c:read(daten.def) */
    E0("load model failed, sleep8s + 2nd try");sleep(8);
    if (initvars(mpi_i, name_file_m)) F1("load model from %s",name_file_m);
  }
  initmasks(nn);	/* initialize bitmasks for lintab/sym/etc. */
  p_mask_ini(nn);       /* initialize bitmasks for parity() */

  // MPI_Barrier(MPI_COMM_WORLD);  // removed 2016-03-16

  /* initialize ud-symmetry data, HFIELD or TBC? */
  sym.wud=+1;           /* initialize symmetry data */
#ifdef Sud
# error "macro Sud removed, switch off sym_ud by setting sym.wud=sym_ud=0"
#endif
#if defined(HFIELD) || defined(CONFIG_TBC) /* reset sym.wud */
  sym.wud=0;		/* do not use ud-sym */
#endif     /* reset sym.wud */
  
  for (i=0;i<MaxSym;i++){ 
    sym.k[i]=0;  /* phase factor in units of 2pi/l */
    sym.w[i]=+1; /* sign = exp(i*2pi*k/l) */
    sym.l[i]=1;  /* possible phase quantum numbers old2015: 2 */
    for (j=0;j<NN;j++) sym.i[i][j]=j; /* no permutation P=1 */
  }

  /* check bond data for errors */
  for (i=0;i<2*nw;i++) if(iw[i>>1][i&1]>=nn) F1("site>=NN iw[%u]",i>>1);
  /* initialize phase parameters */
  for (i=0;i<nw;i++){  ww.ex[i]=1;
    if (iw[i][0]>iw[i][1]){		/* lowest index first */ 
      j=iw[i][0]; iw[i][0]=iw[i][1]; iw[i][1]=j;
      ww.si[i]=-ww.si[i]; ww.ex[i]=-1;
    }
  }

  MPI_Barrier(MPI_COMM_WORLD);
  /* relax nfs, use cache 2016-03-16 kautz 5000 tasks may cause OOM */
  if (mpi_i) usleep(mpi_i/512*10000+100000);
  
  in1=fopen(name_file_i,"rt"); 
  /* fopen may fail after edit nfs-file on another node 2016-01 t100 */
  if (in1==NULL) { 
    if (!mpi_i) E1("ropen \"%s\", sleep 8s + 2nd try",name_file_i);
    sleep(8); in1=fopen(name_file_i,"rt"); } /* try again */
  if(in1==NULL) {
    if (!mpi_i) E1("ropen \"%s\", example doc/daten.i",name_file_i);
    MPI_Abort(MPI_COMM_WORLD,1);
  }

  paini();  /* parameter initialization */
  sprintf(FILE_l1, "tmp/l1_%04d.dat", mpi_i); /* initialize the name */

  /* alloc thread local buffers (not during iterations (old)!)  2014-07-17 */
  if (pt_n > B_NUM) F0("pt_n > B_NUM");
  for (block=0; block<pt_n; block++) {
    hbuf[block] = alloc_hbuf(); /* better alloc local in thread? */
    if (!hbuf[block]) {
      F2("ERROR hamilton: alloc_hbuf failed, mpi=%d thread=%d", mpi_i, block); 
      exit(1);// 2014-07 continue;
    }
    hbuf[block]->block = block;
    hbuf[block]->v0 = NULL;
    hbuf[block]->v1 = NULL;
  }
  
  // ToDo: simpler syntax: [a-z]*[0-9]*= for setting parameters
  // ToDo:                 .[a-z]*[0-9]* for procedures
  for(;;){  /* load and execute input file */
    if (!in1) {
      if((in1=fopen(name_file_i,"rt"))==NULL) { /* was closed for better C/R */
        if (!mpi_i)
          printf("\nERROR: reopen \"%s\" failed\n",name_file_i);
        sleep(1); MPI_Abort(MPI_COMM_WORLD,1);
      }
      fseek(in1,in1pos,SEEK_SET); /* continue on old position */
    }
    if (feof(in1)){ break; } /* removed2012-09: restore .i file */
    s1[0]=0; if (fgets(s1,500,in1)==NULL) break; /* 2012-05 +err-handling */
    in1line++; // infile line number for errors/debugging output
    /* read without \n ?? , .begin ... .end */
    if(s1[0]=='#' || s1[0]=='%' || s1[0]==0 || s1[0]=='\n'){ /* nop */ } else
    if(s1[0]=='!'){ // ToDo: reopen in1 to allow infile-manipulations?
      /* the tmp/pid check dont work on MPI + network_filesystems */
      // 2012-05 printf to fputs (less lib-overhead)
      if (!mpi_i) { 
        // ToDo: produce pid by echo $$PPID | tee spins.pid
        //   for sending signals like CKPT_HELP_REQ
        if (fputs(s1,stdout)<=0) E0("main.scan.fputs");
        if (system((char *)&s1[1])) E0("main.system(s1+1)");/* exec arg */}
    } else
    if(!scan_int(s1,"max_ea",&max_ea)){max_ea=max(1,max_ea);} else
//    if(!scan_int(s1,"chkpt_mode",&chkpt_mode)){ } else
    if(!scan_int(s1,"nev",&nev)){nev=min(max(0,nev),NEV);} else
    if(!scan_int(s1,"ne0",&ne0)){ne0=min(max(0,ne0),NEW-1);} else
    if(!scan_int(s1,"pew",&pew)){pew=min(max(1,pew),NEW);} else
    if(!scan_int(s1,"startvec",&startvec)){} else
    if(!scan_int(s1,"chkpt_load",&checkpoint)){ } else
    if(!scan_int(s1,"chkpt_time",&chkpt_time)){} else // 2012-05-22 checkpointing
    if(!scan_int(s1,"sym_lm",&sym.type_lm)){ sym.type_lm &= 1; } else // 2016-01
    if(!scan_int(s1,"sym_ud",&sym.wud)){ if(abs(sym.wud)>1)ERR(9999);
#if defined(HFIELD) || defined(CONFIG_TBC)
      printf("\nError: sym_ud can not be used for fields and TBC\n\n");
      sym.wud=0;
#endif
    } else
    if(!scan_int(s1,"sym_k",NULL)){ /* in daten.k schreiben? */
      for(i=0;i<MaxSym;i++)sym.kgen[i]=0;
      scan_inta(s1,"sym_k",sym.kgen,MaxSym);
    } else
    if(scan_var(s1,"SL")){ /* set sublattice 0=A 1=B etc. */
      char *s2=scan_var(s1,"SL");
      printf(" set sublattice %d: ", sym.subl);
      for (i=0;i<nn;i++) {
        sym.sl[sym.subl][i]=((s2[i]&~0x20)-'A')%nn;
        printf("%c",'A'+sym.sl[sym.subl][i]);
      } printf("\n");
      /* old: scan_inta(s1,"SL",sym.sl[sym.subl],nn); */
      sym.subl++; sym.subl%=MaxNumSL; /* handle overflow as reset */
    } else
    /* removed2013: if(!scan_dbl(s1,"maxmem" ,&r1)){ maxmem=(ULong) r1;} else */
    if(!scan_dbl(s1,"maxfile",&r1)){ hfmax =(ULong)(r1*HBLen/Hsize);
      if( (double)hfmax+1<r1*HBLen/Hsize ) ERR(612);
      } else
    if(!scan_dbl(s1,"xy_anisotropy+",&r1)){ h_xypara+=r1;} else
    if(!scan_dbl(s1,"z_anisotropy+", &r1)){ h_zpara+=r1; } else
    if(!scan_dbl(s1,"xy_anisotropy", &r1)){ h_xypara=r1; } else
    if(!scan_dbl(s1,"z_anisotropy",  &r1)){ h_zpara=r1;  } else
    if(!scan_int(s1,"verbose",  &vvv)){} else
    if(!scan_int(s1,"sisj",&cfg_sisj)){} else
    if( (ss=scan_var(s1,"nud+"  ))!=NULL ){ scan_ud(ss,1);   } else
    if( (ss=scan_var(s1,"param+"))!=NULL ){ scan_para(ss,1); } else
    if( (ss=scan_var(s1,"nud"   ))!=NULL ){ scan_ud(ss,0);   } else
    if( (ss=scan_var(s1,"param" ))!=NULL ){ scan_para(ss,0); } else
    /* allow "p1=-1.0", "p1= -1.0" or " p1= -1.0" as input per line */
    if (sscanf(s1," p%d= %lf",&i1,&r1)==2) {
        para[i1]=r1; printf("set parameter p%d= %f\n",i1,r1); } else
 // if(s1[0]=='x'){ sscanf(s1,"%*c%d%lu",&x_p,&x_max); } else /* ToDo: replace by function xout= */
    if((ss=scan_var(s1,"xout"))!=NULL){ sscanf(ss,"%d%lu",&x_p,&x_max); } else
 // if(s1[0]=='l'){ loadvec(ev[0],s1+1,-1,-1); } else
    if((ss=scan_var(s1,"loadvec"))!=NULL){ loadvec(ev[0],ss,-1,-1); } else
    if(strncmp(s1,"newtime",7)==0){starttime=0l;itime();} else	/* reset clock */
    if(strncmp(s1,".out",4)==0){if(!mpi_i)outpara(stdout);} else
    if(strncmp(s1,"exit",4)==0){exit(0);} else /* 2015-09-08 for testing */
    if(strncmp(s1,"quit",4)==0){exit(0);} else /* 2015-09-08 for testing */
    // removed2012-10: pout could also be done by !echo "text"
    if(s1[0]=='a' && s1[1]>='0' &&  s1[1]<='9'){ /* add 0-9 check 2012-10 */
      sscanf(s1,"%*c%d%lu",&algo,&n1max);
      if (!mpi_i) outpara(stdout);
      /* 2015-02-26: getfilepos + close + work(+C/R) + reopen in1 */
      in1pos= ftell(in1); fclose(in1); in1=NULL; /* closed for simpler C/R */
      
      MPI_Barrier(MPI_COMM_WORLD); /* debug */
      /* --- test and if neccessary create Sym symmetries P --- */
      symcreate(mpi_i);
      sym_bn_m_ini_flag=1;
      sym_bn_m_ini(0); /* 2015-09-22 was buggy for 2nd sym_k */

      MPI_Barrier(MPI_COMM_WORLD); /* debug */

      if (vvv&1) testsym();

      MPI_Barrier(MPI_COMM_WORLD);

      if (!mpi_i) { /* output on node 0 */
        printf("\n nn,nu,nd %d %d %d   nw= %d"
              "  (number of sites, up/down-spins and bounds)\n",nn,nu,nd,nw);
        /* wud=(+1,-1, 0) to kud=(0,1,* old=-1), wud=exp(kud*I*Pi) */
        if (sym.wud && nu==nd) printf(" k_ud %d", 1-(sym.wud+1)/2);
                          else printf(" k_ud *"); /* for the eyes */
        printf(" k_T ");
        for (i=0;i<sym.numsym;i++) if (sym.g[i]) printf("_%01d",sym.k[i]);
#if     CM(1,0,0)
        sym.type_lm=0; /* 1 not possible in JJ model */
#else
        /* 2016-01-16 LM-type 0=maxS(S+1)  1=maxJ(J+1) */
        if (sym.num_lm != nn) printf(" type_lm %d", sym.type_lm);
#endif
        /* ToDo: move output of parameters to a function? */
        printf("\n parameters ");
        //  for (j=0;j<nw;j++){ printf("t[%d]=%d ",j,ww.wt[j]); }
        for(i=1;i<numpara;i++){ int j,k=0;
          /* output human readable names, ex: J1=J2= 0.3 t1= -0.7 ... */
          for (j=0;j<nw;j++) if (ww.wj[j]==i) {printf("J%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.wjz[j]==i) {printf("Jz%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.wt[j]==i) {printf("t%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.w[j]==i)  {printf("w%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.wv[j]==i) {printf("V%d=",(k=i)); break;}
          for (j=0;j<nw;j++) if (ww.wh[j]==i) {printf("h%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.we[j]==i) {printf("e%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.wu[j]==i) {printf("U%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.wf[j]==i) {printf("f%d=",(k=i)); break;}
          for (j=0;j<nn;j++) if (ww.wd[j]==i) {printf("D%d=",(k=i)); break;}
          if (!k) printf("p%d=",i); /* unused or unknown parameter */
          printf(" %g  ",(double)para[i]);
        } printf("\n");
        printf(" Bsize=%d (size of each basecfg, NN/8[Byte]), tbase[bits]=%d\n",
               (int)Bsize, (int)sizeof(tbase)*8);
        // printf(" Bsize=%d (size of each basecfg, NN/8[Byte]) t+=%s\n",(int)Bsize,sclock(~0l));
      }

      /* early speed tests for better tuning: 2014-05-26 speed_mpi.gpl */
      // 2015-03-26 add pthread too see cache coherence effects
      // 2015-07-07 AHEAD=2K segfault (meggie, ulimit -s=10240, 10240000)
      // 2016-11-04 segfault if ulimit -s 18000 (stacksize) to small
      // 2016-12-20 hangs on LRZ.icpc B_NUM=128 -D CFG_CPUSET=0xffffffff -t32
      // ToDo replace by omp-version? gcc only 600% (no pinning, t100?)
      for(block=0; block<pt_n; block++)
              /* hbuf[] + .block is be initialized above */
              parallel_start(block, speed_tests, hbuf[block]);
      for(block=0; block<pt_n; block++) parallel_wait(block);
      /* ompi will hang on MPI_Barrier here on oversubscription above 16 */
      /*  also using --mca mpi_yield_when_idle 1 --mca mpi_preconnect_mpi 1 */
      /* MPI_Barrier(MPI_COMM_WORLD); */ /* removed for oversubscr. 2016-03 */
      /* ri2016 tri96+4 4c: 64.bnV4=510ns SH=0.02m 128.bn4=1000ns SH=0.03m
                           256.bnV4=1850ns SH=crash 255,257=ok 256=crash
                   98+2 4c: 8-9=ok 10,16,32=crash 10:n1=51=8*6+1*3+1*0 */
      
      /* ===================== new-method l1+H storage ================ */
      /* generate hilbert space+matrix by applying hilbert Op
         on single configs iterative, ToDo: check speed */	
      if (algo&8) {  // ToDo: needs to review this code (buggy for sure)
        if (mpi_n!=1 || pt_n!=1) { 
          printf("WARNING: (mpi_n!=1 || pt_n!=1) is untested, please report\n");
        } /* not tested, ToDo */
        n1=(n1max)?n1max:100000;	/* n1max<=n1 variiere Basis! */
        if (((XY_TYPE)(n1/(mpi_n*pt_n)))!=n1/(mpi_n*pt_n)) {
          printf(" block_n1 dont fit to XY_TYPE, see h_file.h\n");
          ERR(9999);
        }
        for (n1_bits=0;n1_bits<(int)sizeof(ULong)*8;n1_bits++) 
           if (((ULong)1)<<n1_bits > n1-1) break;
        b_ini(n1,vvv);          /* init b_len[] */
        if((algo&7)<2){
          /* malloc by thread 0, first touched by working threads for ccNUMA */
          if ((!mpi_i) && (!v0)) printf(" a8.malloc v0,v1 ...\n");
          v0=(tvec *)mymalloc(b_ofs[pt_n]*Csize,&m_v0);
          v1=(tvec *)mymalloc(b_ofs[pt_n]*Csize,&m_v1);
          if ((!mpi_i) && ((!v0) || (!v1)))
            F0(" malloc failed, consider mmap?");
          touchvec(v0); // 2015-03-09 touch before storeH eats local memory
          touchvec(v1);
        }
        {
          if (1.*(node_n1*(off_t)Bsize)!=node_n1*1.*Bsize)
            printf("\nERROR mymap(n1*Bsize): off_t overflow");
          fl1=fopen(FILE_l1,"r+b");
          if (!fl1) fl1=fopen(FILE_l1,"w+b");
          if (!fl1) F1("fopen %s",FILE_l1);
          l1 = (unsigned char *)
                mymap(FILE_l1, fl1, node_n1 *(off_t)Bsize,
                                    node_ofs*(off_t)Bsize);
          fclose(fl1); fl1=NULL;  /* 2015-02-24 close for BLCR, mmap ok? */
          if (node_n1 && !l1) {
            E1("mymap(l1) failed for mpi=%d, skipping", mpi_i);
#ifdef CONFIG_MPI
            if (mpi_n>1) MPI_Abort(MPI_COMM_WORLD,1);
#endif
            continue;
          }
        }
        for(i=0;i<nev;i++){
          int vvv2=vvv; if (pt_n*mpi_n>4) vvv&=~2; /* avoid to much output */
          if ((!mpi_i) && (!ev[i])) printf(" malloc ev%d", i);
          ev[i]=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_ev[i]);
          vvv=vvv2;
        }
// sprintf(m_n2.name, "tmp/n2_%04d.map", mpi_i);
// norm2=(ATnorm2 *)mymalloc(node_n1*sizeof(ATnorm2),&m_n2);base2ok=0;
        l2b  =  (ULong *)mymalloc(node_n1*Lsize,&m_l2); /* hilfsarray2 */
        n1max=node_n1;       /* save memory block len before n1 changed */
        if (mpi_n!=1 && !mpi_i) E0("storeh2 will fail for mpi");
        storeh2(0, hbuf[0]);	/* ??? provisorium - testphase */
        for (n1_bits=0;n1_bits<(int)sizeof(ULong)*8;n1_bits++) 
           if (((ULong)1)<<n1_bits > n1-1) break;
        /* l2b=(ULong *)myfree((void *) l2b,&m_l2); / * zur Zeit noch!!! */
        randvec(v1,startvec,0);itr=inc1(nev,ne0,v0,v1,hbuf); /* EW(T),EV(T)  */
        randvec(v1,startvec,0);incv1(nev,v0,v1,ev,itr,hbuf); /* EV(H)   */
        /* repeat computation with new start vector */
        if(n1==n1max && nev>0)F0("Fatal, should not break here");
        if (nev) for(j=0;j<nev;j++){
          if (!mpi_i) printf("\n\nstart chk3.ev%d ...", j);
          chk3(ev[j],v1,0,hbuf); /* (inp_vec, work_vec, ?, workbuf) */
        }
        /* ToDo: reduce to / replace by better code structure:
         *   if (algo&8) { alloc } else { numsymconf + alloc }
         *   ...
         *   if (algo&8) { storeh2 } else { storeh }
         */
      } /* (algo&8)!=0  ====== experimental storeh2 (one-step-space+matrix) */
      if ((algo&8)==0){	/* ============ old-good-ns+storeh-method ========= */
        /* generating Hilbert space, and Hilbert matrix in two steps */

        // MPI_Barrier(MPI_COMM_WORLD);

        /* 2015-12-30 l1-generation only for high S, one task recursive */
        if((!mpi_i) && (algo&64)) printf("\nDBG: algo&64 stop after l1 ...");
        n1=(ULong)numsymconf();  /* slow, calculate n1, write l1 */
        /* ToDo17: store to memory or disk like matrix STOREH */
        /* ToDo: use l1 as argument, not global var, better code reading */
        if (opt_perf_test) { /* assuming speed/n1 is constant */
          ULong perf_n1=opt_perf_test; /* 1e6 = about size of sq40 9+31 */
          if (2*NUM_AHEAD_LINES > perf_n1) perf_n1= 2*NUM_AHEAD_LINES;
          if (n1<perf_n1) perf_n1=n1; /* not reduced */
          if (n1!=perf_n1 && !mpi_i)
            printf("\nWARNING: n1 reduced to %lu for perf_test"
                  " (check speed/n1)\n\n", perf_n1);
          n1=perf_n1; /* 1st idea, see if it works */
        }
        // checkpoint2 old=1
        if((!mpi_i) && (algo&64))
          printf("\nDBG: algo&64 stop, use --ckpt_load=2 to continue.");
        if (algo&64) MPI_Abort(MPI_COMM_WORLD,0);
        
        for (n1_bits=0;n1_bits<(int)sizeof(ULong)*8;n1_bits++) 
           if (((ULong)1)<<n1_bits > n1-1) break;
        
        if (!n1) { if (!mpi_i) E0("n1==0"); checkpoint=0; continue; }
        if (!mpi_i) {
          printf("\n N,nu,nd= %u  %u %u  n1= %lu  log2(n1)= %.1f",
                  nn, nu, nd, n1, log(n1)/log(2.0)); /* n1_bits */
          /* Tru64@alpha: log2(0) segfault */
          /* (long+sdouble*2)*n1/nodes, Dsize=sizeof(sdouble) */
          printf("\n MinMem %.2e", (Bsize+Dsize*(2+nev))*1.0*n1/mpi_n);
          printf(" userMaxHfile=hfmax*Hsize= %.2e (per node)\n",
                  (double)hfmax*(double)Hsize/HBLen);
        }
        /* we need a bit for the XY_FLAG, ToDo: change for MPI */
        if (((n1<<1)>>1)!=n1) { if (!mpi_i) E0("n1>max"); continue; }
        if (((XY_TYPE)(n1/mpi_n))!=n1/mpi_n) {
          printf(" node_n1 dont fit to XY_TYPE, see h_file.h\n");
          ERR(9999);
        }

        MPI_Barrier(MPI_COMM_WORLD);
        b_ini(n1,vvv);	/* block initialization (b_len[], b_ofs[] etc.) */
        MPI_Barrier(MPI_COMM_WORLD);

        /* --- l1 is now initialized by numsymconf --- */
        if (!l1) {
          if (!mpi_i) printf(" mymap(%s + node_ofs to *l1)\n",FILE_l10);
          /* Oct07: every node maps only its part of l1[]!
           * may be this is not necessary because only accessed 
           * parts loaded into mem? but if mmap failes, its loaded!
           */
          if (1.*(node_n1*(off_t)Bsize)!=node_n1*1.*Bsize)
            printf("ERROR mymap(n1*Bsize): off_t overflow\n");
// ToDo: may fail on cached NFS and fast write_task0/read_task3 cycle
// void *mymap(const char *nam, FILE *f1, off_t len, off_t ofs)
// 5700*parallel read is bad idea via nfs,
//        zuletzt kautz 4.3.2016 nfs-node-OOM nach 8Tagen alte version 2015-09
// 2014-07-17: use mpi token for coordinated disk access (buggy, still parallel)
/* ToDo: Zeit messen! double tm1=fstime() */
#ifdef CONFIG_MPI  // until 2016-01-20 token was nearly useless
          if (mpi_i>0) { int token=0; MPI_Status status;
            MPI_Recv(&token,1,MPI_INT,mpi_i-1,1,MPI_COMM_WORLD,&status);
            if (0==(mpi_i&(mpi_i-1))) printf(" l1_load.node=%d\n",mpi_i); }
#endif
          fl1=fopen(FILE_l10,"r+b"); /* 2016-01 moved behind token for nfs */
          if (!fl1) { F0("main.fopen_l1"); }
          l1 = (unsigned char *)
               mymap(FILE_l10, fl1, node_n1 *(off_t)Bsize,
                                    node_ofs*(off_t)Bsize);
          fclose(fl1); fl1=NULL;  /* 2015-02-24 close for BLCR, mmap ok? */
#ifdef CONFIG_MPI
          if (mpi_i<mpi_n-1) { int token=0;
            MPI_Send(&token,1,MPI_INT,mpi_i+1,1,MPI_COMM_WORLD); }
#endif
          // ToDo: malloc + load via task0 serial read
          if (node_n1 && !l1) {
            E1("mymap(l1) failed for mpi=%d, skipping", mpi_i);
#ifdef CONFIG_MPI
            if (mpi_n>1) MPI_Abort(MPI_COMM_WORLD,1);
#endif
            continue;
          }
        } /* if(!l1) */

        MPI_Barrier(MPI_COMM_WORLD);
        /* ToDo: add vec_malloc[_block](tvec,m_v0) to vector.h */
        if((algo&7)<2){ /* we need this 2 vectors for Lanczos algorithm */
          int localerror=0, sumerror=0;
          /* ToDo: alloc within threads or remove thread-blocks */
          if ((!mpi_i) && (!v0))
            printf(" a0.malloc %s%3s(%dB) v0,v1,%d*ev a %8.2eB/mpitask\n",
                   ((VecType&4)?"c":" "), ((VecType&1)?"dbl":"flt"),
                   (int)Csize, nev, 1.*b_ofs[pt_n] * (int)Csize);
          /* JS2012-10 for mpi_n > n1  b_ofs */
          // 2015-03-05 touching memory to avoid vm overbooking by storeH
          v0=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_v0); touchvec(v0);
          v1=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_v1); touchvec(v1);
          // ToDo: mpi_sum num_errors 0=ok, 1...mpi_n = num failes
          // compute log2 group of nodenumber or simply sum of errors
          // 2012-11-01: no errors if n1 < mpi_n (b_ofs==0)
          if (b_ofs[pt_n] && ((!v0) || (!v1))) sumerror=localerror=1;
#ifdef CONFIG_MPI
          MPI_Allreduce(&localerror, &sumerror, 1,
                        MPI_INT, MPI_SUM, MPI_COMM_WORLD);
#endif
          if ((!mpi_i) && sumerror)
            F2(" malloc failed on %d of %d nodes", sumerror, mpi_n);
          // if (sumerror) continue;
          // test 8MB allocatable for MPI 32?
        }
        MPI_Barrier(MPI_COMM_WORLD);
        if(((algo&7)==2) && nev<1){ nev=1; E0("nev is 0, set 1"); }
        for(i=0;i<nev;i++){
          int vvv2=vvv; if (mpi_i>3) vvv&=~2; /* avoid to much output */
          if ((!mpi_i) && (!ev[i]) && (vvv&128)) printf(" malloc ev%d", i);
          ev[i]=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_ev[i]);
          touchvec(ev[i]);
          vvv=vvv2;
        }
// sprintf(m_n2.name, "tmp/n2_%04d.map", mpi_i);
// norm2=(ATnorm2 *)mymalloc(node_n1*sizeof(ATnorm2),&m_n2);base2ok=0;
        /* if( checkpoint < 3 ) */ 
        { /* norm2 is not mapped! change it ??? */
          double tm1=fstime();
          ini_norm2();
          if (!mpi_i) printf("PRF: n2.t[m]= %6.2f", (fstime()-tm1)/60.);
        } /* else load_l1(l1,n1); --- mmap used */
        MPI_Barrier(MPI_COMM_WORLD);
        outbasis(stdout);

        /* 2015-04-08 malloc Hbuffer (rest of memory), fail means its slower */
/* ReserveMB need for libMPI
  # sc072.asgard extrapolated to kautz-mpich2: 972*6/18*64KB=20.7MB 5+49
  srun -p sca -n 72 bash -c "(ulimit -v 23232;./spin)"
  srun -p sca -n 55 bash -c "(ulimit -v 23232;./spin)"  # +64K/18tasks
  srun -p sca -n 54 bash -c "(ulimit -v 23168;./spin)"
  srun -p sca -n 37 bash -c "(ulimit -v 23168;./spin)"  # +64K/18tasks
  srun -p sca -n 36 bash -c "(ulimit -v 23104;./spin)"
  srun -p sca -n 19 bash -c "(ulimit -v 23104;./spin)"  # +64K/18tasks
  srun -p sca -n 18 bash -c "(ulimit -v 23040;./spin)"
  srun -p sca -n  1 bash -c "(ulimit -v 23040;./spin)"
  # comp2.ompi-1.6.5-ft reg-mem-warning ???
  mpiexec -np $((4* 8)) bash -c "(ulimit -v 185000;./spin)" # OK +4M/16 +MPIw
  mpiexec -np $((4* 4)) bash -c "(ulimit -v 181000;./spin)" # OK +4M/8 +MPIwarn
  mpiexec -np $((4* 2)) bash -c "(ulimit -v 177000;./spin)" # OK +2M/4 +MPIwarn
  mpiexec -np $((4* 1)) bash -c "(ulimit -v 175000;./spin)" # OK + reg mem limit low
  # comp2.ompi-1.8.4 8+32 2015-04
  mpiexec -np $((6*16)) bash -c "(ulimit -v 255000;./spin)" # OK +138MB
  mpiexec -np $((5*16)) bash -c "(ulimit -v 259000;./spin)" # OK +142MB
  mpiexec -np $((4*16)) bash -c "(ulimit -v 259000;./spin)" # OK +142MB
  mpiexec -np $((3*16)) bash -c "(ulimit -v 254000;./spin)" # OK +137MB
  mpiexec -np $((2*16)) bash -c "(ulimit -v 254000;./spin)" # OK +137MB
  mpiexec -np $((1*16)) bash -c "(ulimit -v 254000;./spin)" # OK 
  mpiexec -np $((1* 2)) bash -c "(ulimit -v 256000;./spin)" # OK +139MB
  mpiexec -np $((6* 1)) bash -c "(ulimit -v 121000;./spin)" # OK +4MB
  mpiexec -np $((4* 1)) bash -c "(ulimit -v 122000;./spin)" # OK 
  mpiexec -np $((3* 1)) bash -c "(ulimit -v 122000;./spin)" # OK 
  mpiexec -np $((2* 1)) bash -c "(ulimit -v 123000;./spin)" # OK +6MB
  mpiexec -np $((1* 1)) bash -c "(ulimit -v 117000;./spin)" # OK 
  bash -c "(ulimit -v 117000;./spin)" # OK (but mpicc)
  # comp1.ompi-1.8.4 5+49
  mpiexec -np 8 bash -c "(ulimit -v 245000;./spin)" # OK +4M/8
  mpiexec -np 7 bash -c "(ulimit -v 241000;./spin)" # OK +4M
  mpiexec -np 6 bash -c "(ulimit -v 237000;./spin)" # OK +4M
  mpiexec -np 5 bash -c "(ulimit -v 233000;./spin)" # OK +3M
  mpiexec -np 4 bash -c "(ulimit -v 230000;./spin)" # OK +4M
  mpiexec -np 3 bash -c "(ulimit -v 226000;./spin)" # OK +2M
  mpiexec -np 2 bash -c "(ulimit -v 223355;./spin)" # OK +156M
  mpiexec -np 2 bash -c "(ulimit -v 223354;./spin)" # hanging SH!
  mpiexec -np 1 bash -c "(ulimit -v  67464;./spin)"
  (ulimit -v 84000;./spin -t8) # ca. 8MB/task
  (ulimit -v 67000;./spin -t6)
  (ulimit -v 51000;./spin -t4)
  (ulimit -v 34000;./spin -t2)
  (ulimit -v 26000;./spin -t1)
 */     /*  sub_ini() must be called before! */
        /* hfmax=maxfile/8B per core or thread! 600MB..24GB  */
        if (STOREH == 0 /* 2017-01 dont want it if disk used ?? */ 
         && (algo&7)<3) /* 2016-11-10 malign alloc big Hbuf for fulldiag */
        { size_t memHbuf= ((size_t)hfmax)*(Hsize/HBLen); /* 2015-05-11 32bit */
          /* n1=1 should give 1 per task, not 0 (therefore n1/mpi_n+1) */
          /* ToDo2016: replace NZXMAX by real nzxmax as better approx */
          /* estimate upper border per Core, do not alloc more than that */
          size_t maxHbuf= (n1/(mpi_n*pt_n)+1)*NZXMAX*Hsize/HBLen; 
          /* 2017-01-12 override maxHbuf by hfmax (if we want no SH) */
 // ToDo         if (hfmax>0) maxHbuf= ((size_t)hfmax)*(Hsize/HBLen);
 // ToDo some disk + memory 2017
          /* ToDo: do a lower estimation of NZXMAX */
          /* ToDo2016: check size_t overflow! 32bit.max=16G */
          /* hfmax =(ULong)(FloatByte*HBLen/Hsize) */
          int lok=0, gok=0; char *reserve;
#if STOREH != 0 /* disk, 2017-01-13 ToDo: on runtime maxhfile+maxhmem */
          maxHbuf=0; /* temporary fix for high level on disk runs 2017-01 */
#endif
          if (1.*memHbuf != hfmax*(1.*Hsize)/HBLen && !mpi_i)
              E0("hfmax does not fit to size_t");
          /* biggest we have is 512GB/32=16GB per core comp2_2015.mpi=32 */
          if (!hfmax) { /* range 1/8..8/8 z.B. 256MB..2GB */
            /* better set it on daten.i? ulimit -v or maxRAM/threads */
            // memHbuf= (16L<<30); // 64bit++ dflt-range: 16GB/thread...1GB/node
            /* 2016-04-21: 16GB is not enough, when part of cores used */
            memHbuf = maxHbuf;
            /* reduce if its bigger than expected maximum */
            /* not needed now: if (memHbuf > maxHbuf) memHbuf = maxHbuf; */
            if (memHbuf < Hsize) memHbuf = Hsize; /* small models */
          } /* hfmax==0 autodetect */
          if (!mpi_i){
if (mpi_n>1) {              
  printf("\nWARNING: spinpack may crash here, if libmpi causes OutOfMemory");
  if (pt_n<B_NUM)           
    printf("\nWARNING: ... but B_NUM=CFG_PTHREAD is set bigger than needed,"
            " please reduce!");
  else      
    printf("\nWARNING: ... if so minimize NZXMAX, reduce NUM_AHEAD_LINES"
            " or maxmem (old:maxfile)");
}
#ifdef ReserveMB
            printf("\nDBG: ReserveMB= %d (ompi chkpt ..., not eaten by SH)",
              ReserveMB); 
            /* Todo: use also --mca mpi_preconnect_mpi 1 or speedtest.mpi 
             */
#endif
            printf("\nDBG: maxfile[B/core]=hfmax*%d= %8.2e  autoMax= %8.2e",
                   (int)(Hsize/HBLen), 1.*memHbuf, 1.*maxHbuf);
          }
          /* better let some free memory for mpich xmalloc calls! + min16MB
           *   ompi-1.4 + pthread? failed for free=16MB
           *  allocating a reserve and free it for mpi-use does not
           *    seems to help in every case (bugs 2015-04-06), but no choice
           *   hopefully it crashs at start and not later as in v2.49
           * ToDo: replace NZXMAX by nzxmax (computed on runtime)
           */
          gok=lok= 1; /* gok: global num threads with reserve = OK */
#ifdef ReserveMB
          reserve=(char *) malloc(ReserveMB<<20); /* MPI reserve memory */
          /* if no MPI, reduce reserve for stack + functions */
          gok=lok= (reserve) ? 1 : 0; /* thread-global and local OK */
#endif
#ifdef CONFIG_MPI
          MPI_Allreduce(&lok, &gok, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
#endif
          if (gok<mpi_n && !mpi_i) {
            printf("\nWARN: OOM %d no reserve %dMB ...", mpi_n-gok, ReserveMB);
#ifdef CONFIG_MPI
            /* 2015-04-09 8MB was not enough ompi-1.4 */
            printf("\nWARN: ... on OOM MPI_Sendrecv may hang later ...");
#endif
          }
          if (reserve && !mpi_i)
            printf("\nDBG: reserve=OK, try Hbuf/thread (min(1/16...1)* %.2e",
                 1.*maxHbuf);
          if ((!reserve) && !mpi_i) printf("\nDBG: reserve=FAIL");
          gok=0; /* reused to count mpi_n*pt_n Hbuffer-mallocs */
          if (reserve) /* no Hbuffer without reserveMB */
          /* try to reduce memHbuf while bigger than expected maximum/8 */
          /* ToDo: NZXMAX should be .le. nw,
           *   1. start random cfg
           *   2. get all H*cfg, pick subcfg with most H*subcfg nzx-terms
           *   3. repeat until 8*same_nzx or no bigger nzx
           *  or rekursiv scan all in lower sub space (smaller nu on JJ)
           *   and add up-spins and repeat add up-spins on last biggest nzx
           * allow setting by hand (exotic lattices)
           * sometimes Nw is big (N^2) and NZX is computet big too
           *   without syms memHbuf/16 is still 10e9 to big and everything
           *    failes so we add 400MB as a Minimum tryable 
           * 2016-01-22 add hfmax/16 too (hand max limit)
[odin:05295] [ 0] /lib64/libpthread.so.0 [0x35e260ebe0] # CentOS-5.8
[odin:05295] [ 1] /usr/lib64/openmpi/1.4-gcc/lib/libopen-pal.so.0(\
opal_memory_ptmalloc2_int_malloc+0xb21) [0x35e2236f31]
web:   mpirun -x OMPI_MCA_memory_ptmalloc2_disable=1 ...   # hilft!
      oder
       OMPI_MCA_memory_ptmalloc2_disable=1 mpirun  ...  # set in /etc/bashrc
        bad ptmalloc2 part of glibc? 
      https://www.open-mpi.org/community/lists/users/2012/09/20249.php
          */
          while (memHbuf > maxHbuf/16
             || (hfmax && memHbuf>hfmax/8) || memHbuf>400e6){
            if (vvv && !mpi_i)
              printf("\nDBG: try alloc %2d*%2d* %8.2e Byte/core for Hbuffer",
                    mpi_n, pt_n, 1.*memHbuf);
            /* ToDo2016: for OMP better malloc(pt_n*memHbuf) +local_access */
            for (lok=0,i=0;i<pt_n;i++) {
              pipeHbuffer[i]= (char *) malloc(memHbuf); // malloc per thread
              if (pipeHbuffer[i]) lok++;
            } /* MPI_BAND = bitwise and */
            gok=lok; /* number of MPI-threads, where malloc succeeded */
#ifdef CONFIG_MPI
            MPI_Allreduce(&lok, &gok, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
#endif
            if (gok==mpi_n*pt_n && !mpi_i) printf(" = OK");
            if (gok==mpi_n*pt_n) break; /* fine, leave the loop here */
            /* malloc all or nothing! */
            if (gok!=mpi_n*pt_n && !mpi_i) 
              printf(" = FAIL %d of %d", mpi_n*pt_n-gok, mpi_n*pt_n);
            for (i=0;i<pt_n;i++) {
              if (pipeHbuffer[i]) { free(pipeHbuffer[i]); pipeHlen[i]=0; }
            }
            // 2015-06 better next try (ubound+lbound)/2 in case of gok>0
            if (gok>0) {
              memHbuf = ((2*gok+1)*memHbuf/(pt_n*mpi_n))/2; // +0.5
              if (memHbuf<maxHbuf) maxHbuf = memHbuf; // shift lower bond
              /* slowdown on mpi-systems where one node has less free mem */
            }
            /* remove lower 1 bits, binary rounding */
            else {
              while ((memHbuf&(memHbuf-1))>memHbuf-memHbuf/4)
                      memHbuf &= (memHbuf-1);
              if (memHbuf&(memHbuf-1)) memHbuf&=(memHbuf-1);
              else                     memHbuf>>=1;  /* -25%...-50% */
            }
            if (!mpi_i) 
              printf(", next= %.2e", 1.*memHbuf);
          } /* end_while try to alloc rest of mem for Hbuffer */
          if (gok==mpi_n*pt_n) for (i=0;i<pt_n;i++) pipeHlen[i]= memHbuf;
          if (gok==mpi_n*pt_n && vvv && !mpi_i)
            printf("\nDBG: allocated %8.2e Byte/core for Hbuffer",1.*memHbuf);
          if (reserve && gok!=mpi_n*pt_n && vvv && !mpi_i)
            printf("\nDBG: alloc %8.2e Byte/core Hbuffer failed",1.*memHbuf);
          if (reserve) free(reserve);
        } // local memHbuf=f(hfmax)

        /* ToDo: free Hbuffer after computation?, error handling statH_out */
        if ((algo&7)<3) { storeh(hbuf); } /* compute H + store to mem/disk */
        ERR(-1);	/* reset error + print old errors */

        MPI_Barrier(MPI_COMM_WORLD);
        test_all_sym();	/* only array-ov test Mai99 */
#if 0
        if((algo&7)<2){ /* ToDo: remove! its twice!? */
          if ((!mpi_i) && (!v0)) printf("\n malloc v0,v1");
          /* ToDo: alloc within threads or remove thread-blocks */
          if(!v0) v0=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_v0);
          if(!v1) v1=(scplx *)mymalloc(b_ofs[pt_n]*Csize,&m_v1);
          if ((!mpi_i) && ((!v0) || (!v1))) E0(" malloc failed");
        }
        MPI_Barrier(MPI_COMM_WORLD);
#endif
        /* degeneracy loop for different startvectors (2016-04: +FTLM) */
        for(iv=1;iv<=max_ea;iv++){ /* iv = index number of EA start vector */
          if (err_fatal) break;   /* stop on fatal errors */
          if (iv>1) checkpoint=5; /* reset checkpoint old=3 */
          if((algo&7)==0){ /* a0 = Lanczos tridiag */
            /* pseudo random startvector depends only on iv! */
            /* here is the right place to choose iv2=random instead of iv */
            randvec(v1,startvec,iv-1+(startvec&8)*nruns); /* bug? svec&7? */
            itr=inc1(nev,ne0,v0,v1,hbuf);               /* EW(T),EV(T)  */
#ifdef CFG_FTLM
            if (!mpi_i) printf("\nFTLM: next round %d/%d",iv,max_ea);
            continue; /* Todo2016: tridiag enough for degeneracy too? */
#endif
            if (checkpoint<9 || iv>1) {	/* v1.7.14=5, v2.44=6 */
              if(nev){ randvec(v1,startvec,iv-1+(startvec&8)*nruns);
                       incv1(nev,v0,v1,ev,itr,hbuf); }          /* EV(H) */
              if ( iv<=1 ) {checkpoint=9;if(!mpi_i)printf(" checkpoint9");}
            } else { // checkpoint9 && iv==1 = load_ev
              if (!mpi_i) printf("\n checkpoint9=%d=reload_ev,",checkpoint);
              if(nev>0)loadvec(ev[0],FileGS,iv,-1);
            } /* reload j? chkpt */
          } else if((algo&7)==1){ /* a1 = 1-step-lanzcos */
            randvec(v0,   startvec,iv-1+(startvec&8)*nruns);
            itr=lanz1(v0,v1,ev[0],hbuf);   /* 1-step-lanzcos EW,EV(H) */
          } else if((algo&7)==2 && nev>0){ /* a2 = 2x2 diag, only needs ev[0] */
            randvec(ev[0],startvec,iv-1+(startvec&8)*nruns);
            itr=lanz2(ev[0],hbuf);         /* 2x2-Matrix     EW,EV(H) */
          } else if((algo&7)==3 && nev>0){ /* a3 load_EV from disk */
            printf("\n algo=3: load vectors from disk");
            /* %NEV to avoid warning "array access above bounds" for NEV=1 */
            if (loadvec(ev[0],FileGS,iv,-1)) continue; /* v1.7.13+ */
            if (nev>1)
            if (loadvec(ev[1%NEV],FileEX,iv,-1)) continue; /* v2.28 */
          } /* <O> zu H' */
          else if((algo&7)==4){
             if (fulldiag(ev,nev,hbuf[0],ew)) break; }/* a4, 2016-11 +break */
          /* 2016-11 a4 malloc fail gives no output erg= because break */
          if (!mpi_i) {
            printf("\n erg= "); outew(stdout,ew,NEW);
            if (vvv) fprintf(stdout," t+=%s",sclock(~0l));
            printf("\n E0/N= %12.8f E0/nw= %12.8f",D(ew[0]/nn),D(ew[0]/nw)); /* Feb99 */
            if (ne0)
            printf("\n E%d/N= %12.8f E%d/nw= %12.8f E%d= %12.8f",ne0,D(ew[ne0]/nn),ne0,D(ew[ne0]/nw),ne0,D(ew[ne0])); /* Feb99 */
          }
          MPI_Barrier(MPI_COMM_WORLD);

          if((algo&7)==4 || (!nev))break; /* brk here if noEV or a4-diag */
          /* continue with EV-handling and degeneracy loop */
          
          if (v0) /* not for a2 2013-01 */
          for(ii=0,j=1;j<iv;j++){ /* orthogonalize ev against saved vectors */
            loadvec(v0,FileGS,j,-1);
            if((ii=iortho(ev[0],v0))==1)break;
          } /* stop if we have found all orthogonal ev's */
          if (ii!=0) break;
          
#ifdef CONFIG_Fidelity  /* better check for option? */
          /* 1 - <\Psi(a-da)|\Psi(a)> (or 1-|<p1|p2>| because of sign) */
          /* try this with J1-J2-model, j2= 0.09,0.1,0.19,0.2 */
          if (v0) { double r1;
            /* JS-2011-03-25 Fidelity = Overlap to other GS vector */
            loadvec(v0,FileGS,iv,-1);  /* tmp_shared/d%d.vec  */
            /* if (!mpi_i) printf("\n old vector:"); x_out(stdout,v0,0,x_p); */
            /* if (!mpi_i) printf("\n new vector:"); x_out(stdout,ev[0],0,x_p); */
            r1 = 1 - fabs(sprod(v0, ev[0]));
            savevec(ev[0],FileGS,iv,-1);
            if (iv>1 && !mpi_i)
              printf("WARNING: ea>1 + FIDELITY does not work!\n");
            if (!mpi_i)
              printf("\n fidelity= %12.9f (1-overlap to last computed EV)\n",
                     r1);
          }
#endif
          /* we only save if its necessary (very slow) */
          /* %NEV to avoid warning "array access above bounds" for NEV=1 */
          if (max_ea>1)   savevec(ev[0    ],FileGS,iv,-1);
          if (nev>1) {    savevec(ev[1%NEV],FileEX,iv,-1); } /* nev<=NEV! */
          /* 0,1,2 */
          if((algo&7) <= 4)
            for(j=0;j<nev;j++){
              if (!mpi_i) printf("\n\nstart chk3.ev%d (nev=%d NEV=%d) ...",
                                 j, nev, NEV);
              chk3(ev[j],v1,iv,hbuf);
              if((algo&7)==3) break; /* ev0 only */
            }

#ifdef CONFIG_DIMER_CHECK
          if (nev>2) { /* Richter extension to check for dimers */
            /* we break the symmetry by adding lowest two states */
            printf("\n dimer-check+ (ev0+ev1)/2:");
            clrvec(ev[2]);
            add_av(ev[2],+sqrt(0.5),ev[0]);
            add_av(ev[2],+sqrt(0.5),ev[1]);
            normvec(ev[2]); // not needed?  
            chk3(ev[2],v1,1,hbuf);
            printf("\n dimer-check- (ev0-ev1)/2:");
            clrvec(ev[2]);
            add_av(ev[2],+sqrt(0.5),ev[0]);
            add_av(ev[2],-sqrt(0.5),ev[1]);
            normvec(ev[2]);
            chk3(ev[2],v1,1,hbuf);
          }
#endif
        } /* loop: iv=1...max_ea */
        if (max_ea>1 && !mpi_i)
          printf("\n degree_of_degeneracy=%d<%d\n",iv-1,max_ea);
      }
 /* label_end: */

      MPI_Barrier(MPI_COMM_WORLD);


      /* 2015-04-08 free Hbuffer */
      for (i=0;i<pt_n;i++){if(pipeHbuffer[i])free(pipeHbuffer[i]);pipeHlen[i]=0;}
      
// norm2=(ATnorm2 *)myfree((void *) norm2, &m_n2);base2ok=0;
      l2b=    (ULong *)myfree((void *) l2b, &m_l2);
      if (l1) { // ??? 2015-02-26, change fl1 to l1 2016-03-22 fix memleak
        fl1=fopen(FILE_l10,"r+b"); /* 2016-03 added to fix reenable myunmap */
        if(algo&8) myunmap(l1,fl1,n1max*(off_t)Bsize,0);
        else       myunmap(l1,fl1,n1   *(off_t)Bsize,0);
      } // ???
      if (fl1) fclose(fl1);
      l1=NULL; fl1=NULL;
      {
        int vvv2=vvv; if (pt_n*mpi_n>4) vvv&=~2; /* avoid to much output */
        if ((!mpi_i) && (!ev[0])) printf(" free ev");
        for(i=0;i<nev;i++) {
          ev[i]=(scplx *)myfree((void *)ev[i],&m_ev[i]);
        }
        if ((!mpi_i) && (!v0)) printf(" free v0,v1");
        v1=(scplx *)myfree((void *)v1,&m_v1);
        v0=(scplx *)myfree((void *)v0,&m_v0);
        vvv=vvv2;
      }
      n1=0l; /* reset n1 */
      /* reset checkpoint, next start will do new computation */
      // echo 0 > tmp_shared/chkpt
      if (!mpi_i) 
        printf(" reset checkpoint %d to 0 for next cycle", checkpoint);
      checkpoint=0;
      nruns++;  /* can be used for other srand on next computation */
      if(!mpi_i){FILE *f;f=fopen(FCHKPT,"wt");fputs("0\n",f);fclose(f);}
      
      // --- JS2012-05 rausgenommen, nur laestige outputs ohne viel Sinn
      //MPI_Barrier(MPI_COMM_WORLD);
      //if (!mpi_i) printf("\n --- next run + mpi-balance-check:\n");
      //MPI_Barrier(MPI_COMM_WORLD);
      // if (mpi_i<8 || mpi_i>=mpi_n-8) // first and last 8
      //   printf(" %d", mpi_i);
      //MPI_Barrier(MPI_COMM_WORLD);
      if (!mpi_i) printf("\n");  /* flush */

    } /* end a#-algo-command if(s1)-lines */
    else {
      if (!mpi_i)
        E2("infile L%d: unknown parameter or command: %s", in1line, s1);
    }

    if (err_fatal){
      err_fatal=0; 
      if ((mpi_i&(mpi_i-1))==0 && mpi_i<32) /* 2015-02-26 log2 output */ 
        printf("\n reset err_fatal (mpi=%d)\n", mpi_i);
    }

  } /* for(;;) scan input lines from daten.i */ 
  if (!mpi_i) printf(" close(%s)", name_file_i);
  MPI_Barrier(MPI_COMM_WORLD);
  if (in1) { fclose(in1); in1=NULL; }
  if (!mpi_i) printf(" ok\n");
  b_free(); /* release dynamic arrays b_ofs[], b_len[] block.c dummy */
  for (block=0; block<pt_n; block++) free_hbuf(hbuf[block]); /* 2014-07-17 */
  MPI_Barrier(MPI_COMM_WORLD); /* sync output of threads 2012-05 */

#ifdef CONFIG_PTHREAD
#ifdef _OPENMP
#pragma omp critical
{
  for (i=0;i<pt_n;i++) main_pt[i].task=-1;    /* finalize */
}
#else /* OPENMP */
  /* send no further tasks to threads */
  pthread_mutex_lock(&pt_mutex1);
  for (i=0;i<pt_n;i++) main_pt[i].task=-1;    /* finalize */
  pthread_cond_broadcast(&pt_cond1);
  pthread_mutex_unlock(&pt_mutex1);
  for (i=0;i<pt_n;i++) {          /* join main threads */
    pthread_join(pt[i],NULL);
  }
#endif /* ~OPENMP */
} /* pragma omp master|single test at L4539,4547 2017-01-09 */
} /* pragma omp parallel */
#endif /* PTHREAD */
  sub_end();	/* kill child, dummy pipe.c */
#if defined(CONFIG_MPI)
  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();
#endif
  return(0);
}

#ifdef F2C
/* lapack-3.0 for linux needs f2c and f2c needs MAIN__() */
/* if you know a better solution, please tell me */
#if (VecType & 4) || defined(__cplusplus)
extern "C"
#endif
int MAIN__(){ return main(0,NULL); }
#endif
